/* GAUGE^3D Asteroids
 * Placed in the Public Domain by Kenton Varda <temporal@gauge3d.org>.
 * <http://www.gauge3d.org>
 */

// ===================================================================

// includes ==========================================================
#include <gauge3d/stddefs.h>

// system ------------------------------------------------------------
#include <iostream>
#include <vector>
#include <typeinfo>
#include <stdlib.h>
#include <time.h>
using namespace std;

// GAUGE^3D ----------------------------------------------------------
#include <gauge3d/common/error.h>
#include <gauge3d/osabstraction/messagebox.h>
#include <gauge3d/3dmath/angles.h>
#include <gauge3d/files/directory.h>
#include <gauge3d/files/configfile.h>
#include <gauge3d/files/stringfile.h>
#include <gauge3d/plugins/objectloader.h>
#include <gauge3d/renderer/renderer.h>
#include <gauge3d/graphics/spheremodel.h>
#include <gauge3d/graphics/billboardsprite.h>
#include <gauge3d/input/messagequeue.h>
#include <gauge3d/input/inputmap.h>
#include <gauge3d/scheduler.h>
#include <gauge3d/engine.h>

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3DAsteroids {

using namespace GAUGE3D;

//generate random float between low and high.
GCoordinate Random(GCoordinate low, GCoordinate high)
{
	return ((GCoordinate)rand() / RAND_MAX) * (high - low) + low;
}

//size of playfield = 16x12
const GCoordinate MAX_X = 8.0f;
const GCoordinate MAX_Y = 6.0f;

class Entity;   //an object in space

pGRenderer gTheRenderer;
pGScene gTheScene;
Entity* gEntityList = NULL;  //linked list

//this class loads up the asteroids config file.
class AsteroidsConfig
{
	pGVirtualDirectory mConfig;  //the config tree
	
	int mResX;         //screen width
	int mResY;         //screen height
	bool mFullscreen;  //fullscreen or not
	
	//3d models
	pGModel mShipModel;
	vector<pGModel> mAsteroidModels;
	pGModel mMissileModel;
	
	//textures
	pGTexture mShipTexture;
	vector <pGTexture> mAsteroidTextures;
	pGTexture mMissileTexture;
	pGTexture mBackground[4];
	
	GCoordinate mBackgroundBrightness;
	
	//Other settings...
	GCoordinate mInitialAsteroidRadius;
	GCoordinate mAsteroidSpeed;
	
	GCoordinate mMissileSpeed;
	GCoordinate mMissileTime;
	
	GCoordinate mShipScale;
	GCoordinate mShipRadius;
	tTime       mShipSpawnTime;
	GCoordinate mShipSpeed;
	GCoordinate mShipTurnSpeed;
	GCoordinate mShipFireRate;
	
	int mStartLevel;
	
	GString mPreferredRenderer;
	
	//attempt to load a given model.
	//modelName is a file name.  The file will be located inside the game's
	// resource directory, which is different on every OS.
	pGModel LoadModel(GString modelName)
	{
		if(modelName == "sphere")
			return new GSphereModel;
		
		pGModel model;
		
		//open the file
		pGFile file = gResourceDir->OpenFile(modelName);
		if(file == NULL)
			GAUGE3D_FATAL_ERROR(GString("Could not open model file ") + modelName
			                    + "\nHave you installed the game data?");
		
		//find a plugin which can load the file
		model = gMainObjectLoader->LoadObjectForFile(file, "GModel");
		
		if(model == NULL)
			GAUGE3D_FATAL_ERROR(GString("Could not find plugin for model ") + modelName);
		
		return model;
	}
	
	//attempt to load a texture.  Just like LoadModel(), above.
	pGImage LoadTexture(GString imageName)
	{
		if(imageName == "blank")
			return NULL;
		
		pGImage image;
		
		//open the file
		pGFile file = gResourceDir->OpenFile(imageName);
		if(file == NULL)
			GAUGE3D_FATAL_ERROR(GString("Could not open image file ") + imageName
			                    + "\nHave you installed the game data?");
		
		//find a plugin which can load the file
		image = gMainObjectLoader->LoadObjectForFile(file, "GImage");
		
		if(image == NULL)
			GAUGE3D_FATAL_ERROR(GString("Could not find plugin for image ") + imageName);
		
		//cache the texture with the renderer
		pGTexture texture = gTheRenderer->NewTexture();
		texture->Image(image);
		
		return texture;
	}
	
	GString GetSetting(GString name, GString defaultValue)
	{
		pGFile file = mConfig->OpenFile(name);
		if(file == NULL)
		{
			cerr << "Config option not found: " << name.AsciiData() << endl;
			return defaultValue;
		}
		else
			return (GString)*file;
	}
	
public:
	AsteroidsConfig()
	{
		pGFile configFile;
		bool saveConfig = false;
		
		configFile = gConfigDir->OpenFile("asteroids.cfg");
		if(configFile == NULL)
		{
			//asteroids.cfg was not found in the user's config directory.  So, we want
			// to open the default config file and save a copy of it into the user's
			// config dir.
			configFile = gResourceDir->OpenFile("asteroids-defaults.cfg");
			if(configFile == NULL)
			{
				GAUGE3D_FATAL_ERROR("Could not open asteroids-defaults.cfg.\nHave you installed the game data?");
			}
			else
				saveConfig = true;
		}
		
		mConfig = GConfigFile::Load(configFile);  //convert to directory structure
		if(mConfig == NULL)
			GAUGE3D_FATAL_ERROR("Failed to load config file.");
		
		if(saveConfig)
		{
			//save copy of config to user's home directory
			pGFile newFile = gConfigDir->AddFile("asteroids.cfg");
			if(newFile != NULL)
				GConfigFile::Save(newFile, mConfig);
		}
		
		//If the config file is missing stuff, this next part could segfault.
		// This will be fixed in the future, when GAUGE starts using exceptions
		// more regularly.  I just did not want to bother with all the error
		// handling here.
		mResX = (int)GetSetting("video/resx", "800");
		mResY = (int)GetSetting("video/resy", "600");
		GString fullscreen = GetSetting("video/fullscreen", "no");
		if(fullscreen == "yes" || fullscreen == "true")
			mFullscreen = true;
		else
			mFullscreen = false;
		
		mShipModel      = LoadModel(GetSetting("models/ship", "ship.gmdl"));
		mMissileModel   = LoadModel(GetSetting("models/missile", "missile.gmdl"));
		
		mInitialAsteroidRadius = (float)GetSetting("logic/initial asteroid radius", "1.0");
		mAsteroidSpeed         = (float)GetSetting("logic/asteroid speed", "1.0");
		
		mMissileSpeed          = (float)GetSetting("logic/missile speed", "8.0");
		mMissileTime           = (float)GetSetting("logic/missile time", "1.0");
		
		mShipScale             = (float)GetSetting("logic/ship scale", "0.4");
		mShipRadius            = (float)GetSetting("logic/ship radius", "0.5");
		mShipSpawnTime         = (float)GetSetting("logic/ship spawn time", "2.0");
		mShipSpeed             = (float)GetSetting("logic/ship speed", "5.0");
		mShipTurnSpeed         = (float)GetSetting("logic/ship turn speed", "4.0");
		mShipFireRate          = (float)GetSetting("logic/ship fire rate", "20");
		
		mStartLevel            = (int)GetSetting("logic/start level", "1");
		
		mBackgroundBrightness  = (float)GetSetting("textures/background brightness", "0.2");
		
		mPreferredRenderer = GetSetting("video/preferred renderer", "opengl");
		
		GString modelName;
		pGIStream is;
		pGFile setting = mConfig->OpenFile("models/asteroid");
		if(setting == NULL)
			setting = new GStringFile("asteroid1.gmdl asteroid2.gmdl asteroid3.gmdl asteroid4.gmdl asteroid5.gmdl");
		is = setting->GetInputStream();
		is->ReadMode(GIStream::ASCII);
		while((*is) >> modelName)   //read in multiple asteroid models
			mAsteroidModels.push_back(LoadModel(modelName));
	}
	
	//textures have to be loaded after the renderer is started, but the rest
	// of the config file has to be loaded *before* that.  So, we have to use two
	// separate functions.
	void LoadTextures()
	{
		mShipTexture      = LoadTexture(GetSetting("textures/ship", "ship.jpeg"));
		mMissileTexture   = LoadTexture(GetSetting("textures/missile", "missile.jpeg"));
		
		GString imageName;
		pGFile setting = mConfig->OpenFile("textures/asteroid");
		if(setting == NULL)
			setting = new GStringFile("rock1.jpeg rock2.jpeg rock3.jpeg rock4.jpeg rock5.jpeg");
		pGIStream is;
		is = setting->GetInputStream();
		while((*is) >> imageName)   //read in multiple asteroid textures
			mAsteroidTextures.push_back(LoadTexture(imageName));
		
		//load background textures
		setting = mConfig->OpenFile("textures/background");
		if(setting == NULL)
			setting = new GStringFile("nebula-ul.jpeg nebula-ur.jpeg nebula-ll.jpeg nebula-lr.jpeg");
		is = setting->GetInputStream();
		(*is) >> imageName;
		mBackground[0] = LoadTexture(imageName);
		(*is) >> imageName;
		mBackground[1] = LoadTexture(imageName);
		(*is) >> imageName;
		mBackground[2] = LoadTexture(imageName);
		(*is) >> imageName;
		mBackground[3] = LoadTexture(imageName);
	}
	
	~AsteroidsConfig() {}
	
	inline pGVirtualDirectory ControlList() {return mConfig->OpenSubdir("control list");}
	inline pGVirtualDirectory ControlMap()  {return mConfig->OpenSubdir("control map");}
	
	inline int  ResX()       {return mResX;}
	inline int  ResY()       {return mResY;}
	inline bool Fullscreen() {return mFullscreen;}
	
	inline pGModel ShipModel() {return mShipModel;}
	pGModel AsteroidModel()
	{
		//select a random asteroid model
		return mAsteroidModels[rand() % mAsteroidModels.size()];
	}
	inline pGModel MissileModel() {return mMissileModel;}
	
	inline pGTexture ShipTexture() {return mShipTexture;}
	pGTexture AsteroidTexture()
	{
		//select a random asteroid texture
		return mAsteroidTextures[rand() % mAsteroidTextures.size()];
	}
	inline pGTexture MissileTexture() {return mMissileTexture;}
	inline pGTexture Background(int num) {return mBackground[num];}
	
	inline GCoordinate BackgroundBrightness() {return mBackgroundBrightness;}
	
	inline GCoordinate InitialAsteroidRadius() {return mInitialAsteroidRadius;}
	inline GCoordinate AsteroidSpeed() {return mAsteroidSpeed;}
	
	inline GCoordinate MissileSpeed()  {return mMissileSpeed;}
	inline GCoordinate MissileTime()   {return mMissileTime;}
	
	inline GCoordinate ShipScale()     {return mShipScale;}
	inline GCoordinate ShipRadius()    {return mShipRadius;}
	inline tTime       ShipSpawnTime() {return mShipSpawnTime;}
	inline GCoordinate ShipSpeed()     {return mShipSpeed;}
	inline GCoordinate ShipTurnSpeed() {return mShipTurnSpeed;}
	inline GCoordinate ShipFireRate()  {return mShipFireRate;}
	
	inline int StartLevel() {return mStartLevel;}
	
	inline GString PreferredRenderer() {return mPreferredRenderer;}
};

AsteroidsConfig* gConfig;

//Entity is the base class for Ship, Asteroid, and Missile.  It
// represents an abstract object in 3D space.  It includes
// functionality for simple sphere-based collision detection.
class Entity: public GObject
{
public:
	//typeid didn't want to work...
	enum tType
	{
		SHIP,
		ASTEROID,
		MISSILE
	};
	
private:
	pGSprite    mSprite;    //the renderer sprite for this entity
	GVector     mPosition;  //the position of the object in space
	GQuaternion mAngle;     //the direction the object is facing
	Entity* mNext;          //linked list of entities
	Entity* mPrev;          //linked list of entities
	bool mDeleteMe;         //mark for garbage collection?
	tTime mCallbackTime;    //how long before Callback() will be called
	tType mType;            //what sort of object this is
	
protected:
	GVector mVelocity;      //the speed at which the object is traveling
	GVector mAccell;        //the accelleration of the object
	GCoordinate mRadius;    //the radius of the object, for collision detection
	
	//accessors
	inline pGSprite Sprite() {return mSprite;}
	inline void DeleteMe() {mDeleteMe = true;}
	inline void CallbackAfterTime(tTime time) {mCallbackTime = time;}
	inline tTime CallbackTimeRemaining() {return mCallbackTime;}
	
	//these functions can be implemented by derived classes to catch certain
	// events.
	virtual void Touch(Entity* other) {}  //called when the entity touches another entity
	virtual void Think(tTime time, tTime frameTime) {} //called every frame
	virtual void Callback() {} //called when the callback time runs out
	
	Entity(tType type)
		: mDeleteMe(false), mType(type)
	{
		mPosition = GVECTOR_0;
		mAngle = GQuaternion(GVECTOR_0, 0.0f);
		mVelocity = GVECTOR_0;
		mAccell = GVECTOR_0;
		mRadius = 0.0f;
		mDeleteMe = false;
		mCallbackTime = -1.0f;
		
		//add to entity list
		mPrev = NULL;
		mNext = gEntityList;
		if(mNext)
			mNext->mPrev = this;
		gEntityList = this;
		
		mSprite = gTheScene->NewSprite();
		
		//tell gMiscScheduler to call Entity::DoFrame() every frame.
		gMiscScheduler->CallEveryFrame(this, 1);
	}
	
public:
	virtual ~Entity()
	{
		//remove from entity list
		if(mNext)
			mNext->mPrev = mPrev;
		if(mPrev)
			mPrev->mNext = mNext;
		if(gEntityList == this)
			gEntityList = mNext;
		
		//other clean-up
		gMiscScheduler->StopCallingEveryFrame(this);
		mSprite->Remove();
	}
	
	//accessors...
	inline tType Type() {return mType;}
	
	inline Entity* Next() {return mNext;}
	
	inline GVector Position() {return mPosition;}
	inline void Position(GVector position)
	{
		mPosition = position;
		mSprite->Position(position);
	}
	
	inline GQuaternion Angle() {return mAngle;}
	inline void Angle(GQuaternion angle)
	{
		mAngle = angle;
		mSprite->Angle(angle);
	}
	
	inline bool DeleteRequested() {return mDeleteMe;}
	
	//this function will be called by gMiscScheduler every frame.
	void DoFrame(tTime time, tTime frameTime)
	{
		bool callbackEnabled = false;
		if(mCallbackTime > 0.0f)
		{
			mCallbackTime -= frameTime;
			callbackEnabled = true;
		}
		
		//apply accelleration and velocity
		mPosition += mVelocity * frameTime + 0.5f * mAccell * frameTime * frameTime;
		mVelocity += mAccell * frameTime;
		
		//wrap position (if you go off one side of the screen, you end up on the other side)
		while(mPosition.x > MAX_X)
			mPosition.x -= 2.0f * MAX_X;
		while(mPosition.x < -MAX_X)
			mPosition.x += 2.0f * MAX_X;
		while(mPosition.y > MAX_Y)
			mPosition.y -= 2.0f * MAX_Y;
		while(mPosition.y < -MAX_Y)
			mPosition.y += 2.0f * MAX_Y;
		
		//move sprite to new position
		mSprite->Position(mPosition);
		
		//find collisions
		Entity* other;
		for(other = mNext; other != NULL; other = other->mNext)
		{
			if((mPosition - other->mPosition).SquareMagnitude()
			 < ((mRadius + other->mRadius) * (mRadius + other->mRadius)))
			{
				Touch(other);
				other->Touch(this);
			}
		}
		
		//call Think()
		Think(time, frameTime);
		
		//if callback time has run out, call Callback()
		//Funny story:  At one point, this said (mCallbackTime < 0.0f) rather than <=.
		// Somehow, this actually was a problem on one of my friends' computers.  His
		// computer was so punctual that the callback time would actually hit exactly
		// 0.0f about once per every 20-50 missiles fired, causing that missile to
		// live forever (until it hit something).  Let this be a lesson to all.
		if((mCallbackTime <= 0.0f) && callbackEnabled)
			Callback();
	}
};

class Asteroid: public Entity
{
private:
	static int mNumAsteroids;  //current number of astroids in scene
	GQuaternion mTumble;       //tumble speed
	pGTexture mTexture;        //which texture this asteroid is using
	
	//spawn the asteroid with the given size, position, and texture
	void Spawn(GCoordinate radius, GVector position, pGTexture texture)
	{
		pGSprite sprite = Sprite();
		pGModel model = gConfig->AsteroidModel();
		mTexture = texture;
		
		sprite->Model(model);
		sprite->AnimState(model->CreateAnimState());
		sprite->Texture(texture);
		sprite->LightProps(0.4f, 16.0f, 0.0f);
		sprite->Color(1.0, 1.0, 1.0);
		sprite->Scale(radius);
		sprite->Show();
		
		Position(position);
		mVelocity = GVector(Random(-MAX_X, MAX_X), Random(-MAX_Y, MAX_Y), 0.0);
		mVelocity.Normalize();
		mVelocity *= gConfig->AsteroidSpeed() / radius;
		mRadius = radius;
		
		GVector randomVector(Random(-1, 1), Random(-1, 1), Random(-1, 1));
		randomVector.Normalize();
		if(randomVector.SquareMagnitude() < 0.9f)
			randomVector = GVECTOR_I;   //prevent (0, 0, 0) as rotation axis
		mTumble = GQuaternion(randomVector, Random(-3.0f, 3.0f));
		
		randomVector = GVector(Random(-1, 1), Random(-1, 1), Random(-1, 1));
		randomVector.Normalize();
		if(randomVector.SquareMagnitude() < 0.9f)
			randomVector = GVECTOR_I;   //prevent (0, 0, 0) as rotation axis
		Angle(GQuaternion(randomVector, Random(-PI, PI)));
	}
	
public:
	//create a new big asteroid at a random location
	Asteroid()
		: Entity(ASTEROID)
	{
		Spawn(gConfig->InitialAsteroidRadius(),
		      GVector(Random(-MAX_X, MAX_X), Random(-MAX_Y, MAX_Y), 0.0),
		      gConfig->AsteroidTexture());
		mNumAsteroids++;
	}
	
	//create an asteroid which resulted from the destruction of another asteroid.
	Asteroid(Asteroid* other)
		: Entity(ASTEROID)
	{
		Spawn(other->mRadius * 0.5f, other->Position(), other->mTexture);
		mNumAsteroids++;
	}
	
	~Asteroid()
	{
		mNumAsteroids--;
	}
	
	void Destroy()
	{
		if(mRadius > 0.3f * gConfig->InitialAsteroidRadius())
		{
			//spawn pieces of the asteroid
			new Asteroid(this);
			new Asteroid(this);
			new Asteroid(this);
		}
		
		DeleteMe();   //mark for deletion next frame
	}
	
	void Think(tTime time, tTime frameTime)
	{
		//apply tumble
		GQuaternion zero(GVECTOR_0, 0.0f), q;
		q.Interpolate(zero, mTumble, frameTime);
		Angle(Angle() * q);
	}
	
	inline static int NumAsteroids() {return mNumAsteroids;}
};

int Asteroid::mNumAsteroids = 0;

class Missile: public Entity
{
public:
	//spawn missile fired from a ship which is at the given position, pointing in the
	// given direction, and moving with the given momentum.
	Missile(GVector position, GQuaternion angle, GVector momentum)
		: Entity(MISSILE)
	{
		pGSprite sprite = Sprite();
		pGModel model = gConfig->MissileModel();
		
		sprite->Model(model);
		sprite->AnimState(model->CreateAnimState());
		sprite->Texture(gConfig->MissileTexture());
		sprite->LightProps(0.0f, 0.0f, 0.0f);
		sprite->Color(1.0, 1.0, 1.0);
		sprite->Scale(0.7);
		sprite->Show();
		
		GVector direction = GVECTOR_FORWARD;
		angle.Rotate(&direction);
		
		Position(position + gConfig->ShipRadius() * direction);
		Angle(angle);
		mVelocity = gConfig->MissileSpeed() * direction + momentum;
		
		CallbackAfterTime(gConfig->MissileTime());  //disappear after time
	}
	
protected:
	void Touch(Entity* other)
	{
		if(other->Type() == ASTEROID)
		{
			((Asteroid*)other)->Destroy();
			DeleteMe();
		}
	}
	
	void Callback()
	{
		DeleteMe();  //missile time ran out
	}
};

class Ship: public Entity
{
	pGSprite mShieldSprite;  //blue transparent shield sphere
	
	bool mAccellerating;     //is the accelleration button down?
	bool mDecellerating;     //is the decelleration button down?
	bool mTurningLeft;       //is the turn left button down?
	bool mTurningRight;      //is the turn right button down?
	bool mShieldsOn;         //is the shields button down?
	bool mFiring;            //is the fire button down?
	
	bool mAlive;             //is the ship alive and on-screen?
	
	tTime mFireTime;         //time when the ship last fired a missile
	
public:
	Ship()
		: Entity(SHIP)
	{
		pGSprite sprite = Sprite();
		pGModel model = gConfig->ShipModel();
		
		//setup ship sprite, but don't show yet
		sprite->Model(model);
		sprite->AnimState(model->CreateAnimState());
		sprite->Texture(gConfig->ShipTexture());
		sprite->LightProps(1.0f, 32.0f, 0.0f);
		sprite->Color(1.0, 1.0, 1.0);
		sprite->Scale(gConfig->ShipScale());
		
		mRadius = gConfig->ShipRadius();
		
		//setup shield sprite, but don't show yet
		model = new GSphereModel();
		mShieldSprite = gTheScene->NewSprite();
		mShieldSprite->Model(model);
		mShieldSprite->AnimState(model->CreateAnimState());
		mShieldSprite->LightProps(1.0f, 32.0f, 0.5f);
		mShieldSprite->Color(0.0, 0.6, 1.0, 0.5);
		mShieldSprite->Scale(mRadius);
		
		mAccellerating = false;
		mDecellerating = false;
		mTurningLeft   = false;
		mTurningRight  = false;
		mFiring        = false;
		mShieldsOn     = false;
		mAlive         = false;
		mFireTime      = 0.0f;
		
		//appear after a short delay
		CallbackAfterTime(gConfig->ShipSpawnTime());
	}
	
	//Controls for the ship.
	void Accellerate(bool down) {mAccellerating = down;}
	void Decellerate(bool down) {mDecellerating = down;}
	void TurnLeft(bool down)    {mTurningLeft   = down;}
	void TurnRight(bool down)   {mTurningRight  = down;}
	void Fire(bool down)        {mFiring        = down;}
	void Shields(bool down)
	{
		mShieldsOn = down;
		if(mAlive)
		{
			if(down)
				mShieldSprite->Show();
			else
				mShieldSprite->Hide();
		}
	}
	
	void Reset()     //called when the ship is destroyed
	{
		Position(GVECTOR_0);
		Angle(GQuaternion(GVECTOR_0, 0.0f));
		mVelocity = GVECTOR_0;
		mAccell = GVECTOR_0;
		Sprite()->Hide();
		mShieldSprite->Hide();
		
		mAlive = false;
		
		CallbackAfterTime(gConfig->ShipSpawnTime());  //respawn after delay
	}
	
protected:
	void Think(tTime time, tTime frameTime)
	{
		if(!mAlive)
			return;
		
		//adjust shield sprite to be at same position as ship
		mShieldSprite->Position(Position());
		
		//recalculate accelleration
		//first, we need to know the direction the ship is facing
		GVector forward = GVECTOR_FORWARD;
		Angle().Rotate(&forward);
		
		//then, we can figure out the accelleration
		if(mAccellerating && !mDecellerating)
			mAccell = forward * gConfig->ShipSpeed();
		else if(!mAccellerating && mDecellerating)
			mAccell = forward * -gConfig->ShipSpeed();
		else
			mAccell = GVECTOR_0;
		
		//calculate turning speed and bank
		GAngles::tAngle turnSpeed;
		GAngles::tAngle targetRoll;
		if(mTurningLeft && !mTurningRight)
		{
			turnSpeed  = gConfig->ShipTurnSpeed();
			targetRoll = -0.8f;
		}
		else if(!mTurningLeft && mTurningRight)
		{
			turnSpeed  = -gConfig->ShipTurnSpeed();
			targetRoll = 0.8f;
		}
		else
		{
			turnSpeed = 0.0f;
			targetRoll = 0.0f;
		}
		
		//use GAngles for pitch and roll -- Euler angles are easier that quaternions...
		GAngles angle(Angle());
		angle.mYaw += turnSpeed * frameTime;
		angle.mRoll = targetRoll + (angle.mRoll - targetRoll) * exp(-4.0 * frameTime);
		Angle(angle);
		
		if(mFiring && !mShieldsOn
		&& (mFireTime + (1.0f / gConfig->ShipFireRate())) < time)
		{
			//fire a missile
			new Missile(Position(), Angle(), mVelocity);
			mFireTime = time;
		}
	}
	
	void Touch(Entity* other)
	{
		if(other->Type() == ASTEROID)
		{
			if(mAlive && !mShieldsOn)
			{
				//ship collided with an asteroid.  Destroy the ship and the asteroid.
				((Asteroid*)other)->Destroy();
				Reset();
			}
			else
			{
				//extend callback time so that the ship doesn't
				// crash the instant it spawns.
				//Note that the callback time is updated before
				// touch is called, but Callback() is not called
				// until afterwards.  So, we can always catch the
				// callback before it happens from here.
				if(CallbackTimeRemaining() <= 0.1f)
					CallbackAfterTime(0.1f);
			}
		}
	}
	
	void Callback()
	{
		//spawn ship
		mAlive = true;
		Sprite()->Show();
		if(mShieldsOn)
			mShieldSprite->Show();
		else
			mShieldSprite->Hide();
	}
};

class AsteroidsApp: public GMessageQueue::Handler
{
	pGDisplay  mDisplay;   //the display window
	pGCamera   mCamera;    //the view camera
	pGLight    mLight;     //the light
	pGSprite   mBackground[4];  //the background
	
	pGMessageQueue mMessageQueue; //internal message queue
	pGInputMap     mInputMap;     //input map -- makes configurable controls possible
	
	Ship* mShip; //the ship entity
	
	int mLevel;  //level number
	
public:
	//called by message queue when messages arrive
	virtual bool HandleMessage(pGMessage message)
	{
		pGQuitMessage quitmessage = QueryInterface<GQuitMessage>(message);
		if(quitmessage != NULL)
		{
			//quit message received (player clicked the window close button)
			GEngine::Quit();
			return true;
		}
		
		pGInputMessage input = QueryInterface<GInputMessage>(message);
		if(input != NULL)
		{
			//input message received (player pressed a button or moved the mouse or something)
			cout << input->Name().AsciiData() << ' ' << input->Value() << endl;
			
			if(input->Name() == "escape")
			{
				GEngine::Quit();
				return true;
			}
			
			if(input->Name() == "accellerate")
				mShip->Accellerate(input->Value() > 0.5f);
			else if(input->Name() == "decellerate")
				mShip->Decellerate(input->Value() > 0.5f);
			else if(input->Name() == "turn left")
				mShip->TurnLeft(input->Value() > 0.5f);
			else if(input->Name() == "turn right")
				mShip->TurnRight(input->Value() > 0.5f);
			else if(input->Name() == "fire")
				mShip->Fire(input->Value() > 0.5f);
			else if(input->Name() == "shields")
				mShip->Shields(input->Value() > 0.5f);
			else
				return false;
			return true;
		}
		
		return false;
	}
	
	//called every frame
	virtual void DoFrame(tTime time, tTime frameTime)
	{
		Entity *ent, *next;
		
		//delete entities which requested to be deleted
		ent = gEntityList;
		while(ent != NULL)
		{
			next = ent->Next();
			if(ent->DeleteRequested())
				delete ent;
			ent = next;
		}
		
		//if all the asteroids are gone, advance to next level
		if(Asteroid::NumAsteroids() == 0)
		{
			mLevel++;
			StartLevel(mLevel);
		}
		
		//print framerate
		static bool initialized = false;
		static tTime startTime;
		static int framecount;
		
		if(!initialized)
		{
			startTime = time;
			framecount = 0;
			initialized = true;
		}
		
		++framecount;
		if(framecount == 64)
		{
			cout << "framerate: " << framecount / (time - startTime) << endl;
			startTime = time;
			framecount = 0;
		}
	}
	
	void StartLevel(int level)
	{
		Entity *ent, *next;
		
		//delete all entities except for the ship
		ent = gEntityList;
		while(ent != NULL)
		{
			next = ent->Next();
			if(ent->Type() != Entity::SHIP)
				delete ent;
			ent = next;
		}
		
		mShip->Reset();   //reset the ship
		
		//spawn 2n+1 asteroids
		new Asteroid;
		for(int i = 0; i < level; i++)
		{
			new Asteroid;
			new Asteroid;
		}
	}
	
	void Main()
	{
		//load config file
		gConfig = new AsteroidsConfig;
		
		//setup internal message queue and input map
		mMessageQueue = new GMessageQueue;
		mInputMap = new GInputMap(mMessageQueue);
		mInputMap->ControlList(gConfig->ControlList());
		mInputMap->ControlMap(gConfig->ControlMap());
		gMainMessageQueue->AddHandler(mInputMap, 0.1);
		mMessageQueue->AddHandler(this);
		
		//get a rendering plugin with the preferred keyword ("opengl" or "direct3d")
		gTheRenderer = (pGRenderer)gMainObjectLoader->LoadObjectWithKeyword(gConfig->PreferredRenderer(),
		                                                                    "GRenderer");
		if(gTheRenderer == NULL)  //failed...  try to get any rendering plugin at all
			gTheRenderer = (pGRenderer)gMainObjectLoader->LoadObjectOfType("GRenderer");
		if(gTheRenderer == NULL)
			GAUGE3D_FATAL_ERROR("Could not find a rendering plugin.");
		
		//get a display window from the renderer
		mDisplay = gTheRenderer->NewDisplay();
		if(mDisplay == NULL)
			GAUGE3D_FATAL_ERROR("Could not get a display.");
		
		//set the window size
		if(!mDisplay->Open(gConfig->ResX(), gConfig->ResY(), 0, gConfig->Fullscreen()))
			GAUGE3D_FATAL_ERROR("Could not open display.");
		
		//load textures now that the renderer is up
		gConfig->LoadTextures();
		
		//create a scene
		gTheScene = gTheRenderer->NewScene();
		
		//set up the camera
		mCamera = gTheScene->NewCamera();
		mCamera->Position(GVector(0, 0, MAX_X * 4.0f));
		mCamera->Zoom(4.0);                       //almost orthogonal
		mCamera->Direction(GAngles(-PI_2, 0, 0)); //directly down
		mDisplay->Camera(mCamera);    //set the display to use the camera
		
		//set up the main light
		mLight = gTheScene->NewLight();
		mLight->Direction(GVector(-1.0f, 1.0f, 0.7f));
		mLight->Color(1.0f, 1.0f, 1.0f);
		mLight->LightProps(0.2f, 0.6f, 0.2f);
		mLight->Show();
		
		//set up the background
		GCoordinate brightness = gConfig->BackgroundBrightness();
		if(brightness > 0.0f)
		{
			pGBillboardSprite model[4];
			for(int i = 0; i < 4; i++)
			{
				model[i] = new GBillboardSprite(MAX_X * 2, MAX_Y * 2);
				mBackground[i] = gTheScene->NewSprite();
				mBackground[i]->Model(model[i]);
				mBackground[i]->AnimState(model[i]->CreateAnimState());
				mBackground[i]->Texture(gConfig->Background(i));
				mBackground[i]->FullBright(true);
				mBackground[i]->Color(brightness, brightness, brightness);
				mBackground[i]->Show();
			}
			mBackground[0]->Position(GVector(-MAX_X,  MAX_Y, -MAX_X * 4.0));
			mBackground[1]->Position(GVector( MAX_X,  MAX_Y, -MAX_X * 4.0));
			mBackground[2]->Position(GVector(-MAX_X, -MAX_Y, -MAX_X * 4.0));
			mBackground[3]->Position(GVector( MAX_X, -MAX_Y, -MAX_X * 4.0));
			
			//prevent border seams between tiles...
			model[0]->TexCoords(GVector(0.0f, 0.0f, 0.0f),
			                    GVector(1.0f - 0.5f / 256, 1.0f - 0.5f / 256, 0.0f));
			model[1]->TexCoords(GVector(0.5f / 256, 0.0f, 0.0f),
			                    GVector(1.0f, 1.0f - 0.5f / 256, 0.0f));
			model[2]->TexCoords(GVector(0.0f, 0.5f / 256, 0.0f),
			                    GVector(1.0f - 0.5f / 256, 1.0f, 0.0f));
			model[3]->TexCoords(GVector(0.5f / 256, 0.5f / 256, 0.0f),
			                    GVector(1.0f, 1.0f, 0.0f));
		}
		
		//register ourself as a message handler.  Note that our input map is
		// also registered as a handler, but it's priority is lower, so it will
		// be called first.  If it doesn't know how to handle the message, then
		// we will be called directly.
		gMainMessageQueue->AddHandler(this, 0.9);
		gMiscScheduler->CallEveryFrame(this);    //call DoFrame() every frame.
		
		mShip = new Ship;
		
		mLevel = gConfig->StartLevel();
		StartLevel(mLevel);
		
		GEngine::MainLoop();
		
		//delete all entities
		Entity *ent, *next;
		ent = gEntityList;
		while(ent != NULL)
		{
			next = ent->Next();
			ent = next;
		}
		
		//clean up
		gMiscScheduler->StopCallingEveryFrame(this);
		gMainMessageQueue->RemoveHandler(this);
		mDisplay = NULL;
		gTheScene = NULL;
		gTheRenderer = NULL;
	}
};

} //namespace GAUGE3DAsteroids

int main(int argc, char* argv[])
{
	srand(time(NULL));
	
	//init engine
	GAUGE3D::GEngine::CommandLine(argc, argv);
	GAUGE3D::GEngine::ProgName("gauge3d-games");
	GAUGE3D::GEngine::Init(GAUGE3D::GEngine::ALL_SUBSYSTEMS);
	
	//create app
	GAUGE3D::GSmartPointer<GAUGE3DAsteroids::AsteroidsApp> app = new GAUGE3DAsteroids::AsteroidsApp;
	
	try
	{
		app->Main();
	}
	catch(const GAUGE3D::GException& error)
	{
		//display dialog box showing the error that occurred
		GAUGE3D::GString message;
		message = error.Cause();
		message += GAUGE3D::GString("\nFile: ") + GAUGE3D::GString(error.File());
		message += GAUGE3D::GString("\nLine: ") + GAUGE3D::GString(error.Line());
		GAUGE3D::GMessageBox errorBox(message, GAUGE3D::GMessageBox::ERROR, GAUGE3D::GMessageBox::OK);
		errorBox.Run();
	}
	
	//we want to make sure that the app is deleted before the engine
	// shuts down.  Since it is a smart poitner, we set it to NULL to
	// delete it.
	app = NULL;
	
	GAUGE3D::GEngine::Shutdown();
	
	return 0;
}

