/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "angles.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include "vector.h"
#endif

#ifndef __GAUGE3D_3DMATH_QUATERNION_H__INCLUDED__
#include "quaternion.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {


// class member functions ============================================
// lifecycle ---------------------------------------------------------
GAngles::GAngles(const GQuaternion& quat)
{
	GVector forward(GVECTOR_FORWARD),
	        up(GVECTOR_UP);
	
	quat.Rotate(&forward, &forward, 1);
	quat.Rotate(&up, &up, 1);
	
	SetDirection(forward, up);
}

// operations --------------------------------------------------------
void
GAngles::Rotate(const GVector in[], GVector out[], int num) const
{
	GAngles::tAngle pitchSin = (GCoordinate)sin(mPitch);
	GAngles::tAngle pitchCos = (GCoordinate)cos(mPitch);
	GAngles::tAngle yawSin   = (GCoordinate)sin(mYaw);
	GAngles::tAngle yawCos   = (GCoordinate)cos(mYaw);
	GAngles::tAngle rollSin  = (GCoordinate)sin(mRoll);
	GAngles::tAngle rollCos  = (GCoordinate)cos(mRoll);
	
	int i;
	GCoordinate x1, y1, z1, x2, y2, z2;
	
	for(i = 0; i < num; i++)
	{
		x1 = in[i].x;
		y1 = in[i].y;
		z1 = in[i].z;
		
		//roll around y axis
		x2 = x1 * rollCos + z1 * rollSin;
		z2 = z1 * rollCos - x1 * rollSin;
		
		//pitch around x axis
		y2 = y1 * pitchCos - z2 * pitchSin;
		z1 = z2 * pitchCos + y1 * pitchSin;
		
		//yaw around z axis
		x1 = x2 * yawCos - y2 * yawSin;
		y1 = y2 * yawCos + x2 * yawSin;
		
		out[i].x = x1;
		out[i].y = y1;
		out[i].z = z1;
	}
}

//MakeVectors takes an angle and converts it to a vector pointing forward,
// a vector pointing right, and a vector pointing up, if forward is in the
// direction of the angles
//You are free to pass NULL for any of the pointer parameters if you don't
// need that value.
void
GAngles::MakeVectors(GVector* forward, GVector* up, GVector* right) const
{
	if(forward != NULL)
		Rotate(&GVECTOR_FORWARD, forward, 1);
	if(right != NULL)
		Rotate(&GVECTOR_RIGHT, right, 1);
	if(up != NULL)
		Rotate(&GVECTOR_UP, up, 1);
}

//SetDirection takes a vector and makes a set of angles which would
// make that vector point forwards.
void
GAngles::SetDirection(const GVector& vec)
{
	if(IsApproxZero(vec.x) && IsApproxZero(vec.y))
		mYaw = 0.0f;
	else
		mYaw = (GAngles::tAngle)atan2(-vec.x, vec.y);
	mPitch = (GAngles::tAngle)asin(vec.z);
	mRoll = 0.0f;
}

//This version will set pitch, yaw, and roll so that the forward, up, and right
// vectors are valid.
void
GAngles::SetDirection(const GVector& forward, const GVector& up)
{
	if(IsApproxZero(forward.x) && IsApproxZero(forward.y))
	{
		//special case: pointing strait up or down.  Both roll and yaw control
		// the same axis of rotation in this case (the Z axis).  This is also
		// known as gimbal lock.
		//We like yaw better than roll, so we use it and set roll to 0.
		mRoll = 0.0f;
		if(forward.z > 0)
		{
			//pointing strait up
			mPitch = PI_2;
			mYaw = atan2(up.x, -up.y);
		}
		else
		{
			//pointing strait down
			mPitch = -PI_2;
			mYaw = atan2(-up.x, up.y);
		}
	}
	else
	{
		mYaw = (GAngles::tAngle)atan2(-forward.x, forward.y);
		mPitch = (GAngles::tAngle)asin(forward.z);
		
		//Find z component of forward X up.  We only need Z.
		GCoordinate rightZ = forward.x * up.y - forward.y * up.x;
		
		mRoll = atan2(-rightZ, up.z);
	}
}

// operators ---------------------------------------------------------
bool
GAngles::operator==(const GAngles& other) const
{
	return ((IsApproxZero(mPitch - other.mPitch))
	     && (IsApproxZero(mYaw   - other.mYaw))
	     && (IsApproxZero(mRoll  - other.mRoll)));
}

//multiplication actually combines the rotations correctly, but takes longer.
const GAngles&
GAngles::operator*=(const GAngles& other)
{
	GVector vecs[2] = {GVECTOR_FORWARD, GVECTOR_UP};
	
	other.Rotate(vecs, vecs, 2);
	Rotate(vecs, vecs, 2);
	
	SetDirection(vecs[0], vecs[1]);
	
	return *this;
}

GAngles::operator GQuaternion() const
{
	GQuaternion qPitch(GVECTOR_I, mPitch),
	            qYaw  (GVECTOR_K, mYaw),
	            qRoll (GVECTOR_J, mRoll),
	            ret;
	
	ret = qYaw;
	ret *= qPitch;
	ret *= qRoll;
	
	return ret;
}

} //namespace GAUGE3D
