/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_3DMATH_ANGLES_H__INCLUDED__
#define __GAUGE3D_3DMATH_ANGLES_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include "vector.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

class GQuaternion;

// class definition ==================================================
/*! \class GAngles angles.h gauge3d/3dmath/angles.h
 *  \ingroup d3math
 *  \brief Represents a set of Euler angles (pitch, yaw, and roll).
 *  \sa GQuaternion
 *
 *  Euler angles are more intuitive than quaternions, especially in certain
 *  types of games (such as 3D shooters).  However, Euler angles have many
 *  limitations, and are not very elegant mathematically.  Fortunately,
 *  you can convert Euler angles into quaternions and vice versa.
 *
 *  A pitch, yaw, and roll of zero points in the direction of GVECTOR_FORWARD.
 *  Pitch rotates clockwise around the X axis, yaw rotates clockwise around the
 *  Z axis, and roll rotates clockwise around the Y axis.  When rotating a vector,
 *  roll is applied first, then pitch, then yaw. */
class GAUGE3D_API GAngles
{
public:
	typedef float tAngle;

	/*! \name Angle Values
	 * \brief All angles are in radians. */
	//@{
	tAngle mPitch;
	tAngle mYaw;
	tAngle mRoll;
	//@}
	
	// lifecycle -------------------------------------------------------
	GAngles();                       //!< Leaves pitch, yaw, and roll unititialized.
	GAngles(tAngle pitch, tAngle yaw, tAngle roll); //!< Sets pitch, yaw, and roll manually.
	GAngles(const GQuaternion& quat);      //!< Converts a quaternion into Euler angles.
	
	// operations ------------------------------------------------------
	//! Convert degrees to radians.
	static tAngle Degrees2Radians(tAngle degrees);
	//! Convert radians to degrees.
	static tAngle Radians2Degrees(tAngle radians);
	
	/*! \brief Rotate a set of vectors by the angles.
	 *  \param in  An array of vectors to rotate.
	 *  \param out Location to which to write the rotated vectors.  May be
	 *             the same as in.
	 *  \param num Number of vectors to rotate.
	 *  \remarks Rotating many vectors at once is much faster than rotating vectors individually.
	 */
	void Rotate(const GVector in[], GVector out[], int num) const;
	
	/*! \brief Finds forward, up, and right vectors for the angles.
	 *  \param forward The location to write the forward vector, or NULL if it should
	 *                 not be calculated.
	 *  \param up      The location to write the up vector, or NULL if it should
	 *                 not be calculated.
	 *  \param right   The location to write the right vector, or NULL if it should
	 *                 not be calculated.
	 *
	 * MakeVectors takes an angle and converts it to a vector pointing forward,
	 * a vector pointing right, and a vector pointing up, according to the direction
	 * the angles are pointing.  In other words, it is similar to calling Rotate()
	 * with the input being the vectors \ref GVECTOR_FORWARD, \ref GVECTOR_UP, and
	 * \ref GVECTOR_RIGHT. */
	void MakeVectors(GVector* forward, GVector* up = NULL, GVector* right = NULL) const;
	
	/*! \brief Sets the angles to point in the given direction.
	 *
	 * SetDirection takes a vector and sets the pitch and yaw such that that
	 * vector points forwards.  Roll is set to zero. */
	void SetDirection(const GVector& forward);
	
	/*! \brief Sets the angles so that the given forward and up vectors are valid.
	 *
	 * This version sets pitch, yaw, <i>and roll</i> so that both the given vectors
	 * are equal to what you'd get if you called MakeVectors. */
	void SetDirection(const GVector& forward, const GVector& up);
	
	// operators -------------------------------------------------------
	//const GAngles& operator=(const GAngles& other);  //use default
	
	//! Component-wise addition (fast).
	const GAngles& operator+=(const GAngles& other);
	//! Component-wise addition (fast).
	GAngles operator+(const GAngles& other);
	//! Component-wise subtraction (fast).
	const GAngles& operator-=(const GAngles& other);
	//! Component-wise subtraction (fast).
	GAngles operator-(const GAngles& other);
	//! Component-wise multiplication (fast).
	GAngles operator*(GCoordinate a);
	//! Component-wise multiplication (fast).
	const GAngles& operator*=(GCoordinate a);
	
	/*! \brief Combines the angles "correctly" (slow).
	 *
	 * Multiplication between GAngles objects is done by converting them to
	 * GQuaternion's, multiplying those, and then converting the result back
	 * to a GAngles.  This results in a more correct combination than simply
	 * adding the components does, but it takes longer. */
	GAngles operator*(const GAngles& other);
	/*! \brief Combines the angles "correctly" (slow).
	 *
	 * Multiplication between GAngles objects is done by converting them to
	 * GQuaternion's, multiplying those, and then converting the result back
	 * to a GAngles.  This results in a more correct combination than simply
	 * adding the components does, but it takes longer. */
	const GAngles& operator*=(const GAngles& other);
	
	//! Checks for equality with error tolerance.
	bool operator==(const GAngles& other) const;
	
	//! Converts to a GQuaternion.
	operator GQuaternion() const;
};

inline
GAngles::GAngles()
{
}

inline
GAngles::GAngles(tAngle pitch, tAngle yaw, tAngle roll)
	: mPitch(pitch),
	  mYaw(yaw),
	  mRoll(roll)
{
}

inline GAngles::tAngle
GAngles::Degrees2Radians(tAngle degrees)
{
	return degrees * (tAngle)(PI / 180.0);
}

inline GAngles::tAngle
GAngles::Radians2Degrees(tAngle radians)
{
	return radians * (tAngle)(180.0 / PI);
}

inline const GAngles&
GAngles::operator+=(const GAngles& other)
{
	mPitch += other.mPitch;
	mYaw   += other.mYaw;
	mRoll  += other.mRoll;
	return *this;
}
inline GAngles
GAngles::operator+(const GAngles& other)
{
	GAngles ret(*this);
	ret += other;
	return ret;
}

inline const GAngles&
GAngles::operator-=(const GAngles& other)
{
	mPitch -= other.mPitch;
	mYaw   -= other.mYaw;
	mRoll  -= other.mRoll;
	return *this;
}
inline GAngles
GAngles::operator-(const GAngles& other)
{
	GAngles ret(*this);
	ret -= other;
	return ret;
}

inline const GAngles&
GAngles::operator*=(GCoordinate a)
{
	mPitch *= a;
	mYaw   *= a;
	mRoll  *= a;
	return *this;
}
inline GAngles
GAngles::operator*(GCoordinate a)
{
	GAngles ret(*this);
	ret *= a;
	return ret;
}
/*! \brief Component-wise subtraction (fast).
 *  \ingroup d3math
 *  \relates GAngles */
inline GAngles
operator*(GCoordinate coord, const GAngles& angles) //scalar multiplication
{
	GAngles ret(angles);
	return ret *= coord;
}

//multiplication actually combines the rotations correctly, but takes longer.
inline GAngles
GAngles::operator*(const GAngles& other)
{
	GAngles a(*this);
	a *= other;
	return a;
}

} // namespace GAUGE3D

#endif //included
