/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "quaternion.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_MATH_H__INCLUDED__
#include <math.h>
#endif

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class member functions ============================================
// lifecycle ---------------------------------------------------------
//axis must be a unit vector, or the quaternion won't work.
GQuaternion::GQuaternion(const GVector& axis, GCoordinate angle)
{
	angle /= 2.0f;
	GCoordinate s = (GCoordinate)sin(angle);
	x = s * axis.x;
	y = s * axis.y;
	z = s * axis.z;
	w = (GCoordinate)cos(angle);
}

// operations --------------------------------------------------------
//Rotate() applies the quaternion rotation to a set of vectors.
void
GQuaternion::Rotate(const GVector in[], GVector out[], int num) const
{
	GQuaternion q;
	
	//TODO:  For large numbers of vectors, build a temporary 3x3 matrix and use
	//       it to do the rotations.
	
	for(int i = 0; i < num; i++, in++, out++)
	{
		// q = this * vec
		q.x = y * in->z - z * in->y + x + w * in->x;
		q.y = z * in->x - x * in->z + y + w * in->y;
		q.z = x * in->y - y * in->x + z + w * in->z;
		q.w = w - x * in->x - in->y * y - z * in->z;
		
		//vec = q * conjugate(this)
		out->x = q.z * y - q.y * z + w * q.x - q.w * x;
		out->y = q.x * z - q.z * x + w * q.y - q.w * y;
		out->z = q.y * x - q.x * y + w * q.z - q.w * z;
//		out->w = q.w * w + q.x * x + y * q.y + q.z * z;
	}
}

//Interpolate() interpolates along a sphere between q1 and q2 using
// time value t, which is between 0.0 (this), and 1.0 (other).  The
// results are stored in this.  This is useful if you want to find
// the orientation between two quaternions, or want to animate an object
// moving between these two orientations.
void
GQuaternion::Interpolate(const GQuaternion& q1, const GQuaternion& q2, GCoordinate t)
{
	GCoordinate angle, dotprod;
	GCoordinate s1, s2;
	
	dotprod = q1.x * q2.x + q1.y * q2.y + q1.z * q2.z + q1.w * q2.w;
	if(IsApproxZero(dotprod - 1.0f) || IsApproxZero(dotprod + 1.0f))
	{
		*this = q1;
		return;      //quaternions are equal or opposite
	}
	angle = (GCoordinate)acos(dotprod);
	
	s1 = (GCoordinate)(1.0f / sin(angle));
	s2 = (GCoordinate)(sin(angle * t) * s1);
	s1 = (GCoordinate)(sin(angle * (1.0f - t)) * s1);
	
	x = s1 * q1.x + s2 * q2.x;
	y = s1 * q1.y + s2 * q2.y;
	z = s1 * q1.z + s2 * q2.z;
	w = s1 * q1.w + s2 * q2.w;
}

// access ------------------------------------------------------------
//note that these functions have to extract the return values from the
// quaternion, so they are not extremely fast.
GCoordinate
GQuaternion::Angle() const
{
	return (GCoordinate)acos(w) * 2.0f;
}

GVector
GQuaternion::Axis() const
{
	GVector ret(x, y, z);
	if(ret == GVECTOR_0)
		return GVECTOR_I;  //angle is 0 or 180 -- just return something that won't crash the caller.
	ret.Normalize();
	return ret;
}

//axis must be a unit vector, or the quaternion won't work.
void
GQuaternion::Set(const GVector& axis, GCoordinate angle)
{
	angle /= 2.0f;
	GCoordinate s = (GCoordinate)sin(angle);
	x = s * axis.x;
	y = s * axis.y;
	z = s * axis.z;
	w = (GCoordinate)cos(angle);
}

// operators ---------------------------------------------------------
GQuaternion
GQuaternion::operator*(const GQuaternion& other) const
{
	GQuaternion ret;
	
	ret.x = y * other.z - z * other.y + other.w * x + w * other.x;
	ret.y = z * other.x - x * other.z + other.w * y + w * other.y;
	ret.z = x * other.y - y * other.x + other.w * z + w * other.z;
	ret.w = w * other.w - x * other.x - other.y * y - z * other.z;
	
	return ret;
}

const GQuaternion&
GQuaternion::operator*=(const GQuaternion& other)
{
	GQuaternion ret;
	
	ret.x = y * other.z - z * other.y + other.w * x + w * other.x;
	ret.y = z * other.x - x * other.z + other.w * y + w * other.y;
	ret.z = x * other.y - y * other.x + other.w * z + w * other.z;
	ret.w = w * other.w - x * other.x - other.y * y - z * other.z;
	
	return (*this = ret);
}

} //namespace GAUGE3D
