/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "triangle.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_LINE_H__INCLUDED__
#include "line.h"
#endif //redundant include guard

#ifndef __GAUGE3D_3DMATH_PLANE_H__INCLUDED__
#include "plane.h"
#endif //redundant include guard

// begin namespace ===================================================
namespace GAUGE3D {

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GTriangle::GTriangle(const GVector& vertex1, const GVector& vertex2, const GVector& vertex3)
{
	mVertices[0] = vertex1;
	mVertices[1] = vertex2;
	mVertices[2] = vertex3;
}

GTriangle::GTriangle(GVector vertices[])
{
	mVertices[0] = vertices[0];
	mVertices[1] = vertices[1];
	mVertices[2] = vertices[2];
}

// operations --------------------------------------------------------
void
GTriangle::GetPlane(GPlane* plane) const
{
	GVector vec, normal;
	
	//Find normal vector.
	//first, find vector from vertex 0 to vertex 2
	normal = mVertices[2];
	normal -= mVertices[0];
	//then, find vector from vertex 0 to vertex 1
	vec = mVertices[1];
	vec -= mVertices[0];
	//and cross them
	normal = normal.X(vec);
	
	plane->n = normal;
	plane->d = -(mVertices[0] * normal);
}

void
GTriangle::Reverse()
{
	GVector swap;
	
	swap = mVertices[0];
	mVertices[0] = mVertices[2];
	mVertices[2] = swap;
}

//Contains returns true if the point is on the triangle.
bool
GTriangle::Contains(const GVector& point)
{
	GPlane plane;
	GetPlane(&plane);
	
	if(!plane.Contains(point))
		return false;    //point is not on plane.
	
	GLine l(point, plane.n);
	
	if(Intersects(l))
		return true;
	return false;
}

//When checking intersection with a line, 0 is returned if there is no
// intersection, 1 if the line hits the front of the triangle, or -1 if the
// line hits the back of the triangle.  (the front side is the side from which
// the vertices of the polygon appear clockwise)
//Note:  This function is faster than the one above.
//Note2: This function is very, *very* fast. >:)
int8
GTriangle::Intersects(const GLine& line)
{
	GVector u, v;  //u and v of edge line in Plucker coordinates.
	GCoordinate dir;
	int8 result;
	
	// L = {P-Q:PxQ}, for P and Q distinct points on L.
	u = mVertices[0] - mVertices[1];
	v = mVertices[0].X(mVertices[1]);
	
	dir = u * line.v + line.u * v;
	if(dir > 0.0f)
		result = -1;   //first edge is counter-clockwise
	else
		result = 1;    //first edge is clockwise
	
	//Now do the second edge.
	u = mVertices[1] - mVertices[2];
	v = mVertices[1].X(mVertices[2]);
	
	dir = u * line.v + line.u * v;
	if(dir > 0.0f)
		if(result > 0)
			return false;  //edge has different direction from first edge
	else
		if(result < 0)
			return false;  //edge has different direction from first edge
	
	//Now do the third edge.
	u = mVertices[2] - mVertices[0];
	v = mVertices[2].X(mVertices[0]);
	
	dir = u * line.v + line.u * v;
	if(dir > 0.0f)
		if(result > 0)
			return false;  //edge has different direction from first edge
	else
		if(result < 0)
			return false;  //edge has different direction from first edge
	
	return result;
}

} // namespace GAUGE3D
