/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "engine.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#include "gauge3d/osabstraction/filesystem.h"
#include "gauge3d/osabstraction/timer.h"
#include "gauge3d/plugins/objectloader.h"
#include "gauge3d/input/messagequeue.h"
#include "gauge3d/files/diskdirectory.h"

// local package -----------------------------------------------------
#include "scheduler.h"

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GEngine::Private
{
	// state variables -------------------------------------------------
	static GArray<GString> mCommandLine;
	static GString mProgname;
	
	static bool mInitialized;
	static int  mSubSystems;
	
	static bool mQuitRequested;
	
	// private functions -----------------------------------------------
	static void InitBase()
	{
		GTimer::Init(0.0f);
		gMiscScheduler = new GScheduler;
	}
	
	static void ShutdownBase()
	{
		gMiscScheduler = NULL;
	}
	
	
	static void InitInput()
	{
		gInputScheduler   = new GScheduler;
		gMainMessageQueue = new GMessageQueue;
	}
	
	static void ShutdownInput()
	{
		gMainMessageQueue = NULL;
		gInputScheduler   = NULL;
	}
	
	
	static void InitVideo()
	{
		gVideoScheduler = new GScheduler;
	}
	
	static void ShutdownVideo()
	{
		gVideoScheduler = NULL;
	}
	
	
	static void InitFilesystem()
	{
		GFileSystem::Init(mProgname);
		
		//FIXME:  Error checking
		gWorkingDir  = new GDiskDirectory(GFileSystem::WorkingDir());
		gConfigDir   = new GDiskDirectory(GFileSystem::ConfigDir());
		gResourceDir = new GDiskDirectory(GFileSystem::ResourceDir());
	}
	
	static void ShutdownFilesystem()
	{
		GFileSystem::ShutDown();
		
		gWorkingDir  = NULL;
		gConfigDir   = NULL;
		gResourceDir = NULL;
	}
	
	
	static void InitObjectLoader()
	{
		gMainObjectLoader = new GObjectLoader;
		
		pGDiskDirectory pluginDir = new GDiskDirectory(GFileSystem::PluginDir());
		GArray<GString> plugins = pluginDir->ListFiles();
		
		GString extension;
#ifdef PLATFORM_WINDOWS
		extension = ".dll";
#elif defined(PLATFORM_POSIX)
		extension = ".so";
#else
		#error "Don't know file extension for plugins on your platform"
#endif
		
		int pos;
		pGFile file;
		
		for(int i = 0; i < plugins.Size(); i++)
		{
			pos = plugins[i].FindLastOf('.');
			if(pos >= 0)
			{
				if(plugins[i].SubString(pos) == extension)
					gMainObjectLoader->AddPlugin(pluginDir->OpenFile(plugins[i]));
			}
		}
	}
	
	static void ShutdownObjectLoader()
	{
		gMainObjectLoader = NULL;
	}
};

GArray<GString> GEngine::Private::mCommandLine;
GString         GEngine::Private::mProgname("gauge3d");

bool            GEngine::Private::mInitialized = false;
int             GEngine::Private::mSubSystems  = 0;

bool            GEngine::Private::mQuitRequested = false;

// class member functions ============================================
void
GEngine::CommandLine(int argc, char* argv[])
{
	Private::mCommandLine = GArray<GString>(argc);
	
	for(int i = 0; i < argc; i++)
		Private::mCommandLine[i] = argv[i];
}

void
GEngine::CommandLine(char* args)
{
	//TODO: Parse args
}

void
GEngine::CommandLine(int argc, GString argv[])
{
	Private::mCommandLine = GArray<GString>(argc);
	
	for(int i = 0; i < argc; i++)
		Private::mCommandLine[i] = argv[i];
}

void
GEngine::CommandLine(GArray<GString> args)
{
	Private::mCommandLine = args;
}

GArray<GString>
GEngine::CommandLine()
{
	return Private::mCommandLine;
}

void
GEngine::ProgName(GString name)
{
	Private::mProgname = name;
}

GString
GEngine::ProgName()
{
	return Private::mProgname;
}

void
GEngine::Init(int subsystems)
{
	if(!Private::mInitialized)
	{
		Private::InitBase();
		Private::mInitialized = true;
	}
	
	//don't initialize stuff that is already initialized.
	subsystems &= ~Private::mSubSystems;
	
	if(subsystems & INPUT)
		Private::InitInput();
	if(subsystems & VIDEO)
		Private::InitVideo();
	if(subsystems & FILESYSTEM)
		Private::InitFilesystem();
	if(subsystems & OBJECT_LOADER)
		Private::InitObjectLoader();
	
	Private::mSubSystems |= subsystems;
}

int
GEngine::InitializedSubSystems()
{
	return Private::mSubSystems;
}

void
GEngine::ShutdownSubSystem(int subsystems)
{
	//don't shut down anything that is not initialized.
	subsystems &= Private::mSubSystems;
	
	if(subsystems & INPUT)
		Private::ShutdownInput();
	if(subsystems & VIDEO)
		Private::ShutdownVideo();
	if(subsystems & FILESYSTEM)
		Private::ShutdownFilesystem();
	if(subsystems & OBJECT_LOADER)
		Private::ShutdownObjectLoader();
	
	Private::mSubSystems &= ~subsystems;
}

void
GEngine::Shutdown()
{
	ShutdownSubSystem(ALL_SUBSYSTEMS);
	
	if(Private::mInitialized)
		Private::ShutdownBase();
	
	Private::mInitialized = false;
}

void
GEngine::MainLoop()
{
	tTime time, lasttime, frametime;
	Private::mQuitRequested = false;
	
	if(!Private::mInitialized)
		return;  //ERROR: not initialized
	
	lasttime = GTimer::Time();
	
	while(!Private::mQuitRequested)
	{
		time = GTimer::Time();
		frametime = time - lasttime;
		
		gMiscScheduler->DoFrame(time, frametime);
		
		if(Private::mSubSystems & VIDEO)
			gVideoScheduler->DoFrame(time, frametime);
		if(Private::mSubSystems & INPUT)
			gInputScheduler->DoFrame(time, frametime);
		
		lasttime = time;
	}
}

void
GEngine::Quit()
{
	Private::mQuitRequested = true;
}

} //namespace GAUGE3D
