/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "configfile.h"

// system ------------------------------------------------------------
#include <vector>
using namespace std;

// external packages -------------------------------------------------
// local package -----------------------------------------------------
#include "stringfile.h"

// begin namespace ===================================================
namespace GAUGE3D {

namespace {

void
RemovePadding(GString* str)
{
	int begin, end;
	wchar_t curchar;
	
	for(begin = 0; begin < str->Length(); begin++)
	{
		curchar = (*str)[begin];
		if(curchar != ' ' && curchar != '\t')
			break;
	}
	for(end = str->Length() - 1; end > 0; end--)
	{
		curchar = (*str)[end];
		if(curchar != ' ' && curchar != '\t')
			break;
	}
	
	*str = str->SubString(begin, end - begin + 1);
}

void
ReadSection(pGVirtualDirectory section, GString lines[], int numLines)
{
	GString entryName, data;
	pGStringFile key;
	pGVirtualDirectory subSection;
	
	for(int i = 0; i < numLines; i++)
	{
		if((lines[i][0] == '[')
		&& (lines[i][lines[i].Length() - 1] == ']'))
		{
			if(lines[i][1] != '/')
			{
				//begin sub-section
				entryName = lines[i].SubString(1, lines[i].Length() - 2);
				RemovePadding(&entryName);
				subSection = new GVirtualDirectory;
				GString endSection = "[/" + entryName + "]";
				
				i++;
				int j;
				
				//find the length of the section
				for(j = i; j < numLines; j++)
				{
					if(lines[j] == endSection)
						break;
				}
				
				//parse it
				ReadSection(subSection, lines + i, j - i);
				
				//add it
				section->AddSubdir(entryName, subSection);
				
				i = j;
			}
		}
		else
		{
			int equalsPos = lines[i].FindFirstOf('=');
			if(equalsPos >= 0)
			{
				entryName = lines[i].SubString(0, equalsPos);
				RemovePadding(&entryName);
				data = lines[i].SubString(equalsPos + 1);
				RemovePadding(&data);
				key = new GStringFile(data);
				section->AddFile(entryName, key);
			}
		}
	}
}

void
WriteSection(pGOStream output, pGVirtualDirectory section, int level)
{
	GArray<GString> entrylist;
	pGVirtualDirectory subdir;
	pGFile file;
	GString data;
	GString indent;
	int i;
	
	for(i = 0; i < level; i++)
		indent += '\t';
	
	//read files first and write them as key/data pairs.
	entrylist = section->ListFiles();
	for(i = 0; i < entrylist.Size(); i++)
	{
		file = section->OpenFile(entrylist[i]);
		(*output) << indent << entrylist[i] << " = " << (GString)(*file) << '\n';
	}
	file = NULL;
	
	//now read the sub-sections
	entrylist = section->ListSubdirs();
	for(i = 0; i < entrylist.Size(); i++)
	{
		(*output) << indent << '[' << entrylist[i] << ']' << '\n';
		subdir = (pGVirtualDirectory)section->OpenSubdir(entrylist[i]);
		WriteSection(output, subdir, level + 1);
		(*output) << indent << "[/" << entrylist[i] << ']' << '\n';
	}
}

} //anonymous namespace

pGVirtualDirectory
GConfigFile::Load(pGFile file)
{
	pGRawIStream inSource = file->GetRawInputStream();
	if(inSource == NULL)
		return NULL;  //error
	
	pGVirtualDirectory target = new GVirtualDirectory;
	
	GIStream input(inSource);
	input.ReadMode(GIStream::ASCII);
	
	GString line;
	vector<GString> data;
	
	//read in all lines.
	while(input.Read(&line, 1, "^\n"))
	{
		RemovePadding(&line);
		if(line != "")
		{
			data.push_back(line);
		}
	}
	
	ReadSection(target, data.begin(), data.size());
	
	return target;
}

bool
GConfigFile::Save(pGFile file, pGVirtualDirectory config)
{
	pGOStream output = file->GetOutputStream();
	output->WriteMode(GOStream::ASCII);
	
	if(output == NULL)
		return false;  //error
	
	WriteSection(output, config, 0);
	
	return true;
}

} // namespace GAUGE3D
