/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_FILES_DIRECTORY_H__INCLUDED__
#define __GAUGE3D_FILES_DIRECTORY_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_ARRAY_H__INCLUDED__
#include "gauge3d/common/array.h"
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_FILES_FILE_H__INCLUDED__
#include "file.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

class GDirectory;

/*! \typedef pGDirectory
 *  \ingroup files
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GDirectory GDirectory\endlink. */
typedef GSmartPointer<GDirectory> pGDirectory;

// class definition ==================================================
/*! \class GDirectory directory.h gauge3d/files/directory.h
 *  \ingroup files
 *  \brief Abstract interface to a directory.
 *
 *  This could represents a directory on the hard drive, but it could
 *  also represent anything from a compressed archive to an ftp site.
 *  Archive plugins should be derived from this. */
class GAUGE3D_API GDirectory: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f41);
public:
	// lifecycle -------------------------------------------------------
	virtual ~GDirectory();
	
	// operations ------------------------------------------------------
	//! Get a list of all the files in the directory.
	virtual GArray<GString> ListFiles() = 0;
	//! Get a list of all the subdirectories of the directory.
	virtual GArray<GString> ListSubdirs() = 0;
	
	/*! \brief Get a pointer to a file in the directory.
	 * \param name The name of the file to open.
	 * \remarks
	 * Use this to open the contents of a directory.  This way, if you have
	 * a pointer to the root directory of any filesystem, you can access
	 * everything in the system.<br>
	 * \note It is perfectly legal to say:<br>
	 * <ul><li>myDirectory.OpenFile("this/is/a/deeply/nested/file");</ul>
	 * Likewise with OpenSubdir().  You need not manually traverse the
	 * directory tree.
	 * <br>Always use forward slashes ('/') to separate directories.
	 *    Using back slashes ('\\') will not work, even on Windows. */
	virtual pGFile OpenFile(GString name) = 0;
	/*! \brief Get a pointer to a sub directory in the directory.
	 * \param name The name of the sub directory to open.
	 * \remarks
	 * Use this to open the contents of a directory.  This way, if you have
	 * a pointer to the root directory of any filesystem, you can access
	 * everything in the system.<br>
	 * \note It is perfectly legal to say:<br>
	 * <ul><li>myDirectory.OpenFile("this/is/a/deeply/nested/subdir");</ul>
	 * Likewise with OpenSubdir().  You need not manually traverse the
	 * directory tree.
	 * <br>Always use forward slashes ('/') to separate directories.
	 *    Using back slashes ('\\') will not work, even on Windows. */
	virtual pGDirectory OpenSubdir(GString name) = 0;
	
	/*! \brief Create a new file in the directory.
	 *  \param name The name of the file to create.
	 *  \returns A pointer to the new file, or NULL if the file already
	 *           existed or could not be created. */
	virtual pGFile AddFile(GString name) = 0;
	/*! \brief Create a new sub directory to the directory.
	 *  \param name The name of the sub directory to create.
	 *  \returns A pointer to the new directory, or NULL if the directory already
	 *           existed or could not be created. */
	virtual pGDirectory AddSubdir(GString name) = 0;
	
	//! Remove a file from the dircetory.
	virtual void RemoveFile(GString name) = 0;
	//! Remove a sub directory from the directory (recursively if necessary).
	virtual void RemoveDirectory(GString name) = 0;
};

/*! \brief The current working directory.
 *  \ingroup files
 *
 *  This is the directory from which the program was run. */
extern GAUGE3D_API pGDirectory gWorkingDir;

/*! \brief The directory in which config files should be placed.
 *  \ingroup files
 *
 *  All config files for your program should be placed here. */
extern GAUGE3D_API pGDirectory gConfigDir;

/*! \brief The directory which contains program resources.
 *  \ingroup files
 *
 *  The data files for your program reside in this directory. */
extern GAUGE3D_API pGDirectory gResourceDir;

} //namespace GAUGE3D

#endif //included
