/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "diskdirectory.h"

// system ------------------------------------------------------------
#include <vector>
using namespace std;

// external packages -------------------------------------------------
#ifndef __GAUGE3D_OSABSTRACTION_FILESYSTEM_H__INCLUDED__
#include "gauge3d/osabstraction/filesystem.h"
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_FILES_DISKFILE_H__INCLUDED__
#include "diskfile.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// GDiskDirectory state definition ===================================
struct GDiskDirectory::Private
{
	// state variables -------------------------------------------------
	bool mOpen;
	GString mDirname;
	
	// private functions -----------------------------------------------
};

// GDiskDirectory member functions ===================================
// lifecycle ---------------------------------------------------------
GDiskDirectory::GDiskDirectory()
	: mPrivate(*(new Private))
{
	mPrivate.mOpen = false;
}

GDiskDirectory::GDiskDirectory(GString dirname)
	: mPrivate(*(new Private))
{
	Open(dirname);
}

GDiskDirectory::~GDiskDirectory()
{
	if(mPrivate.mOpen)
		Close();
}

bool
GDiskDirectory::Open(GString dirname)
{
	if(mPrivate.mOpen)
		Close();
	
	mPrivate.mDirname = dirname;
	
	if(!GFileSystem::Exists(dirname))
		return mPrivate.mOpen = false;
	if(!GFileSystem::IsDirectory(dirname))
		return mPrivate.mOpen = false;
	
	mPrivate.mOpen = true;
	return true;
}

void
GDiskDirectory::Close()
{
	mPrivate.mOpen = false;
}

// operations --------------------------------------------------------
GArray<GString>
GDiskDirectory::ListFiles()
{
	GFileSystem::Directory* dir = GFileSystem::OpenDirectory(mPrivate.mDirname);
	if(dir == NULL)
		return GArray<GString>();   //ERROR (returns NULL)
	
	vector<GString> list;
	GString entry;
	GString path = mPrivate.mDirname + "/";
	
	while(GFileSystem::ReadDirectory(&entry, dir))
	{
		if(!GFileSystem::IsDirectory(path + entry))
			list.push_back(entry);
	}
	
	GArray<GString> ret(list.size());
	
	for(int i = 0; i < ret.Size(); i++)
		ret[i] = list[i];
	
	return ret;
}

GArray<GString>
GDiskDirectory::ListSubdirs()
{
	GFileSystem::Directory* dir = GFileSystem::OpenDirectory(mPrivate.mDirname);
	if(dir == NULL)
		return GArray<GString>();   //ERROR (returns NULL)
	
	vector<GString> list;
	GString entry;
	GString path = mPrivate.mDirname + "/";
	
	while(GFileSystem::ReadDirectory(&entry, dir))
	{
		if(GFileSystem::IsDirectory(path + entry))
		{
			//skip "." and "..", since we usually don't want to worry about them.
			if(entry != "." && entry != "..")
				list.push_back(entry);
		}
	}
	
	GArray<GString> ret(list.size());
	
	for(int i = 0; i < ret.Size(); i++)
		ret[i] = list[i];
	
	return ret;
}

//Use these to open the contents of a directory.  This way, if you have
// a pointer to the root directory of any filesystem, you can access
// everything in the system.
//Note that it is perfectly legal to say:
//   myDirectory.OpenFile("this/is/a/deeply/nested/file");
// Likewise with OpenSubdir().  You need not manually traverse the
// directory tree.
//flags is a bitfield using GDiskFile::tOpenFlags
pGFile
GDiskDirectory::OpenFile(GString name)
{
	GString fullName = mPrivate.mDirname + '/' + name;
	GDiskFile* file = new GDiskFile(fullName, GDiskFile::NOCREATE);
	if(!file->IsOpen())
	{
		delete file;
		file = NULL;
	}
	
	return file;
}

pGDirectory
GDiskDirectory::OpenSubdir(GString name)
{
	GString fullName = mPrivate.mDirname + '/' + name;
	GDiskDirectory* subdir = new GDiskDirectory(fullName);
	if(!subdir->IsOpen())
	{
		delete subdir;
		subdir = NULL;
	}
	
	return subdir;
}

//AddFile returns NULL if the file already exists
pGFile
GDiskDirectory::AddFile(GString name)
{
	GString fullName = mPrivate.mDirname + '/' + name;
	GDiskFile* ret = new GDiskFile(fullName, GDiskFile::CREATE | GDiskFile::NOREPLACE);
	if(!ret->IsOpen())
	{
		delete ret;
		ret = NULL;
	}
	
	return ret;
}

//AddSubdir returns NULL if the directory already exists
pGDirectory
GDiskDirectory::AddSubdir(GString name)
{
	GString fullName = mPrivate.mDirname + '/' + name;
	GFileSystem::CreateDirectory(fullName);
	if(GFileSystem::IsDirectory(fullName))
		return new GDiskDirectory(fullName);
	else
		return NULL;   //failed
}

//these two remove files and sub directories from the directory.
void
GDiskDirectory::RemoveFile(GString name)
{
	GString fullName = mPrivate.mDirname + '/' + name;
	GFileSystem::DeleteFile(fullName);
}

void
GDiskDirectory::RemoveDirectory(GString name)
{
	GString fullName = mPrivate.mDirname + '/' + name;
	GFileSystem::DeleteDirectory(fullName);
}

// inquiry -----------------------------------------------------------
bool
GDiskDirectory::IsOpen()
{
	return mPrivate.mOpen;
}


GAUGE3D_DECLARE_INTERFACES(GDiskDirectory)
	GAUGE3D_DECLARE_BASECLASS(GDirectory)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
