/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "diskfile.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_OSABSTRACTION_FILESYSTEM_H__INCLUDED__
#include "gauge3d/osabstraction/filesystem.h"
#endif

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

//just to make stuff more readable...
typedef GFileSystem GFileSystem;

// GDiskFile state definition ========================================
struct GDiskFile::Private
{
	class DiskIStream;
	class DiskOStream;
	
	class DiskFileMap;
	
	// state variables -------------------------------------------------
	bool mOpen;
	GString mFilename;
	
	// private functions -----------------------------------------------
};

// DiskIStream =======================================================
// definition --------------------------------------------------------

class GDiskFile::Private::DiskIStream: public GRawIStream
{
	GFileSystem::File* mFile;
	
public:
	DiskIStream(GString _filename);
	virtual ~DiskIStream();
	
	virtual void Seek(int pos, int seekType);
	virtual int TellPos();
	
	//Read attempts to fill the buffer and returns the amount actually read.
	virtual int Read(void* buffer, int size);
	
	//InputAvailable returns true if there is input available now.
	virtual bool InputAvailable();
	
	//check for errors...
	inline bool IsOpen() {return mFile != NULL;}
};

// member functions --------------------------------------------------
GDiskFile::Private::DiskIStream::DiskIStream(GString filename)
{
	mFile = GFileSystem::Open(filename, GFileSystem::READ);
}
GDiskFile::Private::DiskIStream::~DiskIStream()
{
	if(mFile != NULL)
		GFileSystem::Close(mFile);
}

void
GDiskFile::Private::DiskIStream::Seek(int pos, int seekType)
{
	GFileSystem::tSeekType ourSeekType = GFileSystem::CURRENT;
	switch(seekType)
	{
	case GIStream::BEGINNING:
		ourSeekType = GFileSystem::BEGINNING;
		break;
	case GIStream::CURRENT:
		ourSeekType = GFileSystem::CURRENT;
		break;
	case GIStream::END:
		ourSeekType = GFileSystem::END;
		break;
	}
	
	GFileSystem::Seek(mFile, pos, ourSeekType);
}

int
GDiskFile::Private::DiskIStream::TellPos()
{
	return GFileSystem::TellPos(mFile);
}

//Read attempts to fill the buffer and returns the amount actually read.
int
GDiskFile::Private::DiskIStream::Read(void* buffer, int size)
{
	return GFileSystem::Read(mFile, buffer, size);
}

//InputAvailable returns true if there is input available now.
bool
GDiskFile::Private::DiskIStream::InputAvailable()
{
	if(GFileSystem::TellPos(mFile) !=
	   GFileSystem::FileSize(mFile))
		return true;
	return false;
}

// DiskOStream =======================================================
// definition --------------------------------------------------------

class GDiskFile::Private::DiskOStream: public GRawOStream
{
	GFileSystem::File* mFile;
	
public:
	DiskOStream(GString filename);
	virtual ~DiskOStream();
	
	virtual void Seek(int pos, int seekType);
	virtual int TellPos();
	
	//Write attempts to write the contents of buffer and returns the
	// amount actually written.
	virtual int Write(const void* buffer, int size);
	
	//check for errors...
	inline bool IsOpen() {return mFile != NULL;}
};

// member functions --------------------------------------------------
GDiskFile::Private::DiskOStream::DiskOStream(GString filename)
{
	mFile = GFileSystem::Open(filename, GFileSystem::WRITE);
}
GDiskFile::Private::DiskOStream::~DiskOStream()
{
	if(mFile != NULL)
		GFileSystem::Close(mFile);
}

void
GDiskFile::Private::DiskOStream::Seek(int pos, int seekType)
{
	GFileSystem::tSeekType ourSeekType = GFileSystem::CURRENT;
	switch(seekType)
	{
	case GIStream::BEGINNING:
		ourSeekType = GFileSystem::BEGINNING;
		break;
	case GIStream::CURRENT:
		ourSeekType = GFileSystem::CURRENT;
		break;
	case GIStream::END:
		ourSeekType = GFileSystem::END;
		break;
	}
	
	GFileSystem::Seek(mFile, pos, ourSeekType);
}

int
GDiskFile::Private::DiskOStream::TellPos()
{
	return GFileSystem::TellPos(mFile);
}

//Write attempts to write the contents of buffer and returns the
// amount actually written.
int
GDiskFile::Private::DiskOStream::Write(const void* buffer, int size)
{
	return GFileSystem::Write(mFile, buffer, size);
}


// GDiskFile member functions ========================================
// lifecycle ---------------------------------------------------------
GDiskFile::GDiskFile()
	: mPrivate(*(new Private))
{
	mPrivate.mOpen = false;
}

GDiskFile::GDiskFile(GString filename, int flags)
	: mPrivate(*(new Private))
{
	Open(filename, flags);
}

GDiskFile::~GDiskFile()
{
	if(mPrivate.mOpen)
		Close();
	delete &mPrivate;
}

bool
GDiskFile::Open(GString filename, int flags)
{
	if(mPrivate.mOpen)
		Close();
	
	mPrivate.mOpen = false;
	mPrivate.mFilename = filename;
	GFileSystem::File* file;
	
	if(GFileSystem::Exists(filename))
	{
		//the file already exists
		
		if(flags & NOREPLACE)
			return false;     //failed -- file already exists
		else if(flags & TRUNCATE)
		{
			//truncate the file
			file = GFileSystem::Open(filename,
			          GFileSystem::TRUNCATE |
			          GFileSystem::WRITE);
			if(file == NULL)
				return false;     //failed -- other reason
			GFileSystem::Close(file);
		}
	}
	else
	{
		//the file does not exist
		
		if(flags & CREATE)
		{
			//create the file
			file = GFileSystem::Open(filename,
			          GFileSystem::CREATE |
			          GFileSystem::WRITE);
			if(file == NULL)
				return false;     //failed -- other reason
			GFileSystem::Close(file);
		}
		else
			return false;     //failed -- file not found
	}
	
	//success
	mPrivate.mOpen = true;
	return true;
}

void
GDiskFile::Close()
{
	mPrivate.mOpen = false;
}

// operations --------------------------------------------------------
//Bet you can't guess what this does!
int
GDiskFile::Size()
{
	if(mPrivate.mOpen)
		return GFileSystem::FileSize(mPrivate.mFilename);
	return 0;
}

bool
GDiskFile::Truncate()
{
	if(!mPrivate.mOpen)
		return false;
	
	GFileSystem::File* file;
	file = GFileSystem::Open(mPrivate.mFilename,
	          GFileSystem::TRUNCATE |
	          GFileSystem::WRITE);
	if(file == NULL)
		return false;
	GFileSystem::Close(file);
	return true;
}

//Stream access to file.  May return NULL if not readable/writable.
pGRawIStream
GDiskFile::GetRawInputStream()
{
	Private::DiskIStream* stream;
	
	if(mPrivate.mOpen)
	{
		stream = new Private::DiskIStream(mPrivate.mFilename);
		if(stream->IsOpen())
			return stream;   //success
		delete stream;      //failed to open
	}
	return NULL;           //failed
}

pGRawOStream
GDiskFile::GetRawOutputStream()
{
	Private::DiskOStream* stream;
	
	if(mPrivate.mOpen)
	{
		stream = new Private::DiskOStream(mPrivate.mFilename);
		if(stream->IsOpen())
			return stream;   //success
		delete stream;      //failed to open
	}
	return NULL;           //failed
}

//Memory Mapping class...
class GDiskFile::Private::DiskFileMap: public GFile::FileMap
{
	GFileSystem::FileMap* mFSMap;
	
public:
	DiskFileMap(GFileSystem::FileMap* FSMap);
	virtual ~DiskFileMap();
};

//Use these functions to map a file to memory.  The mapped
// section will begin <offset> bytes from the start of the file and
// continue for <size> bytes.
GFile::pFileMap
GDiskFile::Map(int offset, int size, int flags)
{
	if(!mPrivate.mOpen)
		return NULL;
	
	GFileSystem::FileMap* FSMap = GFileSystem::Map(mPrivate.mFilename, offset, size, flags);
	if(FSMap == NULL)
		return NULL;
	
	return new Private::DiskFileMap(FSMap);
}

GDiskFile::Private::DiskFileMap::DiskFileMap(GFileSystem::FileMap* FSMap)
	: FileMap(FSMap->mMap), mFSMap(FSMap)
{
}
GDiskFile::Private::DiskFileMap::~DiskFileMap()
{
	GFileSystem::UnMap(mFSMap);
}

//This function maps the file to the hard drive if it isn't there already.
// It fills in <location> with the file name.  You should not write to
// this file unless you know that no other part of the engine is using it.
// If the file was not on the hard disk in the first place, it will be
// deleted as soon as the GFile object is closed.
GString
GDiskFile::MapToDisk()
{
	if(!mPrivate.mOpen)
		return GString();
	return mPrivate.mFilename;
}

GString
GDiskFile::Name()
{
	if(!mPrivate.mOpen)
		return GString();
	return mPrivate.mFilename;
}

// inquiry -----------------------------------------------------------
bool
GDiskFile::IsOpen()
{
	return mPrivate.mOpen;
}


GAUGE3D_DECLARE_INTERFACES(GDiskFile)
	GAUGE3D_DECLARE_BASECLASS(GFile)
GAUGE3D_END_INTERFACES

} // namespace GAUGE3D
