/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_FILES_DISKFILE_H__INCLUDED__
#define __GAUGE3D_FILES_DISKFILE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_FILES_FILE_H__INCLUDED__
#include "file.h"
#endif //redundand include guard

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GDiskFile diskfile.h gauge3d/files/diskfile.h
 *  \ingroup files
 *  \brief Access to files on disk. */
class GAUGE3D_API GDiskFile: public GFile
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f4e);
public:
	//! Flags specified with Open().
	//(I offset these flags by 8 bits to prevent errors caused by people
	// using the READ and WRITE flags when openning GDiskFiles.  I made
	// this change after discovering that I myself had been incorrectly
	// using my own freeking interface by adding these flags.  Doh.)
	enum tOpenFlags
	{
		NOCREATE  = 0,       //!< Fail if the file does not exist.
		CREATE    = 1 << 8,  //!< Create the file if it does not exist.
		NOREPLACE = 1 << 9,  //!< Fail if the file already exists.
		TRUNCATE  = 1 << 10  //!< Truncate the file to zero length if it exists.
	};
	
	// lifecycle -------------------------------------------------------
	//! Create without openning a file.
	GDiskFile();
	/*! \brief Create and open a file.
	 *  \param filename The path and name of the file to open.
	 *  \param flags One or more of \ref tOpenFlags, bitwise or'd. */
	GDiskFile(GString filename, int flags);
	//! Close the file.
	virtual ~GDiskFile();
	
	/*! \brief Create and open a file.
	 *  \param filename The path and name of the file to open.
	 *  \param flags One or more of \ref tOpenFlags, bitwise or'd.
	 *  \returns     True on success, false on error. */
	bool Open(GString filename, int flags);
	//! Close the file.
	void Close();
	
	// inquiry ---------------------------------------------------------
	//! Returns true if the file is open.
	bool IsOpen();
	
	// operations ------------------------------------------------------
	//Bet you can't guess what this does!
	virtual int Size();
	
	virtual bool Truncate();
	
	//Stream access to file.  May return NULL if not readable/writable.
	virtual pGRawIStream GetRawInputStream();
	virtual pGRawOStream GetRawOutputStream();
	
	//Use these functions to map a file to memory.  The mapped
	// section will begin <offset> bytes from the start of the file and
	// continue for <size> bytes.
	virtual pFileMap Map(int offset, int size, int flags);
	
	//For GDiskFile, MapToDisk returns the filename of the actual file on
	// the disk.
	virtual GString MapToDisk();
	
	//Get the file's internal name.  Not always the same as the name of the
	// directory entry pointing to it. (especially in a virtual filesystem)
	virtual GString Name();
};
/*! \typedef pGDiskFile
 *  \ingroup files
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GDiskFile GDiskFile\endlink. */
typedef GSmartPointer<GDiskFile> pGDiskFile;

} // namespace GAUGE3D

#endif //included
