/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "ostream.h"

// system ------------------------------------------------------------
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

#ifdef PLATFORM_BIG_ENDIAN
const GOStream::tEndianness HOST_ENDIANNESS = GOStream::BIG;
#else
const GOStream::tEndianness HOST_ENDIANNESS = GOStream::LITTLE;
#endif

// class state definition ============================================
struct GOStream::Private
{
	enum
	{
		DEFAULT_BUFFER_SIZE = 1024,   //fixme:  what is a good size?
	};
	
	// state variables -------------------------------------------------
	GOStream* mThis;
	pGRawOStream mDest;
	
	char* mBuffer;
	int mBufferSize;
	int mBufferPos;
	int mBufferStart;
	int mBufferFree;
	
	tEndianness mEndianness;
	bool mBinary;
	char mSeparator;
	bool mUseSeparator;
	tNumberBase mNumberBase;
	
	bool mLastWrite;
	
	// private functions -----------------------------------------------
	int CopyToBuffer(const void* source, int num);
	
	int BinaryWrite(const void* source, int num, int elementSize);
	bool WriteTextUInt(ulong source);
	bool WriteTextInt(long source);
	bool WriteTextFloat(double source);
	
	//more of MSVC not supporting standards...
//	template<class tType>
//	int WriteUInt(const tType* source, int num);
//	template<class tType>
//	int WriteInt(const tType* source, int num);
//	template<class tType>
//	int WriteFloat(const tType* source, int num);
};

int
GOStream::Private::CopyToBuffer(const void* source, int num)
{
	//want to copy num bytes from source to the buffer.
	
	if(num > mBufferFree)
		num = mBufferFree;
	
	if((mBufferPos > mBufferStart)
	&& (mBufferSize - mBufferPos <= num))
	{
		//neet to copy in two chunks:  Fill the end of the buffer, and
		// then continue at the beginning.
		memcpy(mBuffer + mBufferPos, source, mBufferSize - mBufferPos);
		memcpy(mBuffer, source, num - (mBufferSize - mBufferPos));
		mBufferPos = num - (mBufferSize - mBufferPos);
	}
	else
	{
		//Copy in one chunk
		memcpy(mBuffer + mBufferPos, source, num);
		mBufferPos += num;
	}
	
	mBufferFree -= num;
	
	return num;
}

int
GOStream::Private::BinaryWrite(const void* source, int num, int elementSize)
{
	int elementsWritten = 0;
	
	if((mEndianness == HOST_ENDIANNESS)
	|| (elementSize == 1))
	{
		//no need to swap bytes, so we can write faster
		
		int bytesWritten = 0;
		int numBytes = num * elementSize;
		
		if(mBufferSize == 0)
		{
			//no buffer, so just write the data
			bytesWritten = mDest->Write(source, numBytes);
		}
		else if(numBytes > mBufferFree)
		{
			//need to write more than will fit in the buffer
			bytesWritten += CopyToBuffer(source, numBytes);
			
			if(mThis->Flush() != 0)
			{
				//the flush was not able to write the whole buffer.
				//we may not be able to write everything.
				//all we can do is copy what we have into whatever space
				// did get flushed.
				bytesWritten += CopyToBuffer((uint8*)source + bytesWritten, numBytes - bytesWritten);
			}
			else if(numBytes - bytesWritten > mBufferSize)
			{
				//we have more than another buffer full to write, so
				// just write it all at once.
				bytesWritten += mDest->Write((uint8*)source + bytesWritten, numBytes - bytesWritten);
			}
			else
			{
				//the rest will fit in the buffer.
				bytesWritten += CopyToBuffer((uint8*)source + bytesWritten, numBytes - bytesWritten);
			}
		}
		else
		{
			//into the buffer we go!
			bytesWritten = CopyToBuffer(source, numBytes);
		}
		
		elementsWritten += bytesWritten / elementSize;
	}
	else
	{
		//damnit, we need to swap each element!
//		char swapped[elementSize];
		uint8 swapped[20];   //MSVC is gimpy
		const uint8* unswapped = (uint8*)source;
		
		elementsWritten = 0;
		while(elementsWritten < num)
		{
			//swap the bytes
			for(int i = 0; i < elementSize; i++)
				swapped[i] = unswapped[elementSize - i - 1];
			
			//incerement unswapped to point at the next element
			unswapped += elementSize;
			
			if(BinaryWrite(swapped, elementSize, 1) != elementSize)
				break;
			
			++elementsWritten;
		}
	}
	
	return elementsWritten;
}

bool
GOStream::Private::WriteTextUInt(ulong source)
{
	char buffer[20];
	
	switch(mNumberBase)
	{
	case OCT:
		sprintf(buffer, "%lo", source);
		break;
	case DEC:
		sprintf(buffer, "%lu", source);
		break;
	case HEX:
		sprintf(buffer, "%lx", source);
		break;
	}
		
	const char* bufptr = buffer;
	mThis->Write(&bufptr);
	
	return mLastWrite;
}

bool
GOStream::Private::WriteTextInt(long source)
{
	char buffer[20];
	
	switch(mNumberBase)
	{
	case OCT:
		sprintf(buffer, "%lo", source);
		break;
	case DEC:
		sprintf(buffer, "%li", source);
		break;
	case HEX:
		sprintf(buffer, "%lx", source);
		break;
	}
	
	const char* bufptr = buffer;
	mThis->Write(&bufptr);
	
	return mLastWrite;
}

bool
GOStream::Private::WriteTextFloat(double source)
{
	char buffer[20];
	sprintf(buffer, "%f", source);
	
	const char* bufptr = buffer;
	mThis->Write(&bufptr);
	
	return mLastWrite;
}

//Member templates?  Not for MSVC!
//template<class tType>
//int GOStream::Private::WriteUInt(const tType* source, int num)
//{
#define WRITE_UINT_TEMPLATE(tType, source, num)\
	int numWritten;\
	\
	if(mPrivate.mBinary)\
	{\
		numWritten = mPrivate.BinaryWrite(source, num, sizeof(tType));\
	}\
	else\
	{\
		for(numWritten = 0; numWritten < num; numWritten++)\
		{\
			if(!mPrivate.WriteTextUInt(source[numWritten]))\
				break;\
		}\
	}\
	\
	if(numWritten == 0)\
		mPrivate.mLastWrite = false;\
	else\
		mPrivate.mLastWrite = true;\
	\
	return numWritten;
//}

//template<class tType>
//int GOStream::Private::WriteInt(const tType* source, int num)
//{
#define WRITE_INT_TEMPLATE(tType, source, num)\
	int numWritten;\
	\
	if(mPrivate.mBinary)\
	{\
		numWritten = mPrivate.BinaryWrite(source, num, sizeof(tType));\
	}\
	else\
	{\
		for(numWritten = 0; numWritten < num; numWritten++)\
		{\
			if(!mPrivate.WriteTextInt(source[numWritten]))\
				break;\
		}\
	}\
	\
	if(numWritten == 0)\
		mPrivate.mLastWrite = false;\
	else\
		mPrivate.mLastWrite = true;\
	\
	return numWritten;
//}

//template<class tType>
//int GOStream::Private::WriteFloat(const tType* source, int num)
//{
#define WRITE_FLOAT_TEMPLATE(tType, source, num)\
	int numWritten;\
	\
	if(mPrivate.mBinary)\
	{\
		numWritten = mPrivate.BinaryWrite(source, num, sizeof(tType));\
	}\
	else\
	{\
		for(numWritten = 0; numWritten < num; numWritten++)\
		{\
			if(!mPrivate.WriteTextFloat(source[numWritten]))\
				break;\
		}\
	}\
	\
	if(numWritten == 0)\
		mPrivate.mLastWrite = false;\
	else\
		mPrivate.mLastWrite = true;\
	\
	return numWritten;
//}

// class member functions ============================================
// lifecycle -------------------------------------------------------
GOStream::GOStream(pGRawOStream streamDest)
	: mPrivate(*(new Private))
{
	mPrivate.mThis = this;
	mPrivate.mDest = streamDest;
	
	mPrivate.mBufferSize   = Private::DEFAULT_BUFFER_SIZE;
	mPrivate.mBuffer       = new char[mPrivate.mBufferSize];
	mPrivate.mBufferPos    = 0;
	mPrivate.mBufferStart  = 0;
	mPrivate.mBufferFree   = mPrivate.mBufferSize;
	
	mPrivate.mEndianness   = HOST_ENDIANNESS;
	mPrivate.mBinary       = true;

	mPrivate.mSeparator    = ' ';
	mPrivate.mUseSeparator = false;
	mPrivate.mNumberBase   = DEC;
	
	mPrivate.mLastWrite    = true;
}

GOStream::GOStream(pGRawOStream streamDest, int bufferSize)
	: mPrivate(*(new Private))
{
	mPrivate.mThis = this;
	mPrivate.mDest = streamDest;
	
	mPrivate.mBufferSize   = bufferSize;
	if(bufferSize != 0)
		mPrivate.mBuffer    = new char[mPrivate.mBufferSize];
	mPrivate.mBufferPos    = 0;
	mPrivate.mBufferStart  = 0;
	mPrivate.mBufferFree   = mPrivate.mBufferSize;
	
	mPrivate.mEndianness   = HOST_ENDIANNESS;
	mPrivate.mBinary       = true;
	
	mPrivate.mSeparator    = ' ';
	mPrivate.mUseSeparator = false;
	mPrivate.mNumberBase   = DEC;
	
	mPrivate.mLastWrite    = true;
}

GOStream::~GOStream()
{
	Flush();
	
	if(mPrivate.mBufferSize != 0)
		delete [] mPrivate.mBuffer;
	
	delete &mPrivate;
}

// operations ------------------------------------------------------
void
GOStream::Seek(int pos, tSeekType seekType)
{
	Flush();
	mPrivate.mDest->Seek(pos, seekType);
}

int
GOStream::TellPos()
{
	return mPrivate.mDest->TellPos() + mPrivate.mBufferSize - mPrivate.mBufferFree;
}

//Use Write() to write out one or more of any basic data type.
//Returns the number of elements successfully written.
int
GOStream::Write(const uint64*  source, int num)
{
//	return mPrivate.WriteUInt(source, num);
	WRITE_UINT_TEMPLATE(uint64, source, num)
}
int
GOStream::Write(const uint32*  source, int num)
{
//	return mPrivate.WriteUInt(source, num);
	WRITE_UINT_TEMPLATE(uint32, source, num)
}
int
GOStream::Write(const uint16*  source, int num)
{
//	return mPrivate.WriteUInt(source, num);
	WRITE_UINT_TEMPLATE(uint16, source, num)
}

int
GOStream::Write(const int64*   source, int num)
{
//	return mPrivate.WriteInt(source, num);
	WRITE_INT_TEMPLATE(int64, source, num)
}
int
GOStream::Write(const int32*   source, int num)
{
//	return mPrivate.WriteInt(source, num);
	WRITE_INT_TEMPLATE(int32, source, num)
}
int
GOStream::Write(const int16*   source, int num)
{
//	return mPrivate.WriteInt(source, num);
	WRITE_INT_TEMPLATE(int16, source, num)
}

int
GOStream::Write(const float32* source, int num)
{
//	return mPrivate.WriteFloat(source, num);
	WRITE_FLOAT_TEMPLATE(float32, source, num)
}
int
GOStream::Write(const float64* source, int num)
{
//	return mPrivate.WriteFloat(source, num);
	WRITE_FLOAT_TEMPLATE(float64, source, num)
}

//This does NOT write a string.  It writes one or more characters.
int
GOStream::Write(const char* source, int num)
{
	int numWritten = 0;
	
	if(mPrivate.mBinary)
	{
		numWritten = mPrivate.BinaryWrite(source, num, sizeof(char));
	}
	else
	{
		for(numWritten = 0; numWritten < num; numWritten++)
		{
			Put(source[numWritten]);
			if(!mPrivate.mLastWrite)
				break;
			if(mPrivate.mUseSeparator)
				Put(mPrivate.mSeparator);
			if(!mPrivate.mLastWrite)
				break;
		}
	}
	
	if(numWritten == 0)
		mPrivate.mLastWrite = false;
	else
		mPrivate.mLastWrite = true;
	
	return numWritten;
};

//Write(char**, int) writes null-terminated strings.
int
GOStream::Write(const char* source[], int num)
{
	int numWritten;
	int stringSize;
	
	for(numWritten = 0; numWritten < num; numWritten++)
	{
		stringSize = strlen(source[numWritten]);
		if(mPrivate.mBinary)
		{
			if(mPrivate.BinaryWrite(source[numWritten], stringSize, 1) != num)
				break;    //write error
		}
		else
		{
			for(int i = 0; i < stringSize; i++)
			{
				Put(source[numWritten][i]);
				if(!mPrivate.mLastWrite)
					break;     //write error
			}
			
			if(!mPrivate.mLastWrite)
				break;   //write error
			if(mPrivate.mUseSeparator)
				Put(mPrivate.mSeparator);
			if(!mPrivate.mLastWrite)
				break;   //write error
		}
	}
	
	if(numWritten == 0)
		mPrivate.mLastWrite = false;
	else
		mPrivate.mLastWrite = true;
	
	return numWritten;
}

int
GOStream::Write(const GString*  source, int num)
{
	int numWritten;
	int stringSize;
	
	for(numWritten = 0; numWritten < num; numWritten++)
	{
		stringSize = source[numWritten].Length();
		if(mPrivate.mBinary)
		{
			if(mPrivate.BinaryWrite(source[numWritten].AsciiData(), stringSize, 1) != num)
				break;    //write error
		}
		else
		{
			for(int i = 0; i < stringSize; i++)
			{
				Put(source[numWritten][i]);
				if(!mPrivate.mLastWrite)
					break;     //write error
			}
			
			if(!mPrivate.mLastWrite)
				break;   //write error
			if(mPrivate.mUseSeparator)
				Put(mPrivate.mSeparator);
			if(!mPrivate.mLastWrite)
				break;   //write error
		}
	}
	
	if(numWritten == 0)
		mPrivate.mLastWrite = false;
	else
		mPrivate.mLastWrite = true;
	
	return numWritten;
}

//void* writes are always binary.  num is the number of bytes.
int
GOStream::Write(const void* source, int num)
{
	return mPrivate.BinaryWrite(source, num, 1);
}

//Put places a character into the output stream.
void
GOStream::Put(char c)
{
#ifdef PLATFORM_WINDOWS
	if(c == '\n')
		Put('\r');       //*sigh*
	if(!mPrivate.mLastWrite)
		return;
#endif
	
	if(mPrivate.mBufferFree <= 0)
	{
		Flush();
		if(mPrivate.mBufferFree <= 0)
		{
			mPrivate.mLastWrite = false;
			return;
		}
	}
	
	mPrivate.mBuffer[mPrivate.mBufferPos] = c;
	++mPrivate.mBufferPos;
	--mPrivate.mBufferFree;
}

//Flush makes sure that the entire buffer has been output.  duh.
int
GOStream::Flush()
{
	int numWritten;
	
	if(mPrivate.mBufferFree == mPrivate.mBufferSize)
		return 0;  //buffer is already empty
	
	if(mPrivate.mBufferPos > mPrivate.mBufferStart)
	{
		//we can write everything in one step.
		numWritten = mPrivate.mDest->Write(mPrivate.mBuffer + mPrivate.mBufferStart,
		                                   mPrivate.mBufferPos - mPrivate.mBufferStart);
		mPrivate.mBufferStart += numWritten;
		mPrivate.mBufferFree  += numWritten;
	}
	else
	{
		//buffer is split (wraps around).  Have to write it in two parts.
		numWritten = mPrivate.mDest->Write(mPrivate.mBuffer + mPrivate.mBufferStart,
		                                   mPrivate.mBufferSize - mPrivate.mBufferStart);
		mPrivate.mBufferStart += numWritten;
		mPrivate.mBufferFree  += numWritten;
		
		if(mPrivate.mBufferStart == mPrivate.mBufferSize)
		{
			//so far so good.  Write the rest.
			
			mPrivate.mBufferStart = 0;
			numWritten = mPrivate.mDest->Write(mPrivate.mBuffer, mPrivate.mBufferPos);
			mPrivate.mBufferStart = numWritten;
			mPrivate.mBufferFree  += numWritten;
		}
	}
	
	if(mPrivate.mBufferStart == mPrivate.mBufferPos)
		mPrivate.mBufferStart = mPrivate.mBufferPos = 0;
	
	return mPrivate.mBufferSize - mPrivate.mBufferFree;
}

// access ----------------------------------------------------------
void
GOStream::WriteMode(tWriteMode mode)
{
	if(mode == BINARY)
		mPrivate.mBinary = true;
	else
		mPrivate.mBinary = false;
}

GOStream::tWriteMode
GOStream::WriteMode()
{
	if(mPrivate.mBinary)
		return BINARY;
	else
		return ASCII;
}

//In text output, if you specify a separator, it will be placed after
// every item output.  This allows you to output an array of numbers
// without them getting all mushed together.  This only applies to text
// i/o, not binary.  It does not affect Put() calls.
// Set enable to false to make all output tightly packed.
void
GOStream::Separator(char separator, bool enable)
{
	mPrivate.mSeparator = separator;
	mPrivate.mUseSeparator = enable;
}
char
GOStream::Separator()
{
	return mPrivate.mSeparator;
}

//Endianness() tells the ostream whether to output big or little endian
// data when in binary mode.
void
GOStream::Endianness(tEndianness endianness)
{
	mPrivate.mEndianness = endianness;
}
GOStream::tEndianness
GOStream::Endianness()
{
	return mPrivate.mEndianness;
}

//NumberBase() sets the number base to write text numbers in.  This
// only applies to text mode, not binary.
void
GOStream::NumberBase(tNumberBase base)
{
	mPrivate.mNumberBase = base;
}
GOStream::tNumberBase
GOStream::NumberBase()
{
	return mPrivate.mNumberBase;
}

// inquiry ---------------------------------------------------------	
//Using the stream as a truth value returns true if the last write
// operation succeeded.
GOStream::operator void*()
{
	if(mPrivate.mLastWrite)
		return this;
	return NULL;
}

//GRawOStream member functions ====================================

GRawOStream::~GRawOStream() {}

void
GRawOStream::Seek(int pos, int seekType)
{
	return;
}
int
GRawOStream::TellPos()
{
	return 0;
}

GAUGE3D_DECLARE_INTERFACES(GOStream)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

GAUGE3D_DECLARE_INTERFACES(GRawOStream)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

} // namespace GAUGE3D
