/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "stringfile.h"

// system ------------------------------------------------------------
#include <string.h>

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

namespace {

class GStringIStream: public GRawIStream
{
	GString* mString;
	int mPos;
	
public:
	GStringIStream(GString* str)
		: mString(str)
	{
		mPos = 0;
	}
	
	virtual ~GStringIStream() {}
	
	virtual void Seek(int pos, int seekType)
	{
		switch(seekType)
		{
		default:
		case BEGINNING:
			mPos = pos;
			break;
		case CURRENT:
			mPos += pos;
			break;
		case END:
			mPos = mString->Length() - pos;
			break;
		}
		
		if(mPos < 0)
			mPos = 0;
		else if(mPos > mString->Length())
			mPos = mString->Length();
	}
	
	virtual int TellPos()
	{
		if(mPos > mString->Length())
			mPos = mString->Length();
		return mPos;
	}
	
	virtual int Read(void* buffer, int size)
	{
		if(mPos >= mString->Length())
		{
			mPos = mString->Length();
			return 0;
		}
		if(mPos + size > mString->Length())
			size = mString->Length() - mPos;
		
		if(mString->IsUnicode())
			memcpy(buffer, (uint8*)mString->UnicodeData() + mPos, size);
		else
			memcpy(buffer, (uint8*)mString->AsciiData() + mPos, size);
		
		mPos += size;
		return size;
	}
	
	virtual bool InputAvailable()
	{
		return mPos < mString->Length();
	}
};

class GStringOStream: public GRawOStream
{
	GString* mString;
	int mPos;
	
public:
	GStringOStream(GString* str)
		: mString(str)
	{
		mPos = 0;
	}
	virtual ~GStringOStream()
	{}
	
	virtual void Seek(int pos, int seekType)
	{
		switch(seekType)
		{
		default:
		case BEGINNING:
			mPos = pos;
			break;
		case CURRENT:
			mPos += pos;
			break;
		case END:
			mPos = mString->Length() - pos;
			break;
		}
		
		if(mPos < 0)
			mPos = 0;
		else if(mPos > mString->Length())
			mPos = mString->Length();
	}
	
	virtual int TellPos()
	{
		return mPos;
	}
	
	virtual int Write(const void* buffer, int size)
	{
		int length = mString->Length();
		
		if(mPos < length)
		{
			if(mPos + size < length)
			{
				*mString = mString->SubString(0, mPos)
				         + GString((char*)buffer, size)
				         + mString->SubString(mPos + size);
			}
			else
			{
				*mString = mString->SubString(0, mPos)
				         + GString((char*)buffer, size);
			}
		}
		else
			*mString += GString((char*)buffer, size);
		
		return size;
	}
};

class GStringFileMap: public GFile::FileMap
{
	GString mString;
	
public:
	GStringFileMap(GString str, void* buffer)
		: FileMap(buffer),
		  mString(str)
	{}
	virtual ~GStringFileMap() {};
};

} //anonymous namespace

// class member functions ============================================
// lifecycle -------------------------------------------------------
GStringFile::GStringFile()
{
}

GStringFile::GStringFile(GString data)
	: mData(data)
{
}

GStringFile::~GStringFile()
{
}

// operations ------------------------------------------------------
pGRawIStream
GStringFile::GetRawInputStream()
{
	return new GStringIStream(&mData);
}

pGRawOStream
GStringFile::GetRawOutputStream()
{
	return new GStringOStream(&mData);
}

GFile::pFileMap
GStringFile::Map(int offset, int size, int flags)
{
	uint8* buffer;
	
	if(offset < 0 || size < 0)
		return NULL;
	
	if(offset + size > mData.Length())
	{
		//make string large enough.
		char* randomCrap = new char[offset + size - mData.Length()];
		mData += GString(randomCrap, offset + size - mData.Length());
		delete randomCrap;
	}
	
	if(mData.IsUnicode())
		buffer = (uint8*)mData.UnicodeData();
	else
		buffer = (uint8*)mData.AsciiData();
	
	buffer += offset;
	
	return new GStringFileMap(mData, buffer);
}

GString
GStringFile::MapToDisk()
{
	//TODO
	return "";
}

GAUGE3D_DECLARE_INTERFACES(GStringFile)
	GAUGE3D_DECLARE_BASECLASS(GFile)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
