/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "spheremodel.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_MATH_H__INCLUDED__
#include <math.h>
#define __SYSTEM_MATH_H__INCLUDED__
#endif

#ifndef __SYSTEM_VECTOR__INCLUDED__
#include <vector>
#define __SYSTEM_VECTOR__INCLUDED__
#endif

using namespace std;

// external packages -------------------------------------------------
#ifndef __GAUGE3D_3DMATH_ANGLES_H__INCLUDED
#include "gauge3d/3dmath/angles.h"
#endif

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

//WARNING:  This file is a mess!  I am sure there is an easier way to do
// all this, but I just can't think of it.

// number of triangles in max LOD = 20 * 4 ^ GSPHERE_MAX_LOD
#define GSPHERE_MAX_LOD 6

struct GSphereModel::Private
{
	class SphereRenderInfo: public RenderInfo
	{
		float mLOD;
		
	public:
		SphereRenderInfo() {}
		~SphereRenderInfo() {}
		
		int StartFrame(pAnimState animState, const GVector& viewPoint,
		               const GQuaternion& viewAngle, const GFrustum& viewFrustum,
		               float geoLod)
		{
			//adjust geometric LOD for distance.
			//since geoLOD is the area of a triangle, and the dimentions of
			// a triangle should increase proportionally to the distance from
			// the viewpoint, the area should increase proportionally to the
			// distance squared.
			mLOD = geoLod * viewPoint.SquareMagnitude();
			
			return 1;
		}
		
		//these two are defined below
		void GetStaticInfo(RenderSet* target, int setNum);
		void GetDynamicInfo(RenderSet* target, int setNum);
	};
	
	
	// stuff for generating the initial mesh --------------------------
	static bool mInitialized;
	
	static void Init();
	
	struct IndexList
	{
		int mNumIndices;
		uint16* mIndices;
	};
	
	static uint16* mBaseIndices[GSPHERE_MAX_LOD];
	static int  mNumIndicesForLod[GSPHERE_MAX_LOD];
	static vector<GVector> mBaseVertices;
	static GVector* mBaseVertexArray;
	static GVector* mBaseTexCoords;
	static int  mNumVerticesForLod[GSPHERE_MAX_LOD];
	static int  mCurrentLod;
	static IndexList mIndexListForLod[GSPHERE_MAX_LOD];
	static float mGeoLod[GSPHERE_MAX_LOD];
	
	//this class represents an edge which can be split on demand.
	class Edge
	{
		int mEndpoints[2];  //indices of the endpoints in mBaseVertices
		int mMidpoint;
		Edge* mSubEdges[2];  //divided first and second half of the edge
		
	public:
		Edge(int end1, int end2)
		{
			mEndpoints[0] = end1;
			mEndpoints[1] = end2;
			mSubEdges[0] = NULL;
			mSubEdges[1] = NULL;
			mMidpoint = -1;
		}
		
		~Edge()
		{
			if(mMidpoint != -1)
			{
				delete mSubEdges[0];
				delete mSubEdges[1];
			}
		}
		
		//Divide() splits the edge if it hasn't been split already.
		void Divide()
		{
			if(mMidpoint == -1)
			{
				GVector midpoint;
				
				midpoint.Slerp(mBaseVertices[mEndpoints[0]], mBaseVertices[mEndpoints[1]], 0.5);
				mBaseVertices.push_back(midpoint);
				++mNumVerticesForLod[mCurrentLod];
				mMidpoint = mBaseVertices.size() - 1;
				
				mSubEdges[0] = new Edge(mEndpoints[0], mMidpoint);
				mSubEdges[1] = new Edge(mMidpoint, mEndpoints[1]);
			}
		}
		
		void Reverse()
		{
			int swapV = mEndpoints[0];
			mEndpoints[0] = mEndpoints[1];
			mEndpoints[1] = swapV;
			
			if(mMidpoint != -1)
			{
				Edge* swapE = mSubEdges[0];
				mSubEdges[0] = mSubEdges[1];
				mSubEdges[1] = swapE;
				
				mSubEdges[0]->Reverse();
				mSubEdges[1]->Reverse();
			}
		}
		
		inline int Midpoint()
		{	
			return mMidpoint;
		}
		inline Edge* SubEdge(int num)
		{
			return mSubEdges[num];
		}
		inline int Endpoint(int num)
		{
			return mEndpoints[num];
		}
	};
	
	class Triangle
	{
		Edge* mEdges[3];
		Edge* mInternalEdges[3];
		Triangle* mSubTriangles[4];
		bool mInternalAreCW;   //are internal edges clockwise?
		bool mDivided;
		
	public:
		Triangle(Edge* e1, Edge* e2, Edge* e3)
			: mDivided(false)
		{
			mEdges[0] = e1;
			mEdges[1] = e2;
			mEdges[2] = e3;
		}
		~Triangle()
		{
			if(mDivided)
			{
				delete mSubTriangles[0];
				delete mSubTriangles[1];
				delete mSubTriangles[2];
				delete mSubTriangles[3];
				
				delete mInternalEdges[0];
				delete mInternalEdges[1];
				delete mInternalEdges[2];
			}
		}
		
		void Divide()
		{
			mEdges[0]->Divide();
			mEdges[1]->Divide();
			mEdges[2]->Divide();
			
			mInternalEdges[0] = new Edge(mEdges[0]->Midpoint(), mEdges[1]->Midpoint()),
			mInternalEdges[1] = new Edge(mEdges[1]->Midpoint(), mEdges[2]->Midpoint()),
			mInternalEdges[2] = new Edge(mEdges[2]->Midpoint(), mEdges[0]->Midpoint());
			
			mSubTriangles[0] = new Triangle(mInternalEdges[0], mInternalEdges[1], mInternalEdges[2]);
			mInternalEdges[0]->Reverse();
			mInternalEdges[1]->Reverse();
			mInternalEdges[2]->Reverse();
			mSubTriangles[1] = new Triangle(mEdges[0]->SubEdge(1), mEdges[1]->SubEdge(0), mInternalEdges[0]);
			mSubTriangles[2] = new Triangle(mEdges[1]->SubEdge(1), mEdges[2]->SubEdge(0), mInternalEdges[1]);
			mSubTriangles[3] = new Triangle(mEdges[2]->SubEdge(1), mEdges[0]->SubEdge(0), mInternalEdges[2]);
			
			mInternalAreCW = false;
		}
		
		Triangle* SubTriangle(int num)
		{
			if(((num == 0) && !mInternalAreCW)
			|| ((num != 0) &&  mInternalAreCW))
			{
				mInternalEdges[0]->Reverse();
				mInternalEdges[1]->Reverse();
				mInternalEdges[2]->Reverse();
				
				mInternalAreCW = !mInternalAreCW;
			}
			
			return mSubTriangles[num];
		}
		
		Edge* OuterEdge(int num)
		{
			return mEdges[num];
		}
	};
	
	static void Tesselate(Triangle* tri, int level);
};

bool GSphereModel::Private::mInitialized = false;
uint16* GSphereModel::Private::mBaseIndices[GSPHERE_MAX_LOD];
int  GSphereModel::Private::mNumIndicesForLod[GSPHERE_MAX_LOD];
vector<GVector> GSphereModel::Private::mBaseVertices;
GVector* GSphereModel::Private::mBaseVertexArray;
GVector* GSphereModel::Private::mBaseTexCoords;
int  GSphereModel::Private::mNumVerticesForLod[GSPHERE_MAX_LOD];
int  GSphereModel::Private::mCurrentLod;
GSphereModel::Private::IndexList GSphereModel::Private::mIndexListForLod[GSPHERE_MAX_LOD];
float GSphereModel::Private::mGeoLod[GSPHERE_MAX_LOD];

void GSphereModel::Private::Tesselate(Triangle* tri, int level)
{
	if(level > 0)
	{
		//divide further.
		tri->Divide();
		
		Tesselate(tri->SubTriangle(0), level - 1);
		Tesselate(tri->SubTriangle(1), level - 1);
		Tesselate(tri->SubTriangle(2), level - 1);
		Tesselate(tri->SubTriangle(3), level - 1);
	}
	else
	{
		//tesselated enough.  Add these vertices to the list.
		uint16* indices = &mBaseIndices[mCurrentLod][mNumIndicesForLod[mCurrentLod]];
		
		indices[0] = tri->OuterEdge(0)->Endpoint(0);
		indices[1] = tri->OuterEdge(1)->Endpoint(0);
		indices[2] = tri->OuterEdge(2)->Endpoint(0);
		
		mNumIndicesForLod[mCurrentLod] += 3;
	}
}

void GSphereModel::Private::Init()
{
	//we start with an icosohedron (20-sides)
#define X 0.525731112119133606f
#define Z 0.850650808352039932f
	static GCoordinate vdata[12][3] = {
		{-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},
		{0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},
		{Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0}
	};
	
	static int tindices[20][3] = {
		{0, 4, 1}, {0, 9, 4}, {9, 5, 4}, {4, 5, 8}, {4, 8, 1},
		{8, 10, 1}, {8, 3, 10}, {5, 3, 8}, {5, 2, 3}, {2, 7, 3},
		{7, 10, 3}, {7, 6, 10}, {7, 11, 6}, {11, 0, 6}, {0, 1, 6},
		{6, 1, 10}, {9, 0, 11}, {9, 11, 2}, {9, 2, 5}, {7, 2, 11}
	};
#undef X
#undef Z
	
	int i, j;
	j = 60;   //3 * number of triangles
	for(i = 0; i < GSPHERE_MAX_LOD; i++)
	{
		mNumIndicesForLod[i] = 0;    //count as we write them.
		mBaseIndices[i] = new uint16[j];
		mIndexListForLod[i].mIndices = mBaseIndices[i];
		mIndexListForLod[i].mNumIndices = j;
		j *= 4;
	}
	
	for(i = 0; i < 12; i++)
		mBaseVertices.push_back(GVector(vdata[i][0], vdata[i][1], vdata[i][2]));
	
	mNumVerticesForLod[0] = 12;
	
	Edge* edges[12][12];
	for(i = 0; i < 12; i++)
		for(j = 0; j < 12; j++)
			edges[i][j] = NULL;
	
	Edge* tri[3];
	
	Triangle* triangles[20];
	for(i = 0; i < 20; i++)
		triangles[i] = NULL;
	
	for(j = 0; j < GSPHERE_MAX_LOD; j++)
	{
		if(j > 0)
			mNumVerticesForLod[j] = mNumVerticesForLod[j - 1];
		mCurrentLod = j;
		
		for(i = 0; i < 20; i++)
		{
			if(edges[tindices[i][1]][tindices[i][0]] != NULL)
			{
				//an edge has already been created between these two
				// vertices.  Use it.  It will have to be reversed.
				tri[0] = edges[tindices[i][1]][tindices[i][0]];
				edges[tindices[i][1]][tindices[i][0]] = NULL;
				edges[tindices[i][0]][tindices[i][1]] = tri[0];
				tri[0]->Reverse();
			}
			else
			{
				//No edge exists between these two vertices yet.
				// Create one.
				tri[0] = new Edge(tindices[i][0], tindices[i][1]);
				edges[tindices[i][0]][tindices[i][1]] = tri[0];
			}
			
			if(edges[tindices[i][2]][tindices[i][1]] != NULL)
			{
				tri[1] = edges[tindices[i][2]][tindices[i][1]];
				edges[tindices[i][2]][tindices[i][1]] = NULL;
				edges[tindices[i][1]][tindices[i][2]] = tri[1];
				tri[1]->Reverse();
			}
			else
			{
				tri[1] = new Edge(tindices[i][1], tindices[i][2]);
				edges[tindices[i][1]][tindices[i][2]] = tri[1];
			}
			
			if(edges[tindices[i][0]][tindices[i][2]] != NULL)
			{
				tri[2] = edges[tindices[i][0]][tindices[i][2]];
				edges[tindices[i][0]][tindices[i][2]] = NULL;
				edges[tindices[i][2]][tindices[i][0]] = tri[2];
				tri[2]->Reverse();
			}
			else
			{
				tri[2] = new Edge(tindices[i][2], tindices[i][0]);
				edges[tindices[i][2]][tindices[i][0]] = tri[2];
			}
			
			if(triangles[i] == NULL)
				triangles[i] = new Triangle(tri[0], tri[1], tri[2]);
			
			Tesselate(triangles[i], mCurrentLod);
		}
		
		//4*PI = surface area of sphere
		//mNumIndicesForLod[j] / 3 = number of triangles
		//                    (/ 6 to find the point between two lods)
		//mGeoLod[j] = area of each triangle at LOD j.
		mGeoLod[j] = (float)((4.0f * PI) / (mNumIndicesForLod[j] / 6));
	}
	
	//Now we need to copy the vertices into the array.
	GVector* vertices = new GVector[mBaseVertices.size()];
	GVector* texcoords = new GVector[mBaseVertices.size()];
	GAngles angles;
	for(i = 0; i < (int)mBaseVertices.size(); i++)
	{
		vertices[i] = mBaseVertices[i];
		angles.SetDirection(vertices[i]);
		texcoords[i].x = angles.mYaw / PI;
		if(texcoords[i].x < 0)     //FIXME:  Is there a better way?
			texcoords[i].x *= -1;
		texcoords[i].y = angles.mPitch / PI + 0.5f;
	}
	mBaseVertices.clear();
	mBaseVertexArray = vertices;
	mBaseTexCoords   = texcoords;
	
	mInitialized = true;
}

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GSphereModel::GSphereModel()
	: mPrivate(*(new Private))
{
	if(!Private::mInitialized)
		Private::Init();
}

GSphereModel::~GSphereModel()
{
	delete &mPrivate;
}

GSphereModel::SphereModelAnimState::SphereModelAnimState()
{
}

GSphereModel::SphereModelAnimState::~SphereModelAnimState()
{
}

// operations --------------------------------------------------------
//CreateAnimState() generates an AnimState object associated with this model.
GModel::pAnimState
GSphereModel::CreateAnimState()
{
	return new SphereModelAnimState;
}

GModel::pRenderInfo
GSphereModel::CreateRenderInfo()
{
	return new Private::SphereRenderInfo;
}

void
GSphereModel::Private::SphereRenderInfo::GetStaticInfo(RenderSet* target, int setNum)
{
	int ourLod;
	
	for(ourLod = GSPHERE_MAX_LOD - 1; ourLod > 0; ourLod--)
	{
		if(mGeoLod[ourLod] > mLOD)
			break;   //use this lod
	}
	
	target->mDynamicFlags  = RenderSet::STATIC;
	target->mNumVertices   = mNumVerticesForLod[ourLod];
	
	target->mVertices      = mBaseVertexArray;
	target->mNormals       = mBaseVertexArray;
	target->mTexCoords     = mBaseTexCoords;
	
	target->mNumIndices    = mIndexListForLod[ourLod].mNumIndices;
	target->mIndices       = mIndexListForLod[ourLod].mIndices;
	target->mIndexType     = RenderSet::TRIANGLES;
}

void
GSphereModel::Private::SphereRenderInfo::GetDynamicInfo(RenderSet* target, int setNum)
{
	return;  //not dynamic
}


GAUGE3D_DECLARE_INTERFACES(GSphereModel)
	GAUGE3D_DECLARE_BASECLASS(GModel)
GAUGE3D_END_INTERFACES

GAUGE3D_DECLARE_INTERFACES(GSphereModel::SphereModelAnimState)
	GAUGE3D_DECLARE_BASECLASS(AnimState)
GAUGE3D_END_INTERFACES

} // namespace GAUGE3D
