/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "supermodel.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#include "gauge3d/scheduler.h"

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GSuperModel::Private
{
	class SuperRenderInfo: public RenderInfo
	{
		Model* mModel;
		float mGeoLod, mLastLod;
		int mStartFrame, mEndFrame;
		GCoordinate mT;
		uint16* mIndexList;        //dynamic index list
		                           // (we want to keep the dynamic list from frame to
		                           // frame because it won't change much each frame, but
		                           // it would be a bitch to recompute it from scratch
		                           // every frame)
		int* mMeshIndices;         //indicates where each mesh is located in mIndexList
		int* mMeshLOD;             //the number of vertices in the mesh at the current LOD
		
	public:
		SuperRenderInfo(GSuperModel* model)
			: mModel(model->ModelData()),
			  mGeoLod(0.0f)
		{
			//prepare for some dynamic detail lovin'
			
			int numIndices = 0;   //count total indices in the model
			int i;
			
			//set up mIndexList and mMeshIndices
			mMeshIndices = new int[mModel->mNumMeshes];
			for(i = 0; i < mModel->mNumMeshes; i++)
			{
				if(mModel->mMeshes[i].mEdgeReductions != NULL)
				{
					//this mesh has dynamic LOD
					mMeshIndices[i] = numIndices;
					numIndices += mModel->mMeshes[i].mNumIndices;
				}
				else
				{
					//this mesh does not have dynamic LOD
					mMeshIndices[i] = -1;
				}
			}
			
			if(numIndices != 0)
			{
				//allocate all those indices
				mIndexList = new uint16[numIndices];
				//and space for the current mesh LODs
				mMeshLOD = new int[mModel->mNumMeshes];
				
				//we will initialize everything to have no vertices -- the lowest LOD
				for(i = 0; i < mModel->mNumMeshes; i++)
					mMeshLOD[i] = 0;
				for(i = 0; i < numIndices; i++)
					mIndexList[i] = 0;
			}
			else
			{
				//no dynamic LOD in the model
				mIndexList = NULL;
				delete [] mMeshIndices;
				mMeshIndices = NULL;
				mMeshLOD = NULL;
			}
		}
		~SuperRenderInfo()
		{
			if(mIndexList != NULL)
				delete [] mIndexList;
			if(mMeshIndices != NULL)
				delete [] mMeshIndices;
		}
		
		int StartFrame(pAnimState animState, const GVector& viewPoint,
		               const GQuaternion& viewAngle, const GFrustum& viewFrustum,
		               float geoLod)
		{
			//adjust geometric LOD for distance.
			//since geoLOD is the area of a triangle, and the dimentions of
			// a triangle should increase proportionally to the distance from
			// the viewpoint, the area should increase proportionally to the
			// distance squared.
			mLastLod = mGeoLod;
			mGeoLod = geoLod * viewPoint.SquareMagnitude();
			
			pSuperAnimState ourAnimState = animState;
			ourAnimState->GetMeshFrame(&mStartFrame, &mEndFrame, &mT);
			
			if(mT <= 0.0)
				mEndFrame = mStartFrame;
			else if(mT >= 1.0)
				mStartFrame = mEndFrame;
			
			return mModel->mNumMeshes;
		}
		
		//these two are defined below
		void GetStaticInfo(RenderSet* target, int setNum);
		void GetDynamicInfo(RenderSet* target, int setNum);
	};
	
	// state variables -------------------------------------------------
	Model* mModel;
	
	// private functions -----------------------------------------------
};

// class member functions ============================================
GSuperModel::GSuperModel()
	: mPrivate(*(new Private))
{
}

GSuperModel::~GSuperModel()
{
	delete &mPrivate;
}

void
GSuperModel::ModelData(Model* model)
{
	mPrivate.mModel = model;
}

GSuperModel::Model*
GSuperModel::ModelData()
{
	return mPrivate.mModel;
}

// SuperAnimState -------------------------------------------------
struct GSuperModel::SuperAnimState::Private
{
	int mFrame1;
	int mFrame2;
	tTime mStart;
	tTime mRate;
};

GSuperModel::SuperAnimState::SuperAnimState()
	: mPrivate(*(new Private))
{
	mPrivate.mFrame1 = mPrivate.mFrame2 = 0;
	mPrivate.mRate = 0.0f;
}

GSuperModel::SuperAnimState::~SuperAnimState()
{
	delete &mPrivate;
}

// control info ------------------------------------------------
//SetFrame sets the model's current frame to <thisFrame>, and then has
// the model interpolate to <nextFrame> over the next <time> seconds.
void
GSuperModel::SuperAnimState::SetMeshFrame(int thisFrame, int nextFrame,
                                          tTime startTime, tTime endTime)
{
	mPrivate.mFrame1 = thisFrame;
	mPrivate.mFrame2 = nextFrame;
	mPrivate.mStart = startTime;
	mPrivate.mRate = 1.0f / (endTime - startTime);
}

//GetFrame gets the start and end frame of the current interpolation,
// and how far between them the model currently is.  Thas it, t will
// be filled in with a number between 0.0 and 1.0, with 0.0 being
// thisFrame and 1.0 being nextFrame.
void
GSuperModel::SuperAnimState::GetMeshFrame(int* thisFrame, int* nextFrame, GCoordinate* t)
{
	*thisFrame = mPrivate.mFrame1;
	*nextFrame = mPrivate.mFrame2;
	*t = (GCoordinate)((gVideoScheduler->Time() - mPrivate.mStart) * mPrivate.mRate);
	
	if(*t > 1.0f)
		*t = 1.0f;
	else if(*t < 0.0f)
		*t = 0.0f;
}

GModel::pAnimState
GSuperModel::CreateAnimState()
{
	return new SuperAnimState;
}

// rendering info ----------------------------------------------------
GModel::pRenderInfo
GSuperModel::CreateRenderInfo()
{
	return new Private::SuperRenderInfo(this);
}

void
GSuperModel::Private::SuperRenderInfo::GetStaticInfo(RenderSet* target, int setNum)
{
	int i;
	Mesh* mesh = &mModel->mMeshes[setNum];
	int32 flags = mesh->mDynamicFlags;
	
	//if we are between two frames, interpolate!
	bool interpolate = mStartFrame != mEndFrame;
	
	if(interpolate)
		target->mDynamicFlags = flags;
	else
		target->mDynamicFlags = RenderSet::STATIC;
	
	if(mesh->mEdgeReductions != NULL)
	{
		//dynamic LOD -- need to check for LOD changes.
		
		int newLOD = mMeshLOD[setNum];
		
		int collapseInto;
		int n;
		int* changedIndices;
		uint16* indices = mIndexList + mMeshIndices[setNum];
		
		//find the new LOD
		if(newLOD > 0)   //avoid array out-of-bounds
		{
			while(mGeoLod >= mesh->mEdgeReductions[newLOD - 1].mLODThreshold)
			{
				--newLOD;
				
				//Apply the reduction mesh->mEdgeReductions[newLOD]
				
				//we just need to change the indices indicated in the edge reduction
				// so that instead of pointing to the vertex being collapsed, they
				// point at the vertex it is collapsing into.
				
				collapseInto = mesh->mEdgeReductions[newLOD].mCollapseInto;
				n = mesh->mEdgeReductions[newLOD].mNumChangedIndices;
				changedIndices = mesh->mEdgeReductions[newLOD].mChangedIndices;
				
				for(i = 0; i < n; i++)
					indices[changedIndices[i]] = collapseInto;
				
				if(newLOD == 0)
					break;  //avoid array out-of-bounds
			}
		}
		
		if(newLOD < mesh->mNumVertices)   //avoid array out-of-bounds
		{
			while(mGeoLod < mesh->mEdgeReductions[newLOD].mLODThreshold)
			{
				//Undo the reduction mesh->mEdgeReductions[newLOD]
				
				//we just undo what we did when applying the reduction
				
				n = mesh->mEdgeReductions[newLOD].mNumChangedIndices;
				changedIndices = mesh->mEdgeReductions[newLOD].mChangedIndices;
				
				for(i = 0; i < n; i++)
					indices[changedIndices[i]] = newLOD;
				
				++newLOD;
				
				if(newLOD == mesh->mNumVertices)
					break;  //avoid array out-of-bounds
			}
		}
		
		mMeshLOD[setNum] = newLOD;
		
		target->mNumVertices = newLOD;
		target->mNumIndices = mesh->mEdgeReductions[newLOD].mNumIndices;
		target->mIndices = indices;
	}
	else
	{
		target->mNumVertices = mesh->mNumVertices;
		target->mNumIndices = mesh->mNumIndices;
		target->mIndices    = mesh->mIndices;
	}
	
	if(!(flags & RenderSet::VERTEX_POSITION))
	{
		target->mVertices = mesh->mFrames[0].mVertices;
		target->mNormals  = mesh->mFrames[0].mNormals;
	}
	else if(!interpolate)
	{
		target->mVertices = mesh->mFrames[mStartFrame].mVertices;
		target->mNormals  = mesh->mFrames[mStartFrame].mNormals;
	}
	
	if(!(flags & RenderSet::VERTEX_TEXCOORD))
		target->mTexCoords = mesh->mFrames[0].mTexCoords;
	else if(!interpolate)
		target->mTexCoords = mesh->mFrames[mStartFrame].mTexCoords;
	
	target->mIndexType  = mesh->mIndexType;
}

void
GSuperModel::Private::SuperRenderInfo::GetDynamicInfo(RenderSet* target, int setNum)
{
	if(target->mDynamicFlags == RenderSet::STATIC)
		return;     //nothing to animate
	
	Mesh* mesh = &mModel->mMeshes[setNum];
	MeshFrame* startFrame = &mesh->mFrames[mStartFrame];
	MeshFrame* endFrame   = &mesh->mFrames[mEndFrame];
	int32 flags = target->mDynamicFlags;
	GCoordinate t = mT;
	GCoordinate oneMinusT = 1.0f - t;
	int i;
	
	if(flags & RenderSet::VERTEX_POSITION)
	{
		//do NOT combine these loops!  We want to write all the vertices at once,
		// and then all of the normals in another pass because the buffers that
		// we are writing to may be in AGP RAM or VRAM.  (and if not, we don't want
		// to thrash the cache)
		for(i = 0; i < target->mNumVertices; i++)
		{
			target->mVertices[i] = startFrame->mVertices[i] * oneMinusT
			                       + endFrame->mVertices[i] * t;
		}
		
		for(i = 0; i < target->mNumVertices; i++)
		{
			//we could use SLERP here, but I don't think the visual quality increase
			// is noticeable, and it would be slower.  So, we'll just use LERP.
			target->mNormals[i]  = startFrame->mVertices[i] * oneMinusT
			                       + endFrame->mVertices[i] * t;
		}
	}
	
	if(flags & RenderSet::VERTEX_TEXCOORD)
	{
		for(i = 0; i < target->mNumVertices; i++)
		{
			target->mTexCoords[i] = startFrame->mTexCoords[i] * oneMinusT
			                        + endFrame->mTexCoords[i] * t;
		}
	}
}

GAUGE3D_DECLARE_INTERFACES(GSuperModel)
	GAUGE3D_DECLARE_BASECLASS(GModel)
GAUGE3D_END_INTERFACES

GAUGE3D_DECLARE_INTERFACES(GSuperModel::SuperAnimState)
	GAUGE3D_DECLARE_BASECLASS(AnimState)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
