/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_GRAPHICS_SUPERMODEL_H__INCLUDED__
#define __GAUGE3D_GRAPHICS_SUPERMODEL_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_OSABSTRACTION_TIMER_H__INCLUDED__
#include "gauge3d/osabstraction/timer.h"
#endif

#ifndef __GAUGE3D_COMMON_ARRAY_H__INCLUDED__
#include "gauge3d/common/array.h"
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_GRAPHICS_MODEL_H__INCLUDED__
#include "model.h"
#endif //redundant include guard


// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GSuperModel supermodel.h gauge3d/graphcs/supermodel.h
 *  \ingroup graphcs
 *  \brief Base class for most normal model formats.
 *
 *  This class is slated to be removed and replaced by a better system in
 *  the near future.  Thus, I don't really feel like documenting this
 *  beast for nothing.
 */
class GAUGE3D_API GSuperModel: public GModel
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39e929ef);
public:
	struct MeshFrame
	{
		//! Array of vertex positions.
		GVector* mVertices;
		//! Array of normal vectors.  Must be unit vectors!
		GVector* mNormals;
		//! Array of texture coordinates.  Can be NULL if texturing is not supported.
		GVector* mTexCoords;
		
		inline MeshFrame(): mVertices(NULL), mNormals(NULL), mTexCoords(NULL) {}
	};
	
	struct EdgeReduction
	{
		float32 mLODThreshold;
		int32 mCollapseInto;
		int32 mNumIndices;
		
		int32  mNumChangedIndices;
		int32* mChangedIndices;
	};
	
	struct Mesh
	{
		int32 mDynamicFlags;
		
		//! Number of elements in each of the arrays.
		int32 mNumVertices;
		int32 mNumFrames;
		GStridedArray<MeshFrame> mFrames;
		
		EdgeReduction* mEdgeReductions;
		
		//! Specifies how to make geometry out of the indices.
		int32   mIndexType;
		//! The number of elements in the index array.
		int32   mNumIndices;
		//! Pointer to an array of indices.
		uint16* mIndices;
		//Each index can be thought of as a half-edge which has endpoints of that vertex
		// and the previous vertex in the triangle.  mAdjacencies stores one number per
		// index, which refers the edge of the adjacent triangle.  (FIXME: that description
		// sucked!)  mAdjacencies is not necessary (it can be NULL), but some rendering features
		// might not work without it.
		int32*  mAdjacencies;
		
		inline Mesh(): mDynamicFlags(0), mNumVertices(0), mNumFrames(0),
		               mEdgeReductions(NULL), mIndexType(RenderInfo::RenderSet::TRIANGLES),
		               mNumIndices(0), mIndices(NULL), mAdjacencies(NULL) {}
	};
	
	struct Model
	{
		int32 mVersion;
		int32 mNumMeshes;
		GStridedArray<Mesh> mMeshes;
		//TODO:  More stuff!
		
		inline Model(): mVersion(0), mNumMeshes(0) {}
	};
	
	GSuperModel();
	virtual ~GSuperModel();
	
protected:
	void ModelData(Model* model);
	
public:
	Model* ModelData();
	
	class GAUGE3D_API SuperAnimState: public GModel::AnimState
	{
		struct Private;
		Private& mPrivate;
		
		GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39e929e8);
	public:
		SuperAnimState();
		virtual ~SuperAnimState();

		/*! \brief Sets the mesh animation parameters.
		 * \param thisFrame The starting frame for the animation.
		 * \param nextFrame The ending frame for the animation.
		 * \param startTime The starting time for the animation.
		 * \param endTime   The ending time for the animation.
		 *
		 * Between startTime and endTime, the model's mesh position will be interpolated
		 * from the starting frame to the ending frame. */
		void SetMeshFrame(int  thisFrame, int  nextFrame, tTime startTime, tTime endTime);
		/*! \brief Get the mesh animation parameters for the current time.
		 *
		 * thisFrame and nextFrame will be filled in to whatever they were last
		 * set to using SetFrame().  t will be filled in with a value between 0
		 * and 1 which represents how far between the two frames the animation is
		 * at the time the function was called.  A value of 0 means that the current
		 * time is equal to the start time while a value of 1 means that the current
		 * time is equal to the end time. */
		void GetMeshFrame(int* thisFrame, int* nextFrame, GCoordinate* t);
	};
	typedef GSmartPointer<SuperAnimState> pSuperAnimState;
	
	pAnimState CreateAnimState();
	
	// rendering info ----------------------------------------------
	pRenderInfo CreateRenderInfo();
};
/*! \typedef pGSuperModel
 *  \ingroup graphics
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GSuperModel GSuperModel\endlink. */
typedef GSmartPointer<GSuperModel> pGSuperModel;

} //namespace GAUGE3D

#endif //included
