/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_GRAPHICS_SURFACE_H__INCLUDED__
#define __GAUGE3D_GRAPHICS_SURFACE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include "gauge3d/3dmath/vector.h"
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_GRAPHICS_IMAGE_H__INCLUDED__
#include "image.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GSurface surface.h gauge3d/graphics/surface.h
 *  \ingroup graphics
 *  \brief Represents a surface map.
 *
 *  GSurface represents a surface map, which is like a texture map but
 *  maps more than just colors and alpha.  A point on a surface map is
 *  referenced by only one set of texture coordinates, but the layers
 *  of the map do not need to line up (see below). */
class GSurface: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x3a1e34e6);
public:
	// lifecycle -------------------------------------------------------
	virtual ~GSurface();
	
	// operations ------------------------------------------------------
	/*! \brief Describes how to interpret a particular channel of a map.
	 *
	 * All maps have between one and four channels.  The channels are split
	 * into the RGB channel(s) and the alpha channel.  Note that the RGB
	 * channel does not necessarily represent a color, and the alpha channel
	 * is not necessarily an opacity value.  They are only named as such for
	 * historical reasons.
	 *
	 * The channels of a map are broken up as follows:
	 * - <b>1 channel map:</b> The one channel is the RGB channel (think grayscale).
	 * - <b>2 channel map:</b> First channel is RGB, second channel is alpha.
	 * - <b>3 channel map:</b> All three channels go into the RGB component.
	 * - <b>4 channel map:</b> The first three channels are the RGB, and the fourth
	 *   is the alpha.
	 *
	 * The RGB and the alpha channels are each assigned a map type from the
	 * following values. */
	enum tMapType
	{
		/*! \brief The color (RGB or grayscale) of the image.
		 *
		 * Duh, this is just regular texture mapping.  Color can be RGB or grayscale.
		 * A surface may contain multiple color maps.  They will be multiplied together
		 * to get the final color.  (hint: Detail maps.) */
		COLOR,
		/*! \brief Transparency map.  This is different from alpha.
		 *
		 * When objects are viewed through a transparent polygon, the colors are
		 * modulated (multiplied) by the transparency map.  This looks much nicer
		 * than opacity maps (alpha transparency) in some situations, and can be
		 * used in combination with an opacity map to give good results in all
		 * situations.  The transparency map can be RGB color or grayscale. */
		TRANSPARENCY,
		/*! \brief Normal map, used for dot product bump mapping.
		 *
		 * A normal map must have three channels (duh).  The channels are the
		 * X, Y, and Z components of the normal vector to the surface.  These are
		 * then dot-product multiplied with the light direction to get per-pixel
		 * lighting. */
		NORMAL,
		
		/*! \brief This is the same thing as alpha mapping.
		 *
		 * An opacity map can only have one channel. */
		OPACITY,
		/*! \brief An opacity map where each pixel is either fully opaque or fully transparent.
		 *
		 * Each pixel in a binary opacity map is either 1.0 or 0.0.  When rendered, the
		 * edges between opaque regions and transparent regions will be drawn sharp rather
		 * than being bi-linearly interpolated like a regular opacity map. */
		BINARY_OPACITY,
		/*! \brief Gives the reflectivity of a surface.
		 *
		 * One channel.  The specular lighting and environmental reflection effects
		 * for each pixel will be multiplied by the gloss map. */
		GLOSS,
		/*! \brief The height of each pixel.
		 *
		 * This is a single-channel map which gives the height of each pixel.
		 * This can be used to draw an elevation map, which is done by drawing
		 * multiple layers, and is rather expensive.  Note that this has nothing
		 * to do with bump mapping. */
		ELEVATION
	};
	
	/*! \struct SurfaceMap surface.h /gauge3d/graphics/surface.h
	 *  \brief Describes a map of one or two aspects of a surface.
	 *
	 * Each surface map can map two aspects of a surface.  For example, an
	 * often-used surface map is one which maps colors and opacity values across
	 * the surface (this is your typical RGBA texture).  Other ideas for surface
	 * maps include:
	 * - Color and gloss (reflectivity)
	 * - Normals (bump map) and elevation
	 * - Transparency and opacity (they are different things!)
	 * - Any other combination where at least one of the two components is a
	 *   single-channel map.
	 *
	 * Each SurfaceMap in a GSurface is referenced by the same texture coordinates,
	 * but the texture coordinates are transformed differently for each map.  This
	 * allows you to apply maps offset from each other to prevent ugly patters from
	 * being visible.  The incoming texture coordinates are first rotated by mAngle,
	 * then scaled by mScale, then offset by mOffset.
	 *
	 * The more surface maps you have, the more rendering passes will be used to
	 * render the polygons that use this map.
	 *
	 * mThresholdNear and mThresholdFar allow you to control level-of-detail.  If the
	 * object to which this surface is being applied is further away than mThresholdFar[1]
	 * or closer than mThresholdNear[0], the map will not be used.  If the object is between
	 * mThresholdNear[1] and mThresholdFar[0], the map will be applied as normal.  In the
	 * area between mThresholdNear[0] and mThresholdNear[1], and in the area between
	 * mThresholdFar[0] and mThresholdFar[1], the map will be smoothly blended in and out
	 * to eliminate popup.  (Note that some renderers may not do anything about popup, or
	 * may use other methods to avoid it.) */
	struct SurfaceMap
	{
		pGImage mImage;         //!< The image to use for this map.
		tMapType mRGBType;      //!< How to interpret the RGB (or grayscale) channel(s) of the image.
		tMapType mAlphaType;    //!< How to interpret the alpha channel of the image.
		GCoordinate mOffset[2]; //!< Offset to apply to texture coordinates.
		GCoordinate mScale[2];  //!< Scale to apply to texture coordinates.
		GCoordinate mAngle;     //!< Angle by which to rotate texture coordinates. (radians)
		GCoordinate mThresholdNear[2]; //!< The distance at which to start showing this map. (see above)
		GCoordinate mThresholdFar[2];  //!< The distance at which to stop showing this map. (see above)
	};
	
	virtual GArray<SurfaceMap> SurfaceMaps() = 0;
};
/*! \typedef pGSurface
 *  \ingroup graphics
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GSurface GSurface\endlink. */
typedef GSmartPointer<GSurface> pGSurface;

} //namespace GAUGE3D

#endif //included
