/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "inputmap.h"

// system ------------------------------------------------------------
#include <map>
#include <list>
using namespace std;

// external packages -------------------------------------------------
#include "gauge3d/scheduler.h"

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GInputMap::Private
{
	struct MapEntry
	{
		GString mOutputName;
		int mOutputType;
		float mSensitivity;
		float mLastValue;
		tTime mLastTime;
	};
	
	typedef map<GString, MapEntry> tInputMap;
	typedef map<GString, int> tControlList;
	
	typedef list<MapEntry*> tTaskList;
	
	// state variables -------------------------------------------------
	pGMessageQueue mQueue;
	
	tInputMap mInputMap;
	tControlList mControlList;
	
	tTaskList mTaskList;
};

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GInputMap::GInputMap(pGMessageQueue outputQueue)
	: mPrivate(*(new Private))
{
	mPrivate.mQueue = outputQueue;
	gInputScheduler->CallEveryFrame(this, 0);
}

GInputMap::~GInputMap()
{
	gInputScheduler->StopCallingEveryFrame(this);
	delete &mPrivate;
}

// operations --------------------------------------------------------
bool
GInputMap::HandleMessage(pGMessage message)
{
	pGInputMessage input = QueryInterface<GInputMessage>(message);
	float outputValue;
	tTime time = gInputScheduler->Time();
	
	if(input == NULL)
		return false;   //not an input message
	
	Private::tInputMap::iterator i;
	
	i = mPrivate.mInputMap.find(input->Name());
	if(i == mPrivate.mInputMap.end())
		return false;   //not on the map
	
	Private::MapEntry* control = &i->second;
	
	//need to translate the value from the input type to the output type,
	// and apply sensitivity.
	switch(input->Type())
	{
	default:
	case (GInputMessage::DIGITAL | GInputMessage::ABSOLUTE):
		switch(control->mOutputType)
		{
		default:
		case (GInputMessage::DIGITAL | GInputMessage::ABSOLUTE):
			//no translation needed.
			outputValue = input->Value();
			break;
		case (GInputMessage::DIGITAL | GInputMessage::RELATIVE):
			//ignore button up.
			if(input->Value() == 0.0)
				return true;
			//sensitivity will be 1.0 or -1.0.
			outputValue = input->Value() * control->mSensitivity;
			break;
		case (GInputMessage::ANALOG | GInputMessage::ABSOLUTE):
			//simple sensitivity translation.
			outputValue = input->Value() * control->mSensitivity;
			break;
		case (GInputMessage::ANALOG | GInputMessage::RELATIVE):
			//need to send every frame...
			if(input->Value() > 0.5f)
			{
				control->mLastValue = input->Value() * control->mSensitivity;
				control->mLastTime = time;
				mPrivate.mTaskList.push_back(control);
			}
			else
				mPrivate.mTaskList.remove(control);
			return true;   //no message to send for now
		}
		break;
		
	case (GInputMessage::DIGITAL | GInputMessage::RELATIVE):
		switch(control->mOutputType)
		{
		default:
		case (GInputMessage::DIGITAL | GInputMessage::ABSOLUTE):
			//no translation possible.
			return false;
		case (GInputMessage::DIGITAL | GInputMessage::RELATIVE):
			//sensitivity may be used to reverse the direction
			outputValue = input->Value() * control->mSensitivity;
			break;
		case (GInputMessage::ANALOG | GInputMessage::ABSOLUTE):
			//each click adds a certain amount to the output value.
			outputValue = control->mLastValue +
			              input->Value() * control->mSensitivity;
			//clamp to [-1.0, 1.0]
			if(outputValue < -1.0)
				outputValue = -1.0;
			else if(outputValue > 1.0)
				outputValue = 1.0;
			//save value for next time
			control->mLastValue = outputValue;
			break;
		case (GInputMessage::ANALOG | GInputMessage::RELATIVE):
			//simple sensitivity translation.
			outputValue = input->Value() * control->mSensitivity;
			break;
		}
		break;
		
	case (GInputMessage::ANALOG | GInputMessage::ABSOLUTE):
		switch(control->mOutputType)
		{
		default:
		case (GInputMessage::DIGITAL | GInputMessage::ABSOLUTE):
			//threshold translation
			if(input->Value() > control->mSensitivity)
				outputValue = 1.0;
			else
				outputValue = 0.0;
			//don't send if not necessary
			if(outputValue == control->mLastValue)
				return true;
			control->mLastValue = outputValue;
			break;
		case (GInputMessage::DIGITAL | GInputMessage::RELATIVE):
			//threshold translation
			if(input->Value() > control->mSensitivity)
				outputValue = 1.0;
			else if(input->Value() < -control->mSensitivity)
				outputValue = -1.0;
			else
				outputValue = 0.0;
			//don't send if not necessary
			if(outputValue == control->mLastValue)
				return true;
			control->mLastValue = outputValue;
			if(outputValue == 0.0)
				return true;
			break;
		case (GInputMessage::ANALOG | GInputMessage::ABSOLUTE):
			//simple sensitivity translation
			outputValue = input->Value() * control->mSensitivity;
			break;
		case (GInputMessage::ANALOG | GInputMessage::RELATIVE):
			//need to send every frame...
			if(input->Value() != 0.0f)
			{
				control->mLastValue = input->Value() * control->mSensitivity;
				control->mLastTime = time;
				mPrivate.mTaskList.push_back(control);
			}
			else
				mPrivate.mTaskList.remove(control);
			return true;   //no message to send for now
		}
		break;
		
	case (GInputMessage::ANALOG | GInputMessage::RELATIVE):
		switch(control->mOutputType)
		{
		default:
		case (GInputMessage::DIGITAL | GInputMessage::ABSOLUTE):
			//no translation possible
			return false;
		case (GInputMessage::DIGITAL | GInputMessage::RELATIVE):
			//when you move past the threshold in either direction,
			// it triggers the output.
			control->mLastValue += input->Value();
			while(control->mLastValue > control->mSensitivity)
			{
				control->mLastValue -= control->mSensitivity * 2;
				outputValue = 1.0;
			}
			while(control->mLastValue < -control->mSensitivity)
			{
				control->mLastValue += control->mSensitivity * 2;
				outputValue = -1.0;
			}
			return false;
		case (GInputMessage::ANALOG | GInputMessage::ABSOLUTE):
			//keep a running total
			outputValue = control->mLastValue +
			              input->Value() * control->mSensitivity;
			//clamp to [-1.0, 1.0]
			if(outputValue < -1.0)
				outputValue = -1.0;
			else if(outputValue > 1.0)
				outputValue = 1.0;
			//save value for next time
			control->mLastValue = outputValue;
			break;
		case (GInputMessage::ANALOG | GInputMessage::RELATIVE):
			//simple sensitivity translation
			outputValue = input->Value() * control->mSensitivity;
			break;
		}
		break;
	}
	
	//The engine is already in the message dispatching phase, so we don't want
	// our messages hanging around in any queues.
	mPrivate.mQueue->DispatchMessage(
		new GInputMessage(input->Origin(), control->mOutputName, outputValue, control->mOutputType));
	return true;
}

void
GInputMap::NewControl(GString name, int type)
{
	mPrivate.mControlList[name] = type;
}

void
GInputMap::MapInput(GString input, GString control, float sensitivity)
{
	Private::MapEntry entry;
	
	entry.mOutputName = control;
	entry.mOutputType = mPrivate.mControlList[control];
	entry.mSensitivity = sensitivity;
	entry.mLastValue = 0.0f;
	entry.mLastTime = gInputScheduler->Time();
	mPrivate.mInputMap[input] = entry;
}

void
GInputMap::ControlList(pGDirectory config)
{
	GArray<GString> list = config->ListFiles();
	GString entry, data;
	GString word;
	pGFile file;
	int start, end;
	int type;
	
	for(int i = 0; i < list.Size(); i++)
	{
		entry = list[i];
		
		//parse entry into words and interpret
		file = config->OpenFile(entry);
		data = (GString)(*file);
		
		start = 0;
		end = data.FindFirstOf(' ');
		type = 0;
		
		while(end >= 0)
		{
			word = data.SubString(start, end - start);
			
			//since DIGITAL and ABSOLUTE are both zero, we don't really have to
			// bother interpreting them.
			if(word == "analog")
				type |= GInputMessage::ANALOG;
			else if(word == "relative")
				type |= GInputMessage::RELATIVE;
			
			start = end + 1;
			end = data.FindNext(' ', start);
		}
		
		word = data.SubString(start);
		
		//since DIGITAL and ABSOLUTE are both zero, we don't really have to
		// bother interpreting them.
		if(word == "analog")
			type |= GInputMessage::ANALOG;
		else if(word == "relative")
			type |= GInputMessage::RELATIVE;
		
		NewControl(entry, type);
	}
}

void
GInputMap::ControlMap(pGDirectory config)
{
	GArray<GString> list = config->ListFiles();
	GString entry, data;
	GString name, modifier;
	float sensitivity;
	pGFile file;
	int start, end;
	int type;
	
	if(list == NULL)
		return;
	
	for(int i = 0; i < list.Size(); i++)
	{
		entry = list[i];
		
		//parse entry into parts and interpret
		
		//Each part has the following syntax:
		//   "input name"*sensitivity
		//for example, the config file section might look like:
		//
		//   move X = "left"*-1.0 "s"*-1.0 "right"*1.0 "d"*1.0 "joystick x"*0.83
		//   move Y = "up"*1.0 "w"*1.0 "down"*-1.0 "a"*-1.0 "joystick y"*0.83
		//   fire = "space" "mouse 1"
		
		file = config->OpenFile(entry);
		data = (GString)(*file);
		
		start = data.FindFirstOf('\"') + 1;
		end = data.FindNext('\"', start);
		
		type = 0;
		
		
		while(end >= 0)
		{
			name = data.SubString(start, end - start);
			
			start = end + 1;
			end = data.FindNext('\"', start);
			
			if(end >= 0)
				modifier = data.SubString(start, end - start);
			else
				modifier = data.SubString(start);
			
			if(modifier.Length() >= 2)
			{
				if(modifier[0] == '*')
					sensitivity = (float)modifier.SubString(1);
				else
					sensitivity = 1.0f;  //default sensitivity
			}
			else
				sensitivity = 1.0f;  //default sensitivity
			
			MapInput(name, entry, sensitivity);
			
			if(end >= 0)
			{
				start = end + 1;
				end = data.FindNext('\"', start);
			}
		}
	}
}

void
GInputMap::ClearControlList()
{
	mPrivate.mControlList.clear();
}

void
GInputMap::ClearControlMap()
{
	mPrivate.mInputMap.clear();
}

void
GInputMap::DoFrame(tTime time, tTime frameTime)
{
	Private::tTaskList::iterator i = mPrivate.mTaskList.begin();
	float value;
	
	for(; i != mPrivate.mTaskList.end(); i++)
	{
		value = (*i)->mLastValue * (time - (*i)->mLastTime);
		(*i)->mLastTime = time;
		
		mPrivate.mQueue->EnqueueMessage(
			new GInputMessage(NULL, (*i)->mOutputName, value, (*i)->mOutputType));
	}
}

GAUGE3D_DECLARE_INTERFACES(GInputMap)
	GAUGE3D_DECLARE_BASECLASS(Handler)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
