/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_INPUT_INPUTMAP_H__INCLUDED__
#define __GAUGE3D_INPUT_INPUTMAP_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_OSABSTRACTION_TIMER_H__INCLUDED__
#include "gauge3d/osabstraction/timer.h"
#endif //redundant include guard

#ifndef __GAUGE3D_FILES_VIRTUALDIRECTORY_H__INCLUDED__
#include "gauge3d/files/virtualdirectory.h"
#endif //redundant include guard

// local package -----------------------------------------------------
#ifndef __GAUGE3D_INPUT_MESSAGEQUEUE_H__INCLUDED__
#include "messagequeue.h"
#endif //redundant include guard

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GInputMap inputmap.h gauge3d/input/inputmap.h
 *  \ingroup input
 *  \brief Maps user-configured input to game-specific controls.
 *
 *  GInputMap makes it easy to allow users to configure their controls
 *  rather than use forced key bindings.  It also translates the input
 *  type to the control type (analog vs. digital, absolute vs. relative)
 *  and applies sensitivity settings to the controls.
 *
 *  To make the input map translate messages, you must register it as a
 *  message handler in the input message queue.  Also, make sure to call
 *  GInputMap::DoFrame() every frame. */
class GAUGE3D_API GInputMap: public GMessageQueue::Handler
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39da68c6);
public:
	// lifecycle -------------------------------------------------------
	/*! \brief Constructs an input map.
	 *  \param outputQueue A message queue to which the translated messages should be sent.
	 *                     This must not be the same as the input queue for obvious reasons. */
	GInputMap(pGMessageQueue outputQueue);
	//! Destructor
	virtual ~GInputMap();
	
	// operations ------------------------------------------------------
	//! Translates and dispatches a message (called by the input message queue).
	virtual bool HandleMessage(pGMessage message);
	
	/*! \brief Add a new control to the list of game-specific controls.
	 *  \param name The name that the translated messages will have.
	 *  \param type The type of the control.  This is a bitwise OR of either
	 *              GInputMessage::DIGITAL or GInputMessage::ANALOG with either
	 *              GInputMessage::ABSOLUTE or GInputMessage::RELATIVE.
	 *
	 *  This function adds a new game-specific control to the list of controls
	 *  which the user can configure, but it does not actually bind an input source
	 *  to it.  Use MapInput to bind one or more input sources to the control.
	 *
	 *  When an input source is mapped to a control, the message is translated from
	 *  the input's type to the control's type, so you don't have to worry about
	 *  handling any type of input other than what you expect, yet the user is free
	 *  to bind any sort of input to any control (almost).
	 *
	 *  Choosing the right type for the control is tricky, especially if it is an
	 *  analog control.  Here are what the four types represent:
	 *  - Absolute Digital:  Any sort of simple button-pressing control, like fire,
	 *    use, jump, etc.  You will recieve messages when the user presses or releases
	 *    the button.
	 *
	 *  - Relative Digital:  Something that cycles in two directions by discreet amounts.
	 *    For example, a control that cycles though your inventory would be relative
	 *    digital.  You will receive messages when the user cycles forward or back.
	 *    Think mouse wheel.
	 *
	 *  - Absolute Analog:  Something which has defined limits.  The only really way
	 *    I can describe the difference between this and relative analog is by explaining
	 *    how the translation is done.  When mapping any sort of absolute input to an
	 *    absolute analog control, the output value is simply the same as the input (after
	 *    sensitivity is applied).  When mapping a relative input to an absolute analog
	 *    output, a running total of the input is kept and clamped to the range [-1.0, 1.0],
	 *    and that total is output.  An example of an absolute analog control would be a
	 *    glance left/right function, where you can only turn your head so far to the left
	 *    or right and the view should snap back to center afterwards.
	 *
	 *  - Relative Analog:  A control with no limits.  Again, I can't think of a good way
	 *    to explain this without explaining exactly how the mapping is done.  When mapping
	 *    any sort of relative input to a relative analog control, the output value is
	 *    simply the same as the input (after sensitivity is applied).  When mapping absolute
	 *    input, messages are sent every frame as long as the input value is not zero.  So,
	 *    the input becomes sort of a rate of change, as if the input were a mouse and were
	 *    moving at exactly that speed.  Unfortunately, the explanation I just gave you
	 *    sucked, so you will have to experiment a little or I will have to get a better
	 *    documentation writer.  Examples of relative analog controls include movement,
	 *    turning, looking, etc.  */
	void NewControl(GString name, int type);
	/*! \brief Maps an input source to a control with a given sensitivity.
	 *  \param input   The name of the input source to map.
	 *  \param control The name of the control to which the input should be mapped.
	 *  \param sensitivity The sensitivity of the input.  How exactly this is used depends
	 *                 on the types of both the input and the control.  Usually, the input
	 *                 value is multiplied by the sensitivity, but in some cases (usually
	 *                 when converting from analog input to a digital control) the
	 *                 sensitivity is actually a threshold value.
	 *  \note You can not map an input to a control before you create the control by
	 *        calling NewControl (or ControlList()). */
	void MapInput(GString input, GString control, float sensitivity = 1.0);
	
	/*! \brief Loads a list of controls from a config file.
	 *  \param config The directory representing the config file.  See GConfigFile for
	 *                more info on how this works.
	 *
	 *  The format of the config file is as follows:
	 *
	 *  <i>config</i> is one section of the config file.  Each line in that section
	 *  should have the form "control name = [absolute|relative] [digital|analog]".
	 *  Here's an example:
	 *
	 *  [controls]<br>
	 *    cycle weapon = relative digital<br>
	 *    fire = absolute digital<br>
	 *    look Y = absolute analog<br>
	 *    move X = analog relative<br>
	 *    move Y = analog relative<br>
	 *  [/controls]
	 *
	 */
	void ControlList(pGDirectory config);
	/*! \brief Loads the input to control map from a config file.
	 *  \param config The directory representing the config file.  See GConfigFile for
	 *                more info on how this works.
	 *  \note You must load the control list before loading the input map.
	 *
	 *  The format of the config file is as follows:
	 *
	 *  <i>config</i> is one section of the config file.  Each line in that section
	 *  should have the form:<br>
	 *  'control name = "input name"[*sensitivity] ["input2 name"[*sensitivity] [...]]
	 *
	 *  Here's an example:
	 *
	 *  [control map]<br>
	 *    cycle weapon = "mouse wheel"<br>
	 *    fire = "space" "mouse 1"<br>
	 *    look Y = "mouse y"*2.0<br>
	 *    move X = "left"*-1.0 "s"*-1.0 "right"*1.0 "d"*1.0 "joystick x"*0.83<br>
	 *    move Y = "up"*1.0 "w"*1.0 "down"*-1.0 "a"*-1.0 "joystick y"*0.83<br>
	 *  [/control map]
	 */
	void ControlMap(pGDirectory config);
	
	//! Empties the list of controls.
	void ClearControlList();
	//! Erases all input-to-control mappings.
	void ClearControlMap();
	
	/*! \brief This will be called every frame by the scheduler.
	 *  \remarks You can safely pretend this function does not exist.  Everything
	 *           will be handled for you. :) */
	virtual void DoFrame(tTime time, tTime frameTime);
};
/*! \typedef pGInputMap
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GInputMap GInputMap\endlink. */
typedef GSmartPointer<GInputMap> pGInputMap;

} //namespace GAUGE3D

#endif //included
