/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "messagebox.h"

// system ------------------------------------------------------------
#ifdef PLATFORM_WINDOWS
# include <windows.h>
# undef ERROR
# undef IGNORE
# include <exception>
#endif

#include <iostream>
using namespace std;

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GMessageBox::Private
{
	GString mMessage;
	tEnum  mType;
	tEnum  mButtons;
};

namespace { //local
	const char* BUTTONS_ABORT_RETRY_IGNORE[] = {"abort", "retry", "ignore", NULL};
	const char* BUTTONS_OK[]                 = {"OK", NULL};
	const char* BUTTONS_OK_CANCEL[]          = {"OK", "cancel", NULL};
	const char* BUTTONS_RETRY_CANCEL[]       = {"retry", "cancel", NULL};
	const char* BUTTONS_YES_NO[]             = {"yes", "no", NULL};
	const char* BUTTONS_YES_NO_CANCEL[]      = {"yes", "no", "cancel", NULL};
} //local namespace

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GMessageBox::GMessageBox(GString message, tEnum type, tEnum buttons)
	: mPrivate(*(new Private))
{
	mPrivate.mMessage = message.ToAscii();
	mPrivate.mType    = type;
	mPrivate.mButtons = buttons;
}

GMessageBox::~GMessageBox()
{
}

// operations --------------------------------------------------------
//Run displays the message box and does not return until the user has
// clicked on a button.  The return value is the button that was clicked.
GMessageBox::tEnum
GMessageBox::Run()
{
#ifdef PLATFORM_WINDOWS
	UINT flags = 0;

	switch(mPrivate.mType)
	{
	case WARNING:
		flags = MB_ICONWARNING;
		break;
	case ERROR:
		flags = MB_ICONERROR;
		break;
	case QUESTION:
		flags = MB_ICONQUESTION;
		break;
	default:
	case INFORMATION:
		flags = MB_ICONINFORMATION;
		break;
	}

	switch(mPrivate.mButtons)
	{
	case ABORT_RETRY_IGNORE:
		flags |= MB_ABORTRETRYIGNORE;
		break;
	default:
	case OK:
		flags |= MB_OK;
		break;
	case OK_CANCEL:
		flags |= MB_OKCANCEL;
		break;
	case RETRY_CANCEL:
		flags |= MB_RETRYCANCEL;
		break;
	case YES_NO:
		flags |= MB_YESNO;
		break;
	case YES_NO_CANCEL:
		flags |= MB_YESNOCANCEL;
		break;
	}
	
	int choice = MessageBox(NULL, mPrivate.mMessage.AsciiData(), "GAUGE^3D", flags);

	switch(choice)
	{
	case IDABORT:
		return ABORT;
		break;
	case IDCANCEL:
		return CANCEL;
		break;
	case IDIGNORE:
		return IGNORE;
		break;
	case IDNO:
		return NO;
		break;
	case IDOK:
		return OK;
		break;
	case IDRETRY:
		return RETRY;
		break;
	case IDYES:
		return YES;
		break;
	default:
		break;
	}
	
	//according to the docs, MessageBox will succeed unless it can't allocate
	// enough memory...
	throw bad_alloc();

#else
	int choice;
//	if(gui available)
//	{
//		Use GUI;
//	}
//	else
	{
		//No GUI options available... fall back to text I/O.
		const char** buttons;
		
		cout << "----------------------------------------" << endl;
		
		switch(mPrivate.mType)
		{
		case WARNING:
			cout << "WARNING:" << endl;
			break;
		case ERROR:
			cout << "ERROR:" << endl;
			break;
		case QUESTION:
			cout << "Question:" << endl;
			break;
		case INFORMATION:
			cout << "Message:" << endl;
			break;
		default:
			cout << "Message:" << endl;
			break;
		}
		
		cout << mPrivate.mMessage.AsciiData() << endl << endl;
		
		switch(mPrivate.mButtons)
		{
		case ABORT_RETRY_IGNORE:
			buttons = BUTTONS_ABORT_RETRY_IGNORE;
			break;
		default:
		case OK:
			buttons = BUTTONS_OK;
			break;
		case OK_CANCEL:
			buttons = BUTTONS_OK_CANCEL;
			break;
		case RETRY_CANCEL:
			buttons = BUTTONS_RETRY_CANCEL;
			break;
		case YES_NO:
			buttons = BUTTONS_YES_NO;
			break;
		case YES_NO_CANCEL:
			buttons = BUTTONS_YES_NO_CANCEL;
			break;
		}
		
		int i = 0;
		while(buttons[i])
		{
			cout << "(" << (i + 1) << ") " << buttons[i] << endl;
			i++;
		}
		
		cout << "----------------------------------------" << endl;
		cout << "Enter choice: " << endl;
		
		while(true)
		{
			cin >> choice;
			if((choice > 0) && (choice <= i))
				break;
			
			cout << "Invalid choice, try again: " << endl;
		}
		--choice;
	}
	
	switch(mPrivate.mButtons)
	{
	case ABORT_RETRY_IGNORE:
		if(choice == 0)
			return ABORT;
		else if(choice == 1)
			return RETRY;
		else
			return IGNORE;
	default:
	case OK:
		return OK;
	case OK_CANCEL:
		if(choice == 0)
			return OK;
		else
			return CANCEL;
	case RETRY_CANCEL:
		if(choice == 0)
			return RETRY;
		else
			return CANCEL;
	case YES_NO:
		if(choice == 0)
			return YES;
		else
			return NO;
	case YES_NO_CANCEL:
		if(choice == 0)
			return YES;
		else if(choice == 1)
			return NO;
		else
			return CANCEL;
	}
	return OK;  //should never get here
#endif
}


GAUGE3D_DECLARE_INTERFACES(GMessageBox)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
