/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_OSABSTRACTION_MESSAGEBOX_H__INCLUDED__
#define __GAUGE3D_OSABSTRACTION_MESSAGEBOX_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include "gauge3d/common/gstring.h"
#endif //redundant include guard

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GMessageBox messagebox.h gauge3d/osabstraction/messagebox.h
 *  \ingroup osabstraction
 *  \brief Opens a simple messagebox using the native GUI.
 *
 *  This class can be used to display simple message boxes before the graphical
 *  sub-system is available.  It should not be used once the graphics system
 *  is running, however, as the message box may not be visible to the user
 *  at that time. */
class GAUGE3D_API GMessageBox: public GObject
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x399aa771);
public:
	/*! \brief Flags and settings used by GMessageBox.
	 *  \remarks Sorry about the limited button sets.  That was Windows' fault. :( */
	enum tEnum
	{
		WARNING,            //!< Message type
		ERROR,              //!< Message type
		QUESTION,           //!< Message type
		INFORMATION,        //!< Message type
		
		//Windows mandates that we use one of these.  :(
		ABORT_RETRY_IGNORE, //!< Button set
		OK_CANCEL,          //!< Button set
		RETRY_CANCEL,       //!< Button set
		YES_NO,             //!< Button set
		YES_NO_CANCEL,      //!< Button set
		OK,                 //!< Button set / User response
		
		ABORT,              //!< User response
		CANCEL,             //!< User response
		RETRY,              //!< User response
		IGNORE,             //!< User response
		YES,                //!< User response
		NO                  //!< User response
	};
	
	// lifecycle -------------------------------------------------------
	/*! \brief Create a message box, but don't display it yet.
	 * \param message The text of the message to display.
	 * \param type    One of the message type values in \ref tEnum.
	 * \param buttons One of the button sets in \ref tEnum. */
	GMessageBox(GString message, tEnum type, tEnum buttons);
	//! Destroy the message box.
	~GMessageBox();
	
	// operations ------------------------------------------------------
	/*! \brief Run the message box (modal).
	 * \returns One of the user response values in \ref tEnum.  This is the
	 *          button which the user clicked on.
	 * \remarks This function does not return until the user has clicked on a
	 *          button. */
	tEnum Run();
};
/*! \typedef pGMessageBox
 *  \ingroup osabstraction
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GMessageBox GMessageBox\endlink. */
typedef GSmartPointer<GMessageBox> pGMessageBox;

} //namespace GAUGE3D

#endif //included
