/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_PLUGINS_OBJECTLOADER_H__INCLUDED__
#define __GAUGE3D_PLUGINS_OBJECTLOADER_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_FILES_FILE_H__INCLUDED__
#include "gauge3d/files/file.h"
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_PLUGINS_PLUGIN_H__INCLUDED__
#include "plugin.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {

class GFile;

class GPlugin;
class GDataFile;

// class definition ==================================================
/*! \class GObjectLoader objectloader.h gauge3d/plugins/objectloader.h
 *  \ingroup plugins
 *  \brief Loads objects from plugins.
 *
 *  This class acts sort of like a CORBA ORB, except for GAUGE plugin objects.
 *  (Disclaimer:  I've never used CORBA, so I don't actually know how a CORBA
 *  ORB works.  Let me know if that last statement was wrong.)  The class
 *  loads plugins and queries them to see what objects can be loaded from
 *  them, and what functions those objects can perform.  It then fields
 *  requests for objects that perform certain types of functions by matching
 *  the request with a plugin object that performs the task and getting
 *  an instance of the object from the corresponding plugin. */
class GAUGE3D_API GObjectLoader: public GObject
{
	// insulated state -------------------------------------------------
	struct Private;
	Private& mPrivate;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f96);
public:
	// lifecycle -------------------------------------------------------
	GObjectLoader();
	~GObjectLoader();
	
	// operations ------------------------------------------------------
	
	// loading objects -------------------
	/*! \brief Find an object with the given class name and create it.
	 *  \returns A pointer to the object, or NULL if an obect with the given
	 *           name could not be found. */
	pGObject LoadObject(const GString& className);
	
	/*! \brief Find an object derived from the specified class and create it.
	 *  \returns A pointer to the object, or NULL if an obect derived from the
	 *           given class could not be found. */
	pGObject LoadObjectOfType(const GString& baseClass);
	
	/*! \brief Find an object with the given keyword and create it.
	 *  \returns A pointer to the object, or NULL if an obect with the given
	 *           keyword could not be found.
	 *
	 *  Each object can have several keywords.  See GPlugin::ObjectInfo::mKeywords
	 *  for more info on them. */
	pGObject LoadObjectWithKeyword(const GString& keyWord, const GString& baseClass);
	
	/*! \brief Find an object that can load the given file and that is
	 *         derived from the given base class and create it.
	 *  \param file      The file that the object should load.
	 *  \param baseClass The class that the object should be derived from.
	 *
	 *  So, if you have a jpeg image file "jpegFile", you might say:
	 *  <ul><li>objectLoader.LoadObject(jpegFile, "GImage");</ul> */
	pGObject LoadObjectForFile(pGFile file, const GString& baseClass);
	
	// searching for objects ------------
	/*! \typedef ObjectInfo
	 *  \brief Just links to \ref GPlugin::ObjectInfo. */
	typedef GPlugin::ObjectInfo ObjectInfo;
	
	//these functions allow you to search for particular objects to fulfill
	// your needs.
	GSmartPointer<ObjectInfo> FindObject(const GString& classname);
	GArray<ObjectInfo> FindObjectsOfType(const GString& baseClass);
	GArray<ObjectInfo> FindObjectsWithKeyword(const GString& keyWord);
	GArray<ObjectInfo> FindObjectsWithKeyword(const GString& keyWord, const GString& baseClass);
	GArray<ObjectInfo> FindObjectsForFile(pGFile file);
	GArray<ObjectInfo> FindObjectsForFile(pGFile file, const GString& baseClass);
	GArray<ObjectInfo> FindAllObjects();
	
	// searching for plugins ------------
	struct PluginInfo
	{
		GString mName;
		GString mDescription;
		GString mLibName;
		GArray<ObjectInfo> mObjects;
	};
	GArray<PluginInfo> FindAllPlugins();
	GSmartPointer<PluginInfo> FindPluginForObject(GString classname);
	
	// adding plugins -------------------
	//! Add a library to the list of plugins to search for objects.
	void AddPlugin(pGFile library);
};
/*! \typedef pGObjectLoader
 *  \ingroup plugins
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GObjectLoader GObjectLoader\endlink. */
typedef GSmartPointer<GObjectLoader> pGObjectLoader;

/*! \brief The main GAUGE object loader.
 *  \ingroup plugins
 *
 *  In almost all cases, you'll want to use this object loader rather than
 *  create your own.  Plugins that need to load other plugins will use this
 *  object loader, and they will fail if this object loader does not have
 *  any plugins in it. */
extern GAUGE3D_API pGObjectLoader gMainObjectLoader;

} // namespace GAUGE3D

#endif //included
