/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

// includes ==========================================================
// component header --------------------------------------------------
#include "scheduler.h"

// system ------------------------------------------------------------
#include <list>
using namespace std;

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class state definition ============================================
struct GScheduler::Private
{
	// state variables -------------------------------------------------
	tTime mTime;
	tTime mFrameTime;
	
	struct tObjectEntry
	{
		GObject* mObject;
		int      mPriority;
	};
	
	typedef list<tObjectEntry> tObjectList;
	tObjectList mObjectList;
};

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GScheduler::GScheduler()
	: mPrivate(*(new Private))
{
	mPrivate.mTime = 0.0f;
	mPrivate.mFrameTime = 0.0f;
}

GScheduler::~GScheduler()
{
	delete &mPrivate;
}

// operations --------------------------------------------------------
void
GScheduler::CallEveryFrame(GObject* object, int priority)
{
	Private::tObjectEntry entry;
	
	entry.mObject   = object;
	entry.mPriority = priority;
	
	Private::tObjectList::iterator i   = mPrivate.mObjectList.begin();
	Private::tObjectList::iterator end = mPrivate.mObjectList.end();
	
	while(i != end)
	{
		if(priority <= i->mPriority)
			break;
		++i;
	}
	
	mPrivate.mObjectList.insert(i, entry);
}

void
GScheduler::StopCallingEveryFrame(GObject* object)
{
	Private::tObjectList::iterator i   = mPrivate.mObjectList.begin();
	Private::tObjectList::iterator end = mPrivate.mObjectList.end();
	
	while(i != end)
	{
		if(i->mObject == object)
		{
			mPrivate.mObjectList.erase(i);
			break;
		}
		++i;
	}
}

int
GScheduler::GetObjectPriority(GObject* object)
{
	Private::tObjectList::iterator i   = mPrivate.mObjectList.begin();
	Private::tObjectList::iterator end = mPrivate.mObjectList.end();
	
	while(i != end)
	{
		if(i->mObject == object)
			return i->mPriority;;
		++i;
	}
	
	return -1;
}

void
GScheduler::SetObjectPriority(GObject* object, int priority)
{
	Private::tObjectList::iterator i   = mPrivate.mObjectList.begin();
	Private::tObjectList::iterator end = mPrivate.mObjectList.end();
	
	//if the object is not in the object list, we don't want to add it.
	while(i != end)
	{
		if(i->mObject == object)
		{
			mPrivate.mObjectList.erase(i);
			CallEveryFrame(object, priority);
			break;
		}
		++i;
	}
}

void
GScheduler::DoFrame(tTime time, tTime frametime)
{
	Private::tObjectList::iterator i   = mPrivate.mObjectList.begin();
	Private::tObjectList::iterator end = mPrivate.mObjectList.end();
	
	mPrivate.mTime      = time;
	mPrivate.mFrameTime = frametime;
	
	while(i != end)
	{
		i->mObject->DoFrame(time, frametime);
		++i;
	}
}

// access ------------------------------------------------------------
tTime
GScheduler::Time()
{
	return mPrivate.mTime;
}

tTime
GScheduler::FrameTime()
{
	return mPrivate.mFrameTime;
}

pGScheduler gMiscScheduler;
pGScheduler gInputScheduler;
pGScheduler gVideoScheduler;


GAUGE3D_DECLARE_INTERFACES(GScheduler)
	GAUGE3D_DECLARE_BASECLASS(GObject)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
