/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_3DMATH_PLANE_H__INCLUDED__
#define __GAUGE3D_3DMATH_PLANE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include"vector.h"
#endif //redundant include guard

// begin namespace ===================================================
namespace GAUGE3D {

class GLine;

// class definition ==================================================
/*! \class GPlane plane.h gauge3d/3dmath/plane.h
 * \ingroup d3math
 * \brief Represents a plane in 3D space.
 *
 * The plane is represented as a normal vector and a scalar so that
 * n * p + d = 0 for all points p on the plane. */
class GAUGE3D_API GPlane
{
public:
	GVector     n;        //!< Normal vector to plane.
	GCoordinate d;        //!< Umm...  d...  it's d...  yeah...

	// lifecycle -------------------------------------------------------
	GPlane();             //!< Creates an uninitialized plane.
//	d3m_Plane(const d3m_Plane& other);  //uses default copy constructor
	//! Creates a plane with the given normal that contains the given point.
	GPlane(const GVector& point, const GVector& normal);
	//! Initialized n and d to the given values.
	GPlane(const GVector& normal, GCoordinate _d);

	// operations ------------------------------------------------------
	/*! \brief Returns true if the plane is facing the point.
	 *
	 * The plane is considered to be "facing" the point if the normal vector
	 * of the plane points in the direction of the point from any point on the
	 * plane.  In other (non-)words: (point * n + d >= 0) */
	bool IsFacing(const GVector& point) const;
	
	/*! \brief Finds the intersection between the plane and a line.
	 * \param result Pointer to a vector which should be overwritten with
	 *               the point of intersection.  If NULL, it will not be
	 *               overwritten.
	 * \param line   The line with which to test for intersection.
	 * \returns      True if a point of intersection was found, or false if
	 *               the line and plane did not intersect or if the line was
	 *               on the plane. */
	bool FindIntersection(GVector* result, const GLine& line) const;
	/*! \brief Finds the intersection between two planes.
	 * \param result Pointer to a line which should be overwritten with
	 *               the line of intersection.  If NULL, it will not be
	 *               overwritten.
	 * \param other  The plane with which to test for intersection.
	 * \returns      True if a line of intersection was found, or false if
	 *               the planes did not intersect or if they were equal. */
	bool FindIntersection(GLine* result, const GPlane& other) const;
	/*! \brief Finds the intersection between three planes.
	 * \param result Pointer to a point which should be overwritten with
	 *               the point of intersection.  If NULL, it will not be
	 *               overwritten.
	 * \param other1 One of the two planes with which to test for intersection.
	 * \param other2 The other of the two planes with which to test for intersection.
	 * \returns      True if a point of intersection was found, or false if
	 *               the planes did not intersect or if two or more of them were equal. */
	bool FindIntersection(GVector* result, const GPlane& other1, GPlane& other2) const;
	
	//! Returns true if the point is on the plane.
	bool Contains(const GVector& point) const;

	//! Finds the shortest distance between the plane and the point.
	GCoordinate FindDistance(const GVector& point) const;
	/*! \brief Finds the square of the shortest distance between the plane and the point (faster).
	 *
	 *  This is faster because the square root of the result does not need to be found. */
	GCoordinate FindSquareDistance(const GVector& point) const;
	
	// operators -------------------------------------------------------
	//! Moves the plane a certain distance.
	const GPlane& operator+=(const GVector& vec);
	//! Moves the plane a certain distance.
	const GPlane& operator-=(const GVector& vec);
};

// inline member functions ===========================================
// lifecycle ---------------------------------------------------------
inline
GPlane::GPlane()
{
}

inline
GPlane::GPlane(const GVector& point, const GVector& normal)
	: n(normal),
	  d(-(point * normal))
{
}

inline
GPlane::GPlane(const GVector& normal, GCoordinate _d)
	: n(normal),
	  d(_d)
{
}

// operations --------------------------------------------------------
inline bool
GPlane::IsFacing(const GVector& point) const
{
	return (point * n + d >= 0);
}

//Contains returns true if the point is within the plane.
inline bool
GPlane::Contains(const GVector& point) const
{
	//P*N+d=0 for all points P on plane [N:d]
	return (IsApproxZero((point * n) + d));
}

// operators ---------------------------------------------------------
inline
const GPlane&
GPlane::operator+=(const GVector& vec)
{
	d -= vec * n;
	return *this;
}

inline
const GPlane&
GPlane::operator-=(const GVector& vec)
{
	d += vec * n;
	return *this;
}

// free operators ====================================================

} // namespace GAUGE3D

#endif //included
