/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_3DMATH_TRIANGLE_H__INCLUDED__
#define __GAUGE3D_3DMATH_TRIANGLE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------
#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#include "vector.h"
#endif //redundant include guard

// begin namespace ===================================================
namespace GAUGE3D {

class GLine;
class GPlane;

// class definition ==================================================
/*! \class GTriangle triangle.h gauge3d/3dmath/triangle.h
 *  \ingroup d3math
 *  \brief Represents a triangle in 3-space. */
class GAUGE3D_API GTriangle
{
	GVector mVertices[3];

public:
	// lifecycle -------------------------------------------------------
	GTriangle(); //!< Leaves vertices uninitialized.
	//! Initializes vertices to specified values.
	GTriangle(const GVector& vertex1, const GVector& vertex2, const GVector& vertex3);
	explicit GTriangle(GVector vertices[]);  //!< Initializes to a three-vertex array.

	// operations ------------------------------------------------------
	/*! \brief Gets the plane which the triangle is on.
	 *  \param plane Points to the location to which to write the plane.
	 *  \note The plane will face in the direction from which the vertices of the
	 *        triangle appear clockwise. */
	void GetPlane(GPlane* plane) const;
	/*! \brief Reverses the direction of the vertices.
	 *  \remarks This does not reverse the order of the vertices.  It just swaps
	 *           two of the vertices so that clockwise becomes counter-clockwise.
	 *           This also means that the direction that the triangle faces will
	 *           be reversed, since that direction is the direction from which the
	 *           vertices appear in clockwise order. */
	void Reverse();

	//! Returns true if the point is on the triangle.
	bool Contains(const GVector& point);
	
	/*! \brief Finds whether or not the line intersects the triangle, and from which
	 *         side if so.
	 * \param line The line with which to test for intersection.
	 * \returns 0 if there is no intersection, 1 if the line hits the front of the
	 *          triangle, or -1 if the line hits the back of the triangle.  (the front
	 *          side is the side from which the vertices of the triangle appear clockwise.)
	 * \remarks This function is very fast.  It is actually <i>faster</i> than Contains(),
	 *          so if you want to find the point of intersection, first call this and then
	 *          call GPlane::FindIntersection if necessary. */
	int8 Intersects(const GLine& line);
	
	// operators -------------------------------------------------------
	//! Array access to vertices.
	GVector& operator[](int index);
	//! Array access to vertices.
	const GVector& operator[](int index) const;
};

// inline member functions ===========================================
// lifecycle ---------------------------------------------------------
inline
GTriangle::GTriangle()
{
}

// operators ---------------------------------------------------------
inline GVector&
GTriangle::operator[](int index)
{
	return mVertices[index];
}

inline const GVector&
GTriangle::operator[](int index) const
{
	return mVertices[index];
}

} // namespace GAUGE3D

#endif //included
