/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__
#define __GAUGE3D_3DMATH_VECTOR_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
#ifndef __SYSTEM_MATH_H__INCLUDED__
#include <math.h>
#define __SYSTEM_MATH_H__INCLUDED__
#endif //redundant include guard

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

/*! \typedef GCoordinate
 *  \brief Represents a coordinate.
 *  \ingroup d3math
 *  This type should be used in all 3D math functions and any functions
 *  that manipulate positions in three dinemtional space. */
typedef float GCoordinate;

/*! \ingroup d3math
 *  \brief Test if a float is equal to zero with a tolerance of +-0.0001. */
inline bool IsApproxZero(float a)
{
	return (a < 0.0001) && (a > -0.0001);
}
/*! \ingroup d3math
 *  \brief Test if a double is equal to zero with a tolerance of +-0.0000001. */
inline bool IsApproxZero(double a)
{
	return (a < 0.0000001) && (a > -0.0000001);
}

// class definition ==================================================
/*! \class GVector vector.h gauge3d/3dmath/vector.h
 *  \ingroup d3math
 *  \brief A basic vector class.
 *
 *  This class represents a vector in three dimentional space.  Consult your
 *  highschool math books for more information. :)
 */
class GAUGE3D_API GVector
{
public:
	/*! \name Coordinates */
	//@{
	GCoordinate x;
	GCoordinate y;
	GCoordinate z;
	//@}

	// lifecycle -------------------------------------------------------
	/*! \name Constructors */
	//@{
	GVector(void);                                           /*!< Leaves x, y, and z uninitialized. */
	GVector(GCoordinate _x, GCoordinate _y, GCoordinate _z); /*!< Initialize to given x, y, and z. */
	/*! Initialize from 3-coordinate array containing x, y, and z. */
	explicit GVector(GCoordinate coords[]);
	//@}

	// operators -------------------------------------------------------
	/*! \name Operators
	 *  These are all standard vector math.*/
	//@{
	const GVector& operator+=(const GVector& other);    //!< Standard vector add.
	const GVector& operator-=(const GVector& other);    //!< Standard vector subtract.
	GCoordinate operator*(const GVector & other) const; //!< Dot product.
	GVector X(const GVector& other) const;              //!< Cross product.
	const GVector& operator*=(GCoordinate other);       //!< Scalar multiplication.
	GVector operator/(GCoordinate div) const;           //!< Scalar division.
	const GVector& operator/=(GCoordinate other);       //!< Scalar division.
	GCoordinate& operator[](int index);                 //!< Array-like access to coordinates.
	const GCoordinate& operator[](int index) const;     //!< Array-like access to coordinates.
	bool operator==(const GVector& other) const;        //!< Comparison with error compensation.
	bool operator!=(const GVector& other) const;        //!< Comparison with error compensation.
	//@}

	// operations ------------------------------------------------------
	/*! \name Advanced Ops*/
	//@{
	GCoordinate Magnitude(void) const;       //!< Find the magnitude (length) of vector.
	GCoordinate SquareMagnitude(void) const; //!< Find the magnitude squared (faster to calculate).
	void Normalize(void);                    //!< Make the vector into a unit vector with the same direction.
	
	//! Spherical Linear Interpolation.
	/*! Slerp() replaces the vector with the result of a spherical linear
	 *  interpolation between v1 and v2.  Imagin interpolating between two
	 *  points on the surface of a sphere.  The resulting vector will be a
	 *  unit vector.
	 */
	void Slerp(const GVector& v1, const GVector& v2, GCoordinate t);
	//@}
};

// inline functions ==================================================
// lifecycle ---------------------------------------------------------
inline
GVector::GVector(void)
{}

inline
GVector::GVector(GCoordinate _x, GCoordinate _y, GCoordinate _z)
	: x(_x),
	  y(_y),
	  z(_z)
{}

inline
GVector::GVector(GCoordinate coords[])
	: x(coords[0]),
	  y(coords[1]),
	  z(coords[2])
{}

// operations --------------------------------------------------------

inline GCoordinate
GVector::Magnitude(void) const
{
	return GCoordinate(sqrt(x * x + y * y + z * z));
}

inline GCoordinate
GVector::SquareMagnitude(void) const
{
	return x * x + y * y + z * z;
}

inline void
GVector::Normalize(void)
{
	*this /= this->Magnitude();
}

// operators ---------------------------------------------------------

inline const GVector&
GVector::operator+=(const GVector& other)
{
	x += other.x;
	y += other.y;
	z += other.z;
	return *this;
}

inline const GVector&
GVector::operator-=(const GVector& other)
{
	x -= other.x;
	y -= other.y;
	z -= other.z;
	return *this;
}

inline GCoordinate
GVector::operator*(const GVector& other) const
{
	return (x * other.x + y * other.y + z * other.z);
}

//cross product
inline GVector
GVector::X(const GVector& other) const
{
	return GVector(y * other.z - z * other.y,
	               z * other.x - x * other.z,
	               x * other.y - y * other.x);
}

inline const GVector&
GVector::operator*=(GCoordinate other)
{
	x *= other;
	y *= other;
	z *= other;
	return *this;
}

inline GVector
GVector::operator/(GCoordinate div) const
{
	GVector ret(*this);
	return ret /= div;
}

inline const GVector&
GVector::operator/=(GCoordinate other)
{
	GCoordinate s = 1 / other;
	x *= s;
	y *= s;
	z *= s;
	return *this;
}

inline GCoordinate&
GVector::operator[](int index)
{
	return (&x)[index];
}

inline const GCoordinate&
GVector::operator[](int index) const
{
	return (&x)[index];
}

inline bool
GVector::operator!=(const GVector& other) const
{
	return !(*this == other);
}

// free operators ====================================================

/*! \brief Vector addition.
 *  \ingroup d3math
 *  \relates GVector */
inline GVector
operator+(const GVector& vec1, const GVector& vec2)	//vector addition
{
	GVector ret(vec1);
	return ret += vec2;
}

/*! \brief Vector subtraction.
 *  \ingroup d3math
 *  \relates GVector */
inline GVector
operator-(const GVector& vec1, const GVector& vec2) //vector subtraction
{
	GVector ret(vec1);
	return ret -= vec2;
}

/*! \brief Vector and scalar multiplication.
 *  \ingroup d3math
 *  \relates GVector */
inline GVector
operator*(const GVector& vec, GCoordinate coord) //scalar multiplication
{
	GVector ret(vec);
	return ret *= coord;
}

/*! \brief Vector and scalar multiplication.
 *  \ingroup d3math
 *  \relates GVector */
inline GVector
operator*(GCoordinate coord, const GVector& vec) //scalar multiplication
{
	GVector ret(vec);
	return ret *= coord;
}

// global constants ==================================================
/*! \name Vector Constants */
//@{
/*! \ingroup d3math
 *  \brief Unit vector in the positive x direction. */
const GVector GVECTOR_I    = GVector(1.0, 0.0, 0.0);
/*! \ingroup d3math
 *  \brief Unit vector in the positive y direction. */
const GVector GVECTOR_J    = GVector(0.0, 1.0, 0.0);
/*! \ingroup d3math
 *  \brief Unit vector in the positive z direction. */
const GVector GVECTOR_K    = GVector(0.0, 0.0, 1.0);
/*! \ingroup d3math
 *  \brief Vector with lenght 0. */
const GVector GVECTOR_0    = GVector(0.0, 0.0, 0.0);


/*! \ingroup d3math
 *  \brief Vector pointing in the "forward" direction.
 *
 *  In GAUGE, the positive Y axis points forward. */
const GVector GVECTOR_FORWARD = GVector(0.0, 1.0, 0.0);
/*! \ingroup d3math
 *  \brief Vector pointing in the "up" direction.
 *
 *  In GAUGE, the positive X axis points right. */
const GVector GVECTOR_RIGHT   = GVector(1.0, 0.0, 0.0);
/*! \ingroup d3math
 *  \brief Vector pointing in the "right" direction.
 *
 *  In GAUGE, the positive Z axis points up. */
const GVector GVECTOR_UP      = GVector(0.0, 0.0, 1.0);
//@}

} // namespace GAUGE3D

#endif //included
