/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_COMMON_ARRAY_H__INCLUDED__
#define __GAUGE3D_COMMON_ARRAY_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

/*! \class GArray array.h gauge3d/array.h
 *  \ingroup common
 *  \brief A garbage-collected array class.
 *
 *  GArray is basically the same as GSmartPointer except that it points
 *  to an array.
 */
template <class tType>
class GArray
{
	GSmartPointer<tType> mArray;
	int mSize;
	
public:
	//! Create a null array.
	inline GArray(): mSize(0) {}
	
	//! Create an array from a pointer.
	inline GArray(GSmartPointer<tType> array, int size)
		: mArray(array), mSize(size) {}
	
	//! Allocate a new array with the given size.
	inline explicit GArray(int size)
		: mSize(size) {if(size > 0) mArray = new tType[size];}
	
	//! Random access to the array.
	inline tType& operator[](int index) const
	{
		return mArray[index];
	}
	
	//! Get the number of elements in the array.
	inline int Size() const {return mSize;}
	
	//! Get a smart pointer to the array.
	inline GSmartPointer<tType> Data() {return mArray;}
	//! Get a smart pointer to the array.
	inline const GSmartPointer<tType> Data() const {return mArray;}
	
	//! Cast to smart pointer.
	inline operator GSmartPointer<tType>() {return mArray;}
	//! Cast to smart pointer.
	inline operator const GSmartPointer<tType>() const {return mArray;}
	
	//! Set the GArray to point to a different data set.
	inline GArray& operator=(const GArray& other)
	{
		mArray = other.mArray;
		mSize  = other.mSize;
		return *this;
	}
	
	//! Check if two arrays point at the same data.
	inline bool operator==(const GArray& other) const
	{
		return mArray == other.mArray;
	}
	
	//! Use to test if array is NULL.
	inline operator void*() const
	{
		return (void*)mArray;
	}
};

/*! \class GStridedArray array.h gauge3d/array.h
 *  \ingroup common
 *  \brief Represents a strided or interleaved array.
 *
 *  This class allows you to create an array with space between
 *  the elements.  This is useful for interleaved data, where you
 *  have two types of interleaved into one big array.
 *
 *  The stride of the array is the distance in bytes between the
 *  beginning of one array element and the beginning of the next.
 *  For example, say you wanted to interleave a set of vertex
 *  positions and normals.  So, you have an array:
 *     GVector* data = new GVector[numVertices * 2];
 *  In that array, the even-numbered elements are vertex positions
 *  while the odd-numbered elements are vertex normals (starting at
 *  zero).  Then, you could create two strided arrays like so:
 *     GStridedArray<GVector> vertices(data, sizeof(GVector)*2);
 *     GStridedArray<GVector> normals(data + 1, sizeof(GVector)*2);
 *  Then, you could access "vertices" and "normals" as if they were
 *  regular arrays.
 *
 *  Note that GStridedArray is NOT reference-counted, unlike
 *  GArray.
 */
template<class tType>
class GStridedArray
{
	uint8* mPtr;
	int mStride;
	
public:
	//! Create NULL array.
	inline GStridedArray()
		: mPtr(NULL)
	{}
	
	//! Create array that starts at ptr with given stride.
	inline GStridedArray(tType* ptr, int stride)
		: mPtr((uint8*)ptr),
		  mStride(stride)
	{}
	
	//! Create array using non-strided data.
	inline explicit GStridedArray(tType* ptr)
		: mPtr((uint8*)ptr),
		  mStride(sizeof(tType))
	{}
	
	//! Array access.
	inline const tType& operator[](int index) const
	{
		return *(tType*)(mPtr + mStride * index);
	};
	//! Array access.
	inline tType& operator[](int index)
	{
		return *(tType*)(mPtr + mStride * index);
	};
	
	//! Get the stride.
	inline int Stride() {return mStride;}
	
	//! Cast to type.
	inline operator tType*() const {return (tType*)mPtr;}
	//! Test for equality with pointer.
	inline bool operator==(const void* other) {return mPtr == other;}
};

} //namespace GAUGE3D

#endif //included
