/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_COMMON_MEMORY_H__INCLUDED__
#define __GAUGE3D_COMMON_MEMORY_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
#ifndef __SYSTEM_MALLOC_H__INCLUDED__
#include <malloc.h>
#endif

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

GAUGE3D_API void* GMalloc(int size);
GAUGE3D_API void GFree(void* ptr);

struct GSmartPointerBase
{
	//intentionally left empty.
};

/*! \class GSmartPointer object.h gauge3d/object.h
 *  \ingroup common
 *  \brief A reference-counting smart pointer class.
 *
 *  GSmartPointer is a reference counting smart pointer class.  It can
 *  point to any class derived from GObject.  For each smart pointer
 *  pointing to an object, the object's reference count will be incremented.
 *  When the smart pointer is destroyed, the reference count will be
 *  decremented.  When the count hits zero, the object will be deleted.
 *
 *  By using smart pointers, you can forget about who has ownership of an
 *  object and should therefore delete it when done.  The object will be
 *  deleted automatically when it is done with.
 *
 *  In GAUGE, every class that is derived from GObject has an associated typedef'd
 *  smartpointer type.  The smartpointer type is named after the object with the
 *  prefix 'p'.  So, pGObject is a smart pointer to a GObject, and pGFile is a smart
 *  pointer to a GFile, etc.
 *
 *  Some important things to remember when using smart pointers:
 *  - Avoid circular links.  This occurs when two objects contain smart
 *    pointers to each other.  In this case, the objects will not be removed
 *    automatically as they will always have reference counts of one or more.
 *  - Smart pointers are meant to be passed by <i>value</i>, not by address
 *    or reference.  Furthermore, smart pointers should always be created
 *    either on the stack or as members of larger structures rather than be
 *    created wich new.  In general, you should never find yourself dealing
 *    with either a pointer or a reference to a smart pointer unless you know
 *    what you are doing.
 */
template<class tType>
class GAUGE3D_API GSmartPointer: public GSmartPointerBase
{
	tType* mObject;
	
	inline void Reference()
	{
		//the reference count is stored in an integer just before the object
		// pointed to.
		if(mObject)
			++(*((int*)mObject - 1));
	}
	
	inline void UnReference()
	{
		//the reference count is stored in an integer just before the object
		// pointed to.
		if(mObject)
			if(--(*((int*)mObject - 1)) == 0)
				delete mObject;
	};
	
	inline void SetPointer(tType* ptr)
	{
		UnReference();
		mObject = ptr;
		Reference();
	};
	
public:
	// lifecycle ------------------------------------------------------
	/*! \name Constructors and Destructors */
	//@{
	//! Sets pointer to NULL.
	inline GSmartPointer()
		: mObject(NULL)
	{
	}
	
	/*! \brief Cast from regular pointer.
	 *
	 *  WARNING:  This can be dangerous.  If you have both regular pointers
	 *  and smart pointers pointing to the same object, and all the smart
	 *  pointers are destroyed, the regular pointers will be left dangling,
	 *  and could cause a program crash.  Try to avoid using both regular and
	 *  smart pointers to an object. */
	inline GSmartPointer(tType* object)
		: mObject(object)
	{
		Reference();
	}
	
	//! Copy constructor
	inline GSmartPointer(const GSmartPointer& other)
		: mObject(other.mObject)
	{
		Reference();
	}
	
	/*! \brief Cast constructor from smart pointers for other types.
	 *
	 *  Note that type checking is not performed.  If you cast to a type
	 *  which the object isn't, you will not be warned.  If you need to
	 *  do dynamic casting, you must do it manually.
	 */
	inline GSmartPointer(const GSmartPointerBase& other)
		: mObject(((GSmartPointer*)&other)->mObject)
	{
		Reference();
	}
	
	//! Decrement reference count and delete if zero.
	~GSmartPointer()
	{
		UnReference();
	}
	//@}
	
	// operators ------------------------------------------------------
	/*! \name Assignment Operators
	 *  \brief These work like you'd expect them to. */
	//@{
	const GSmartPointer& operator=(const GSmartPointer& other)
	{
		SetPointer(other.mObject);
		return *this;
	}

	/*  The same warnings as for the corresponding constructor apply. */
	const GSmartPointer& operator=(const GSmartPointerBase& other)
	{
		SetPointer(((GSmartPointer*)&other)->mObject);
		return *this;
	}

	/*  The same warnings as for the corresponding constructor apply. */
	const GSmartPointer& operator=(tType* other)
	{
		SetPointer(other);
		return *this;
	}
	//@}
	
	/*! \name Equality Tests
	 *  \brief These work like you'd expect them to. */
	//@{
	inline bool operator==(const GSmartPointer& other) const
	{
		return mObject == other.mObject;
	}
	inline bool operator==(const GSmartPointerBase& other) const
	{
		return mObject == ((GSmartPointer*)&other)->mObject;
	}
	inline bool operator==(const tType* other) const
	{
		return mObject == other;
	}
	//@}
	
	/*! \name Dereferencing
	 *  \brief These work like you'd expect them to. */
	//@{
	
	//duh
	inline tType& operator*()
	{
		return *mObject;
	}
	inline const tType& operator*() const
	{
		return *mObject;
	}
	
	//operator-> to make it like a real pointer!
	inline tType* operator->()
	{
		return mObject;
	}
	inline const tType* operator->() const
	{
		return mObject;
	}
	//@}
	
	//! Array access for smart pointers that point to arrays.
	inline tType& operator[](int index) const
	{
		return mObject[index];
	}
	
	//! Cast to standard pointer.
	inline operator tType*() const
	{
		return mObject;
	}
};

} //namespace GAUGE3D

#ifdef PLATFORM_MSVC

//MSVC doesn't know what bad_alloc is!  Stupid!

//well, we never actually throw it anyway, so...

inline void* operator new(size_t sz)
	{return GAUGE3D::GMalloc(sz);}
inline void  operator delete(void* addr)
	{GAUGE3D::GFree(addr);}

inline void* operator new[](size_t sz)
	{return GAUGE3D::GMalloc(sz);}
inline void  operator delete[](void* addr)
	{GAUGE3D::GFree(addr);}

#else

#include <new>

inline void* operator new(size_t sz) throw(bad_alloc)
	{return GAUGE3D::GMalloc(sz);}
inline void  operator delete(void* addr) throw()
	{GAUGE3D::GFree(addr);}

inline void* operator new[](size_t sz) throw(bad_alloc)
	{return GAUGE3D::GMalloc(sz);}
inline void  operator delete[](void* addr) throw()
	{GAUGE3D::GFree(addr);}

#endif

#endif //included
