/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_COMMON_OBJECT_H__INCLUDED__
#define __GAUGE3D_COMMON_OBJECT_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
#ifndef __SYSTEM_ASSERT_H__INCLUDED__
#include <assert.h>
#define __SYSTEM_ASSERT_H__INCLUDED__
#endif

// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \name Support macros for QueryInterface. */
//@{

/*! \brief Set the UID for a class.
 *  \ingroup common
 *
 *  GAUGE3D_SET_UID should be used in the definition for any class
 *  that is derived from GObject.  The numbers are the high, middle,
 *  and low 32-bit integers in the 96-bit GAUGE unique identifier for
 *  the class.  (use gauge3d-uidgen to create them) */
#define GAUGE3D_SET_UID(HIGH, MID, LOW)                                  \
public:                                                                  \
	enum {_UID_HIGH = HIGH, _UID_MID = MID, _UID_LOW = LOW};              \
	virtual pGObject _QueryInterface(uint32 high, uint32 mid, uint32 low);\
private:

/*! \brief Begin interface declaration block for a class.
 *  \ingroup common
 *
 *  GAUGE3D_DECLARE_INTERFACES should appear in the cpp file for any
 *  class that uses GAUGE3D_SET_UID.  It should be followed by a
 *  GAUGE3D_DECLARE_BASECLASS for each base class (but not base class's
 *  base classes, etc.) and a GAUGE3D_DECLARE_EMBEDDED_INTERFACE for every
 *  embedded interface.  These should preferably be ordered with the
 *  most likely requested interface on top.  After all of these,
 *  GAUGE3D_END_INTERFACES should be used. */
#define GAUGE3D_DECLARE_INTERFACES(CLASSNAME)                           \
pGObject CLASSNAME::_QueryInterface(uint32 high, uint32 mid, uint32 low)\
{                                                                       \
	pGObject ret;                                                        \
	if((high == _UID_HIGH) && (mid == _UID_MID) && (low  == _UID_LOW))   \
		return this;

/*! \brief Declare inherited interfaces for a class.
 *  \ingroup common
 *
 *  GAUGE3D_DECLARE_BASECLASS declares direct base classes.  Not
 *  the base classes of base classes, just direct inherited classes.
 *  So, if you are not using multiple inheritance, you should never
 *  need more than one of these. */
#define GAUGE3D_DECLARE_BASECLASS(BASECLASSNAME)                     \
	if((ret = BASECLASSNAME::_QueryInterface(high, mid, low)) != NULL)\
		return ret;

/*! \brief Declare embedded interfaces for a class.
 *  \ingroup common
 *
 *  GAUGE3D_DECLARE_EMBEDDED_INTERFACE declares interfaces from which this
 *  class is not inherited, but which define additional functionality that
 *  this class has.  The parameter should be an actual instantiated copy of
 *  the interface, not the classname of it.  This is because that
 *  interface's QueryInterface function needs to be called. */
#define GAUGE3D_DECLARE_EMBEDDED_INTERFACE(INTERFACE_OBJECT)             \
	if((ret = (INTERFACE_OBJECT)._QueryInterface(high, mid, low)) != NULL)\
		return ret;

/*! \brief End interface declaration block.
 *  \ingroup common
 *
 *  Place GAUGE3D_END_INTERFACES after all you interface declarations. */
#define GAUGE3D_END_INTERFACES                                \
	return ret;                                                \
}

//@}

//And here we have GObject

class GObject;

/*! \typedef pGObject
 *  \ingroup common
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GObject GObject\endlink. */
typedef GSmartPointer<GObject> pGObject;

/*! \class GObject stddefs.h gauge3d/stddefs.h
 * \ingroup common
 * \brief Base class objects with smart pointer and QueryInterface capabilities.
 *
 * GObject is the base class for many, but not all, GAUGE classes.  Using it adds
 * several capabilities to the derived class:
 * - GSmartPointer can point to GObject and classes derived from it.
 * - The QueryInterface system provides COM-like interface building capabilities.
 * - Per-class error reporting system.
 */
class GAUGE3D_API GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781e0b);
public:
	virtual ~GObject();
	
	/*! \brief An error identifier.
	 *
	 *  tError is an error identifier.  In actuality, it is a const char*
	 *  which points to a human-readable string.  Every class which has
	 *  error conditions should define these conditions as member variables
	 *  of type static const tError.  Then, to check which error occured, the
	 *  caller can check the error identifier against each of these member
	 *  variables.  If it doesn't know how to handle the error, it can just
	 *  output it to cerr or something since the error identifier points to
	 *  a human-readable string. */
	typedef const char* tError;
	
private:
	tError mError;
protected:
	//! Set the last error.
	inline void Error(tError error) {mError = error;}
public:
	//! Get the last error.
	inline tError Error()
		{return mError;}
	
	/*! \brief Called by GScheduler every frame if the object requests it.
	 *  \param time      The current time, in seconds.  This is measured relative
	 *                   to a completely arbitrary point (usually the point when
	 *                   the engine was initialized).
	 *  \param frameTime The amount of time which elapsed between the start of the
	 *                   previous frame and the start of this frame.
	 *
	 *  See GScheduler for more info on how to make it call this function every
	 *  frame.  Objects that don't need this functionality can ignore it. */
	virtual void DoFrame(tTime time, tTime frameTime);
	
	/*! \brief Called by GScheduler at a desired time.
	 *  \param time The current time.  This will usually be a bit later than the
	 *              time at which the object requested to be called back, but it
	 *              will never be earlier.
	 *
	 *  See GScheduler for more info on how to make it call this function at a
	 *  particular point in time.  Objects that don't need this functionality can
	 *  ignore it. */
	virtual void TimerCallback(tTime time);
	
#if 0
	//MSVC can't handle this template.  Who knows why?  :(
	template <class tType>
	inline GSmartPointer<tType> QueryInterface()
	{
		return (GSmartPointer<tType>)_QueryInterface(tType::_UID_HIGH, tType::_UID_MID, tType::_UID_LOW);
	}
#endif
};

/*! \fn GSmartPointer<tType> QueryInterface(pGObject object)
 *  \ingroup common
 *  \brief See if an object supports a particular interface.
 *  \param tType The type of the interface you want.
 *  \param object A smart pointer to the object from which you are requesting
 *                the interface.
 *  \return A smart pointer to the interface you requested, or NULL if the object
 *          does not support that interface.
 *
 *  This function determines if the given object supports an interface
 *  of type tType.  The returned interface may be either an inherited
 *  interface or an embedded interface.
 *
 *  For example, if you had an object of type GObject and you wanted to
 *  see if it was a GFile, you would do this:  QueryInterface<GFile>(myObject);
 *
 *  This function would be a member of GObject, but MSVC doesn't like templated
 *  member functions.
 */
template <class tType>
inline GSmartPointer<tType> QueryInterface(pGObject object)
{
	return object->_QueryInterface(tType::_UID_HIGH, tType::_UID_MID, tType::_UID_LOW);
}

} //namespace GAUGE3D

#endif //included
