/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_ENGINE_H__INCLUDED__
#define __GAUGE3D_ENGINE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_ARRAY_H__INCLUDED__
#include <gauge3d/common/array.h>
#endif

// local package -----------------------------------------------------
#ifndef __GAUGE3D_COMMON_STRING_H__INCLUDED__
#include <gauge3d/common/gstring.h>
#endif

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GEngine engine.h gauge3d/engine.h
 *  \ingroup control
 *  \brief Used to initialize the engine.
 *
 *  All the functions in this class are static. */
class GAUGE3D_API GEngine
{
	// insulated state -------------------------------------------------
	struct Private;
	
public:
	//! Bits for subsystems.
	enum tSubSystems
	{
		INPUT          = 1 << 0,      //!< Input subsystem (message handling).
		VIDEO          = 1 << 1,      //!< Video and 3D rendering subsystem.
		FILESYSTEM     = 1 << 2,      //!< File I/O and data stream subsystem.
		OBJECT_LOADER  = 1 << 3,      //!< Object/Plugin loader subsystem.
		ALL_SUBSYSTEMS = 0x7FFFFFFF   //!< All subsystems.
	};
	
	// operations ------------------------------------------------------
	//! Call this when your program starts to tell GAUGE the command-line arguments used.
	static void CommandLine(int argc, char* argv[]);
	//! Call this when your program starts to tell GAUGE the command-line arguments used.
	static void CommandLine(char* args);
	//! Call this when your program starts to tell GAUGE the command-line arguments used.
	static void CommandLine(int argc, GString argv[]);
	//! Call this when your program starts to tell GAUGE the command-line arguments used.
	static void CommandLine(GArray<GString> args);
	//! Get the command-line arguments.
	static GArray<GString> CommandLine();
	
	/*! \breif Set the program's name.
	 *  \parameter name This should be a short name for the program consisting of only
	 *                  lower-case letters, numerals, hyphens, and underscores.  Depending
	 *                  on the OS, this name may be used for file and directory names.
	 *                  For example, on unix-like OS's, the user's config directory will
	 *                  be "~/.progname", where "progname" is the name you specify. */
	static void ProgName(GString name);
	//! Get the program's name.
	static GString ProgName();
	
	/*! \breif Initialize the engine and one or more sub-systems.
	 *  \param subsystems A bitwise OR of one or more of the values of \ref tSubSystems.
	 *
	 * You may call Init() multiple times, or initialize everything at once. */
	static void Init(int subsystems);
	//! Get the initialized subsystems.
	static int  InitializedSubSystems();
	//! Shutdown one or more subsystems.
	static void ShutdownSubSystem(int subsystems);
	//! Shutdown the engine and all active subsystems.
	static void Shutdown();
	
	/*! \brief Execute main loop.
	 *
	 * While running the main loop, GAUGE will repeatedly call the DoFrame() functions of
	 * each subsystem's scheduler.  See \ref GScheduler for more. */
	static void MainLoop();
	/*! \brief Stop the main loop.
	 *
	 * MainLoop() will return on the frame after Quit() is called. */
	static void Quit();
};

} //namespace GAUGE3D

#endif //included
