/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_GRAPHICS_IMAGE_H__INCLUDED__
#define __GAUGE3D_GRAPHICS_IMAGE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_ARRAY_H__INCLUDED__
#include <gauge3d/common/array.h>
#endif

#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include <gauge3d/common/gstring.h>
#endif

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GImage image.h gauge3d/graphics/image.h
 *  \ingroup graphics
 *  \brief Base class for image plugins.
 *
 * GImage describes simple images in an abstract way.  Plugins that
 * support specific image formats (like GIF or JPEG) are derived from
 * this. */
class GAUGE3D_API GImage: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781e6f);
public:
	virtual ~GImage();
	
	/*! \struct FormatDesc image.h gauge3d/graphics/image.h
	 *  \brief GImage::FormatDesc describes the format of raw image data obtained from a GImage.
	 *
	 *  mNumChannels stores the number of channels in the image.  This should be a number
	 *  between 1 and 4, inclusive.  A typical RGB image has three channels (red, green, and
	 *  blue) while RGBA has four and grayscale has one.
	 *
	 *  mBitsPerChannel can 4 (for 16-bit images), 8 (for 32-bit images), or 5.  In the case
	 *  of 5, the image is 16-bits-per-pixel, and must contain three or four channels.  If
	 *  it contains three channels, the last bit of each 16 is ignored.  If it contains four
	 *  channels, the last bit of each 16 is interpreted as the fourth channel (a one-bit
	 *  channel). */
	struct FormatDesc
	{
		int  mWidth;          //!< Width of the image, in pixels.
		int  mHeight;         //!< Height of the image, in pixels.
		int  mNumChannels;    //!< Number of color channels in the image.
		int  mBitsPerChannel; //!< Number of bits per element per channel.
		bool mHasMipmaps;     //!< True if the image has built-in mipmaps.
		GString mEncoding;    //!< The encoding/compression used to store the image. (JPEG, FXT1, etc.)
	};
	
	//! Get the image format.
	virtual FormatDesc Format() = 0;
	
	/*! \brief Gets the raw image data in the format given by GetFormat().
	 *  \param mipmapLevel The mip map to return.  For example, for a 64x64 image,
	 *                     level 0 is 64x64, 1 is 32x32, 2 is 16x16, etc.  Images
	 *                     that don't provide mip maps should ignore this.
	 *  \returns The raw image data in the given format. */
	virtual GArray<uint8> Data(int mipmapLevel = 0) = 0;
	
	/*! \brief Gets the encoded image data.
	 *  \param mipmapLevel The mip map to return.  For example, for a 64x64 image,
	 *                     level 0 is 64x64, 1 is 32x32, 2 is 16x16, etc.  Images
	 *                     that don't provide mip maps and images encodings whing
	 *                     encode multiple mipmap levels together should ignore this.
	 *  \returns The image, encoded in whatever format the image file was in.
	 *           So, it may be jpeg compressed, or FXT1, or whatever. */
	virtual GArray<uint8> EncodedData(int mipmapLevel = 0) = 0;
};
/*! \typedef pGImage
 *  \ingroup graphics
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GImage GImage\endlink. */
typedef GSmartPointer<GImage> pGImage;

} //namespace GAUGE3D

#endif //included
