/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_INPUT_MESSAGE_H__INCLUDED__
#define __GAUGE3D_INPUT_MESSAGE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include <gauge3d/common/gstring.h>
#endif //redundant include guard

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GMessage message.h gauge3d/input/message.h
 *  \ingroup input
 *  \brief Represents a message of some sort.
 *
 *  GMessage itself does not contain any functions for finding out what
 *  sort of message it is.  You must use QueryInterface to determine the
 *  type of the message. */
class GAUGE3D_API GMessage: public GObject
{
	pGObject mOrigin;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39b2ecf1);
public:
	/*! \brief Create a GMessage with the given origin.
	 *  \param origin A pointer to the object which is the logical "originator"
	 *                of the message.  For example, most messages will originate
	 *                from a GDisplay. */
	inline GMessage(pGObject origin)
		: mOrigin(origin) {}
	//! Destructor.
	virtual ~GMessage();
	
	//! Returns a pointer to the object that this message came from.
	inline pGObject Origin() {return mOrigin;}
};
/*! \typedef pGMessage
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GMessage GMessage\endlink. */
typedef GSmartPointer<GMessage> pGMessage;

/*! \class GQuitMessage message.h gauge3d/input/message.h
 *  \ingroup input
 *  \brief Represents a message asking the program to quit.
 *
 *  This message is sent when the user requests that the program exit.
 *  It is up to the receiver of the message to actually shut down the
 *  program when it receives this message.  So, for example, if the user
 *  tries to close the program window, the renderer will send a quit
 *  message, but the window will not actually be closed until some other
 *  part of the engine asks the renderer to close it.  So, if the quit
 *  message is ignored, the program will stay open. */
class GAUGE3D_API GQuitMessage: public GMessage
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39b2eefb);
public:
	/*! \brief Create a GQuitMessage with the given origin.
	 *  \param origin A pointer to the object which is the logical "originator"
	 *                of the message.  For example, most messages will originate
	 *                from a GDisplay. */
	inline GQuitMessage(pGObject origin)
		: GMessage(origin) {}
	//! Destructor.
	virtual ~GQuitMessage();
	
	//Purposely left empty.  This class's type is all that it needs.
};
/*! \typedef pGQuitMessage
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GQuitMessage GQuitMessage\endlink. */
typedef GSmartPointer<GQuitMessage> pGQuitMessage;

/*! \class GInputMessage message.h gauge3d/input/message.h
 *  \ingroup input
 *  \brief Represents an input message.
 *
 *  Input messages can be analog or digital, and they can be relative
 *  or absolute.  Here's what each combination means:
 *  - <b>Absolute Analog</b>:  The value of the input is always between -1.0
 *    and 1.0, and is measured relative to some central point.  For example,
 *    joystick positions are normally measured this way.  0.0 is the center of
 *    the joystick, and anything else is off to one side or the other.
 *  - <b>Relative Analog</b>:  The value of the input could be anything.  It
 *    is measured relative to the last position of the input rather than relative
 *    to some reference position.  Mice are typically relative analog devices.
 *    Each mouse movement message only reports the distance that the mouse moved
 *    since the last message.
 *  - <b>Absolute Digital</b>:  The value of the input is either 0.0 or 1.0.
 *    This is mostly used for buttons, including mouse buttons and keys on the
 *    keyboard.  A value of 1.0 means the button was pressed, whereas 0.0 means
 *    it was released.
 *  - <b>Relative Digital</b>:  The value of the input is either -1.0 or 1.0.
 *    The main use for this is mouse wheels, since they normally are measured in
 *    some sort of clicking fashion.  1.0 means the wheel was clicked up one notch
 *    and -1.0 means it was clicked down one notch. */
class GAUGE3D_API GInputMessage: public GMessage
{
	GString mName;
	float mValue;
	int mFlags;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39b19a2d);
public:
	//! Indicates the type of input.
	enum tFlags
	{
		DIGITAL = 0,
		ANALOG = 1,
		ABSOLUTE = 0,
		RELATIVE = 2
	};
	
	/*! \brief Create a GInputMessage with the given info.
	 *  \param origin   A pointer to the object which is the logical "originator"
	 *                  of the message.  For example, most messages will originate
	 *                  from a GDisplay.
	 *  \param name     See \ref Name(), below.
	 *  \param value    See \ref Value(), below.
	 *  \param flags    A bitwise combination of the enum values of tFlags.  The
	 *                  input must be either ANALOG or DIGITAL, and either ABSOLUTE
	 *                  or RELATIVE. */
	inline GInputMessage(pGObject origin, GString name, float value, int flags)
		: GMessage(origin),
		  mName(name),
		  mValue(value),
		  mFlags(flags)
	{}
	virtual ~GInputMessage();
	
	//! Returns true if the input is analog, false if digital.
	inline bool IsAnalog() {return mFlags & ANALOG;}
	//! Returns true if the input is relate, false if absolute.
	inline bool IsRelative() {return mFlags & RELATIVE;}
	/*! \brief Returns the raw bitpattern type of the message.
	 *  \remarks This does not give you any more info than IsAnalog and IsRelative do. */
	inline int Type() {return mFlags;}
	
	//! Returns the value of the input.
	inline float Value() {return mValue;}
	
	/*! \brief Returns an identifier for the input.
	 *
	 *  The name of the input source is a human-readable unique identifier for
	 *  the source of the input.  In the case of a key on the keyboard, the
	 *  name might be "a" or "left control".  For a mouse button, it would be
	 *  "mouse1" or "mouse3".  For mouse movement, "mouse x" or "mouse y".  Et
	 *  cetera. */
	inline GString Name() {return mName;}
};
/*! \typedef pGInputMessage
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GInputMessage GInputMessage\endlink. */
typedef GSmartPointer<GInputMessage> pGInputMessage;

/*! \class GKeyboardMessage message.h gauge3d/input/message.h
 *  \ingroup input
 *  \brief Message sent when a key is pressed or released.
 *
 *  This class is here for two reasons:
 *   - To make it easier to create common message types.
 *   - To make it possible to determine the source of a message.
 *     (Hint:  Use \ref QueryInterface) */
class GAUGE3D_API GKeyboardMessage: public GInputMessage
{
	GString mModified;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39b2fe95);
public:
	/*! \brief Create a keyboard message with the given info.
	 *  \param origin   A pointer to the object which is the logical "originator"
	 *                  of the message.  For example, most messages will originate
	 *                  from a GDisplay.
	 *  \param key      The name of the key that was pressed or released.
	 *                  If the key has a direct ASCII translation, this
	 *                  should be it.  Otherwise, it should be a short
	 *                  but unique name for the key (like "left control").
	 *  \param modified This is the key with modifier keys (like shift or
	 *                  caps lock) applied.  So, if shift is held down and
	 *                  the "a" key is pressed, <i>modified</i> should be
	 *                  "A", whereas <i>key</i> should be "a".  If the key is
	 *                  not printable, this will be an empty sting.
	 *  \param down     True if the key was pressed, false if released.
	 *
	 *  \remarks Please, please, if you are writing a plugin which reads
	 *           input from a keyboard and sends GKeyboardMessages, please
	 *           make sure to check the system keyboard map!  Do NOT assume
	 *           querty!  I have a dvorak keyboard, so I will not use your
	 *           software if you disregard this. */
	inline GKeyboardMessage(pGObject origin, GString key, GString modified, bool down)
		: GInputMessage(origin, key, down?1.0:0.0, DIGITAL | ABSOLUTE),
		  mModified(modified)
	{}
	//! Destructor.
	virtual ~GKeyboardMessage();
	
	//! Get the key name with modifiers applied.
	inline GString ModifiedKey() {return mModified;}
};
/*! \typedef pGKeyboardMessage
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GKeyboardMessage GKeyboardMessage\endlink. */
typedef GSmartPointer<GKeyboardMessage> pGKeyboardMessage;

/*! \class GMouseButtonMessage message.h gauge3d/input/message.h
 *  \ingroup input
 *  \brief Message sent when a mouse button is pressed or released.
 *
 *  This class is here for two reasons:
 *   - To make it easier to create common message types.
 *   - To make it possible to determine the source of a message.
 *     (Hint:  Use \ref QueryInterface) */
class GAUGE3D_API GMouseButtonMessage: public GInputMessage
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39b30137);
public:
	/*! \brief Create a mouse button message with the given info.
	 *  \param origin A pointer to the object which is the logical "originator"
	 *                of the message.  For example, most messages will originate
	 *                from a GDisplay.
	 *  \param button The name of the button pressed.  Something like
	 *                "left mouse".
	 *  \param down   True if the button was pressed, false if released. */
	inline GMouseButtonMessage(pGObject origin, GString button, bool down)
		: GInputMessage(origin, button, down?1.0:0.0, DIGITAL | ABSOLUTE)
	{}
	virtual ~GMouseButtonMessage();
};
/*! \typedef pGMouseButtonMessage
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GMouseButtonMessage GMouseButtonMessage\endlink. */
typedef GSmartPointer<GMouseButtonMessage> pGMouseButtonMessage;

/*! \class GMouseWheelMessage message.h gauge3d/input/message.h
 *  \ingroup input
 *  \brief Message sent when the mouse wheel is moved.
 *
 *  This class is here for two reasons:
 *   - To make it easier to create common message types.
 *   - To make it possible to determine the source of a message.
 *     (Hint:  Use \ref QueryInterface) */
class GAUGE3D_API GMouseWheelMessage: public GInputMessage
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39d7c604);
public:
	enum tDirection
	{
		UP,
		DOWN
	};
	
	/*! \brief Create a mouse wheel message with the given info.
	 *  \param origin A pointer to the object which is the logical "originator"
	 *                of the message.  For example, most messages will originate
	 *                from a GDisplay.
	 *  \param name   The name of the input.  Just "mouse wheel" should do in most
	 *                cases.
	 *  \param direction Either UP or DOWN, indicating the direction the mouse wheel
	 *                was moved. */
	inline GMouseWheelMessage(pGObject origin, GString name, tDirection direction)
		: GInputMessage(origin, name, (direction == UP)?1.0:-1.0, DIGITAL | RELATIVE)
	{}
	virtual ~GMouseWheelMessage();
};
/*! \typedef pGMouseWheelMessage
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GMouseWheelMessage GMouseWheelMessage\endlink. */
typedef GSmartPointer<GMouseWheelMessage> pGMouseWheelMessage;

/*! \class GMouseMotionMessage message.h gauge3d/input/message.h
 *  \ingroup input
 *  \brief Message sent when the mouse is moved.
 *
 *  This class is here for two reasons:
 *   - To make it easier to create common message types.
 *   - To make it possible to determine the source of a message.
 *     (Hint:  Use \ref QueryInterface) */
class GAUGE3D_API GMouseMotionMessage: public GInputMessage
{
	float mAbsolute;
	
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39b30254);
public:
	/*! \brief Create a mouse motion message with the given info.
	 *  \param origin   A pointer to the object which is the logical "originator"
	 *                  of the message.  For example, most messages will originate
	 *                  from a GDisplay.
	 *  \param button   The name of the axis on which the motion occurred.
	 *                  Something like "mouse x".
	 *  \param change   The distance that the mouse moved.  This should be measured
	 *                  in the same units as <i>absolute</i>.
	 *  \param absolute If the mouse is <i>not</i> captured by the display, then
	 *                  this is the absolute position of the mouse.  The edges of
	 *                  the screen are 0.0 and 1.0.
	 *  \remarks Yes, mouse motion will usually have to be sent as two
	 *           separate messages (one for x and one for y). */
	inline GMouseMotionMessage(pGObject origin, GString axis, float change, float absolute)
		: GInputMessage(origin, axis, change, ANALOG | RELATIVE),
		  mAbsolute(absolute)
	{}
	virtual ~GMouseMotionMessage();
	
	/*! \brief Returns the absolute position of the mouse.
	 *  \remarks This will be zero if the mouse is in captured mode. */
	inline float Absolute() {return mAbsolute;}
};
/*! \typedef pGMouseMotionMessage
 *  \ingroup input
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GMouseMotionMessage GMouseMotionMessage\endlink. */
typedef GSmartPointer<GMouseMotionMessage> pGMouseMotionMessage;

} //namespace GAUGE3D

#endif //included
