/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_OSABSTRACTION_FILESYSTEM_H__INCLUDED__
#define __GAUGE3D_OSABSTRACTION_FILESYSTEM_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include <gauge3d/common/gstring.h>
#endif //redundant include guard

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GFileSystem filesystem.h gauge3d/osabstraction/filesystem.h
 *  \ingroup osabstraction
 *  \brief Low-level filesystem access.
 *
 *  Do not use this class directly.  Use the classes in the File Manipulation
 *  module instead.
 *
 *  This is a static class.  All members are static.  You do not need to
 *  instantiate it.  So, if you want to open a file, you would say something
 *  like GFileSystem::Open("myfile.txt", GFileSystem::READ). */
class GAUGE3D_API GFileSystem
{
	// insulated state -------------------------------------------------
	struct Private;
	
public:
	// lifecycle -------------------------------------------------------
	
	/*! \name Initialization */
	//@{
	
	/*! \brief Initialize the filesystem.
	 *  \param progName A short name for the program, containing only numbers and
	 *                  lower-case letters with no spaces.  For example, "gauge3d".
	 *                  This name will be used for directories and such that belong
	 *                  to this program.  (i.e. on Unix, the config files are stored
	 *                  in "~/." + progName.)
	 *
	 *  Specifically, this function does the following things:
	 *   - Creates the config directory if it isn't there yet.
	 *   - Deletes the temporary directory if it is there, and creates a new one.
	 *   - Sticks a text file in the temporary directory telling the user that they
	 *     can delete it if it is there. */
	static void Init(GString progName);
	/*! \brief Shuts down and cleans up the filesystem.
	 *
	 *  Specifically, this function deletes the temporary directory and all of
	 *  its contents. */
	static void ShutDown();
	
	//@}
	
	// operations ------------------------------------------------------
	
	/*! \name Directory Structure */
	//@{
	
	//! Get the currend working directory.
	static GString WorkingDir();
	//! Get the directory where config files should be placed.
	static GString ConfigDir();
	//! Get the location of the program's resources.
	static GString ResourceDir();
	//! Get the location of GAUGE plugins.
	static GString PluginDir();
	
	//@}
	
	struct File;
	
	//! Flags used in the Open() function.
	enum tOpenFlags
	{
		READ      = 1 << 0,  //!< Open for reading.
		WRITE     = 1 << 1,  //!< Open for writing.
		CREATE    = 1 << 2,  //!< Create the file if it doesn't exist.
		NOREPLACE = 1 << 3,  //!< Don't open if the file already exists.
		TRUNCATE  = 1 << 4,  //!< Truncate the file to zero length if it exists.
		APPEND    = 1 << 5   //!< Place the stream at the end of the file.
	};
	//! Flags for the Seek() function.
	enum tSeekType
	{
		BEGINNING,           //!< Seek relative to the beginning of the stream.
		CURRENT,             //!< Seek relative to the current stream position.
		END                  //!< Seek relative to the end of the stream.
	};
	
	/*! \name  Stream-based File Access */
	//@{
	/*! \brief Open a file.
	 *  \param filename The path and name of the file.
	 *  \param flags    One or more of the flags from \ref tOpenFlags, bitwise or'd.
	 *  \returns        A pointer to the open file, or NULL if an error occured.
	 */
	static File* Open(GString filename, int flags);
	/*! \brief Read from a file.
	 *  \param file   A pointer to a file, obtained by calling Open().
	 *  \param target The location to fill with the data read from the file.
	 *  \param size   The number of bytes to read.
	 *  \returns      The number of bytes read. */
	static int   Read(File* file, void* target, int size);
	/*! \brief Write to a file.
	 *  \param file   A pointer to a file, obtained by calling Open().
	 *  \param source The location of the data to write to the file.
	 *  \param size   The number of bytes to write.
	 *  \returns      The number of bytes written. */
	static int   Write(File* file, const void* source, int size);
	/*! \brief Seek the position of the stream.
	 *  \param file       A pointer to a file, obtained by calling Open().
	 *  \param offset     The position to seek to.
	 *  \param relativeTo The position that offset is relative to.  One of
	 *                    the values of \ref tSeekType. */
	static void  Seek(File* file, int offset, tSeekType relativeTo);
	/*! \brief Get the position of the stream.
	 *  \param file A pointer to a file, obtained by calling Open().
	 *  \returns    The position of the stream relative to the beginning of the file. */
	static int   TellPos(File* file);
	/*! \brief Close the file.
	 *  \param file A pointer to a file, obtained by calling Open().
	 *
	 *  Call this when you are done with the file. */
	static void  Close(File* file);
	//@}
	
	struct FileMapInternal;
	
	/*! \class FileMap filesystem.h gauge3d/osabstraction/filesystem.h
	 *  \brief Contains information related to a memory mapped file.
	 *
	 *  The values of mOffset and mSize will always be exactly what was requested
	 *  when GFileSystem::Map() was called. */
	struct FileMap
	{
		void* mMap;       //!< The location in memory where the file has been mapped.
		int   mOffset;    //!< The offset from the beginning of the file at which the map starts.
		int   mSize;      //!< The size of the map.
		FileMapInternal* mInternal; //!< A pointer to internal info needed by the filesystem.
	};
	
	/*! \name Memmory Mapped File Access */
	//@{
	
	/*! \brief Map a file to memory.
	 *  \param filename The path and name of the file to map.
	 *  \param offset   The offset in bytes from the beginning of the file at which the
	 *                  map should start.
	 *  \param size     The number of bytes of the file to map.
	 *  \param flags    One or more of the flags \ref READ, \ref WRITE, \ref CREATE, and
	 *                  \ref NOREPLACE, from \ref tOpenFlags, bitwise or'd.  These have the
	 *                  same effect here as they do for Open().
	 *  \returns        A FileMap structure describing the map in memory, or NULL if an error
	 *                  occured.
	 *
	 *  This function allows you to access a file as if the entire file (or a segment of it)
	 *  were in RAM.  Where available, features of the operating system are used to accellerate
	 *  this functionality. */
	static FileMap* Map(GString filename, int offset, int size, int flags);
	/*! \brief Close a memory mapped file.
	 *  \param map A pointer to a FileMap, obtained by calling Map()
	 *
	 *  Call this when you are done with the file map. */
	static void UnMap(FileMap* map);
	
	//@}
	
	/*! \name Other File Manipulation Functions */
	//@{
	
	/*! \brief Get the size of a file.
	 *  \param filename The path and name of the file to get the size of.
	 *  \returns        The size of the file in bytes. */
	static int FileSize(GString filename);
	/*! \brief Get the size of an already open file.
	 *  \param file     A pointer to a file, obtained by calling Open().
	 *  \returns        The size of the file in bytes. */
	static int FileSize(File* file);
	
	/*! \brief Check if a file or directory exists.
	 *  \param filename The path and name of the file to check.
	 *  \returns        True if the file exists, or false if it does not.
	 *  \remarks        This function will also return true if the file is actually
	 *                  a directory.  Use IsDirectory() to find out if it is a file or
	 *                  a directory. */
	static bool Exists(GString filename);
	
	/*! \brief Delete a file.
	 *  \param filename The path and name of the file to delete. */
	static void DeleteFile(GString filename);
	
	//@}
	
	/*! \name Directory Manipulation Functions */
	//@{
	
	struct Directory;
	
	/*! \brief Open a directory for listing.
	 *  \param dirName The path and name of the directory.
	 *  \returns       A pointer to a directory, or NULL if an error occured. */
	static Directory* OpenDirectory(GString dirName);
	/*! \brief Read the name of one object in the directory.
	 *  \param filename  A pointer to a string which should be replaced with the name of the
	 *                   file or sub directory.
	 *  \param directory A pointer to a directory, obtained by calling OpenDirectory().
	 *  \returns         True if a name was read, or false if there were no more objects to
	 *                   read */
	static bool ReadDirectory(GString* filename, Directory* directory);
	/*! \brief Restart reading from the beginning of a directory.
	 *  \param directory A pointer to a directory, obtained by calling OpenDirectory(). */
	static void RewindDirectory(Directory* directory);
	/*! \brief Close the directory.
	 *  \param directory A pointer to a directory, obtained by calling OpenDirectory().
	 *
	 *  Call this when you are done reading the directory. */
	static void CloseDirectory(Directory* directory);
	
	/*! \brief Create a directory.
	 *  \param dirname The path and name of the directory to create. */
	static void CreateDirectory(GString dirname);
	/*! \brief Remove an empty directory.
	 *  \param dirname The path and name of the directory to remove.  This directory
	 *                 must be empty. */
	static void DeleteDirectory(GString dirname);
	/*! \brief Recursively delete a directory and all of its contents.
	 *  \param dirname The path and name of the directory to remove.  This directory
	 *                 does <i>not</i> have to be empty. */
	static void DeleteTree(GString dirname);
	/*! \brief Check if an object is a file or a directory.
	 *  \param dirname The path and name of the filesystem object to check.
	 *  \returns       True if the object was a directory, or false if it was not. */
	static bool IsDirectory(GString dirname);
	
	//@}
};

} // namespace GAUGE3D

#endif //included
