/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

#ifndef __GAUGE3D_PLUGINS_PLUGIN_H__INCLUDED__
#define __GAUGE3D_PLUGINS_PLUGIN_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include <gauge3d/stddefs.h>
#endif //redundant include guard

// system ------------------------------------------------------------
// external packages -------------------------------------------------
#ifndef __GAUGE3D_COMMON_GSTRING_H__INCLUDED__
#include <gauge3d/common/gstring.h>
#endif //redundant include guard

#ifndef __GAUGE3D_FILES_FILE_H__INCLUDED__
#include <gauge3d/files/file.h>
#endif

#ifndef __GAUGE3D_COMMON_ARRAY_H__INCLUDED__
#include <gauge3d/common/array.h>
#endif

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

// class definition ==================================================
/*! \class GPlugin plugin.h gauge3d/plugins/plugin.h
 *  \ingroup plugins
 *  \brief Base class for objects that tell GAUGE what a plugin can do.
 *
 *  You only need to know about this class if you are writing a plugin,
 *  not if you are only using them.
 *
 *  Derivatives of GPlugin are used to describe what sort of objects a
 *  plugin contains, and create those objects for the object loader.
 *
 *  Here is what happens when a plugin is opened:
 *  <ol>
 *   <li>GObjectLoader uses GLibrary to load a shared library on disk.
 *   <li>The shared library exports an object derived from GPlugin.
 *       GObjectLoader imports this object.
 *   <li>GObjectLoader calls GPlugin::ObjectList() on the imported plugin
 *       object to find out what other types of objects the plugin contains.
 *   <li>Later on, when an object is requested, GObjectLoader finds a plugin
 *       that contains a suitable object.  It calls one of this plugin's
 *       GPlugin::CreateObject() functions to create the object.
 *   <li>The plugin creates the object and returns a pointer to it.
 *  </ol> */
class GAUGE3D_API GPlugin: public GObject
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39781f8f);
public:
	// lifecycle -------------------------------------------------------
	virtual ~GPlugin();
	
	// inquiry ---------------------------------------------------------
	//! Returns the human-readable name of the plugin.
	virtual GString Name() = 0;
	//! Returns a human-readable description of the plugin.
	virtual GString Description() = 0;
	
	//InUse returns true if the plugin is being used.  Otherwise, false.
	// The object loader uses this to determine when it is safe to unload
	// a plugin from RAM.
	/*! \brief Returns true if the plugin is in use.
	 *  \remarks GObjectLoader uses this to determine when it is safe to
	 *           unload the library from RAM.  It should return true as
	 *           long as an instance of any class in the plugin exists. */
	virtual bool InUse() = 0;
	
	// operations ------------------------------------------------------
	//! Creates an instance of the class with the given name.
	virtual pGObject CreateObject(const GString& className) = 0;
	/*! \brief Creates an object of the class with the given name using the
	 *         given file.
	 *
	 * For example, if the class loads jpeg images, the file would be a jpeg
	 * image to load.
	 *
	 *  \returns A pointer to the created object, or NULL if the file could
	 *           not be loaded for some reason. */
	virtual pGObject CreateObject(const GString& className, pGFile file) = 0;
	
	/*! \struct ObjectInfo plugin.h gauge3d/plugins/plugin.h
	 *  \brief Used to describe the objects that a plugin contains.
	 *
	 * The plugin tells the object loader what objects are available from
	 * it by giving it an array of ObjectInfo objects.  Each one gives
	 * info for one of the available objects. */
	struct ObjectInfo
	{
		//! The class name of the object.
		GString mClassname;
		
		//! The human-readable name of the object.
		GString mName;
		//! A human-readable description of the object.
		GString mDescription;
		
		//! A list of all the base classes of the object (excluding GObject).
		GArray<GString> mBaseClasses;
		
		/*! \brief A list of keywords that users can use to request the object.
		 *
		 * For example, the OpenGL renderer has the keyword "opengl", and
		 * the D3D renderer has "direct3d".  Keywords should be all
		 * lower-case letters. */
		GArray<GString> mKeywords;
		
		/*! \brief A list of file types which are supported by the object.
		 *
		 * Types are represented by the extension of the file.
		 * For example:  "mp3", "bsp", "tgz".
		 * Names should be all lower-case and should not include the '.'. */
		GArray<GString> mFileTypes;
	};
	
	//! Gets ObjectInfos for every object available from this plugin.
	virtual GArray<ObjectInfo> ObjectList() = 0;
};
/*! \typedef pGPlugin
 *  \ingroup plugins
 *  \brief \link GAUGE3D::GSmartPointer Smart pointer\endlink to a
 *         \link GAUGE3D::GPlugin GPlugin\endlink. */
typedef GSmartPointer<GPlugin> pGPlugin;


//Use this macro to declare a plugin.  CLASSNAME is a class
// derived from GPlugin.
#if defined(PLATFORM_MSVC) && !defined(DOXYGEN)

#define GAUGE3D_EXPORT_PLUGIN(CLASSNAME)                      \
CLASSNAME __gauge3d_plugin;                                   \
                                                              \
/* needs to be extern "C" to avoid name mangling.           */\
extern "C"                                                    \
{                                                             \
__declspec(dllexport)                                         \
void* __gauge3d_plugin_create() {return &__gauge3d_plugin;}   \
}

#else //PLATFORM_MSVC

/*! \ingroup plugins
 *  \brief Export a GPlugin derivative from the library.
 *  \param CLASSNAME A class derived from GPlugin to export.
 *
 *  This macro should be used once in every plugin.  It exports
 *  the given class so that GObjectLoader can load that class
 *  and access the rest of the plugin through it. */
#define GAUGE3D_EXPORT_PLUGIN(CLASSNAME)                      \
CLASSNAME __gauge3d_plugin;                                   \
                                                              \
/* needs to be extern "C" to avoid name mangling.           */\
extern "C"                                                    \
{                                                             \
void* __gauge3d_plugin_create() {return &__gauge3d_plugin;}   \
}

#endif //!PLATFORM_MSVC

} // namespace GAUGE3D

#endif //included
