/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: gifimage.cpp
 *      Package: GIF image file plugin
 *    Component: ImgPak
 *  Description: Loads GIF Images
 * Dependencies: GImage
 *               GFile
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "gifimage.h"

// system ------------------------------------------------------------
//#include <windows.h>    //bad Steve!  :)
// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

const uint16 MAX_CODES = 4096;
const char *GIF_HEADER = "GIF";
const char *VERSION_87 = "87a";
const char *VERSION_89 = "89a";
const uint8 IMAGE_BLOCK = 0x2C;
const uint8 EXTENSION_BLOCK = 0x21;
const uint8 GRAPHIC_CTRL_EXT = 0xF9;
const uint8 COMMENT_EXT = 0xFE;
const uint8 PLAIN_TEXT_EXT = 0x01;
const uint8 APPLICATION_EXT = 0xFF;
const uint8 TRAILER = 0x3B;

unsigned short GpGifImage::mCodeMask[] = {
	0x0000, 0x0001, 0x0003,	0x0007, 0x000f, 0x001f,	0x003f, 
	0x007f, 0x00ff,	0x01ff, 0x03ff, 0x07ff,	0x0fff
};

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GpGifImage::GpGifImage(pGFile giffile)
	: mReadOk(false)
{
	mColorTable.mpData = NULL;
	mReadOk = Load(giffile);
}

GpGifImage::~GpGifImage()
{
	if(mColorTable.mpData)
		delete [] mColorTable.mpData;
}

// operations --------------------------------------------------------
//Data() retuns the raw image data in the format given by GetFormat().
//mipmapLevel is the mip map to retun.  For example, for a 64x64 image,
// level 0 is 64x64, 1 is 32x32, 2 is 16x16, etc.  Images that don't provide
// mip maps should ignore this.
GArray<uint8>
GpGifImage::Data(int mipmapLevel)
{
	GArray<uint8> ImageData(mWidth * mHeight * 4);
	uint8 *FormattedData = &ImageData[0];
	int DataSize = mWidth * mHeight;
	for(int c=0;c<DataSize;c++) {
		int ind = mData[c] * 3;
		*(FormattedData++) = mColorTable.mpData[ind++];
		*(FormattedData++) = mColorTable.mpData[ind++];
		*(FormattedData++) = mColorTable.mpData[ind++];
		if(!mColorTable.mHasTransparentColor || (mColorTable.mTransparentColor != mData[c]))
			*(FormattedData++) = 0xFF;
		else
			*(FormattedData++) = 0x00;
	}
	return ImageData;
}

bool
GpGifImage::Load(pGFile gifFile)
{
	char buf[3];
	unsigned char id,data,skip[16];

	pGRawIStream inSrc = gifFile->GetRawInputStream();
	if(inSrc == NULL)
		return false;
	GIStream input(inSrc);
	input.Endianness(GIStream::LITTLE);

	input.Read(buf,3);
	if(memcmp(buf,GIF_HEADER,3) != 0)
		return false;
	input.Read(buf,3);
	if(memcmp(buf,VERSION_87,3) == 0) {
		mVersion = 87;
	}
	else if(memcmp(buf,VERSION_89,3) == 0) {
		mVersion = 89;
	}
	else {
		return false;
	}
	input.Read(&mWidth,1);
	input.Read(&mHeight,1);
	input.Read(&data,1);
	input.Read(skip,2);
	if(data & 0x80) {
		mColorTable.mSize = 1 << ((data & 7) + 1);
		mColorTable.mpData = new uint8[mColorTable.mSize * 3];
		input.Read(mColorTable.mpData,mColorTable.mSize * 3);
	}
	else {
		memset(&mColorTable,0,sizeof(mColorTable));
	}
	do {
		input.Read(&id,1);
		switch(id) {
		case EXTENSION_BLOCK:
			{
				unsigned char type;

				input.Read(&type,1);
				switch(type) {
				case GRAPHIC_CTRL_EXT:
					input.Read(skip,1);
					input.Read(&data,1);
					mColorTable.mHasTransparentColor = (data & 0x01);
					input.Read(skip,2);
					input.Read(&mColorTable.mTransparentColor,1);
					input.Read(skip,1);
					break;
				case COMMENT_EXT:
					do {
						input.Read(&data,1);
					} while(data != 0);
					break;
				case PLAIN_TEXT_EXT:
					input.Read(skip,13);
					do {
						input.Read(&data,1);
					} while(data != 0);
					break;
				case APPLICATION_EXT:
					input.Read(skip,12);
					break;
				default:
					return false;
				}
			}
			break;
		case IMAGE_BLOCK:
			input.Read(skip,4);
			input.Read(&mWidth,1);
			input.Read(&mHeight,1);
			input.Read(&data,1);
			mInterlaced = ((data & 0x40) != 0);
			mData = GArray<uint8>(mWidth * mHeight);
			mGroup = 1;
			mLine = 0;
			if(!Decoder(input,mWidth))
				return false;
			if(data & 0x80) {
				if(mColorTable.mpData) {
					delete [] mColorTable.mpData;
					mColorTable.mpData = NULL;
				}
				mColorTable.mSize = 1 << ((data & 7) + 1);
				mColorTable.mpData = new uint8[mColorTable.mSize * 3];
				input.Read(mColorTable.mpData,mColorTable.mSize * 3);
			}
			break;
		default:
			return false;
		}
	} while(id != IMAGE_BLOCK);
	mFormat.mWidth = mWidth;
	mFormat.mHeight = mHeight;
	mFormat.mNumChannels = 4;
	mFormat.mBitsPerChannel = 8;
	mFormat.mHasMipmaps = false;
	mFormat.mEncoding = "";  //TODO
	return true;
}

bool
GpGifImage::Decoder(GIStream& input,int lineWidth)
{
	unsigned char stack[MAX_CODES];
	unsigned short prefix[MAX_CODES];
	unsigned char suffix[MAX_CODES];
	unsigned char *sp,*bufptr,*buf;
	unsigned char curByte = 0;
	int size,curSize;
	int slot,topSlot;
	int clear,ending;
	int newCodes,code;
	int bytesLeft,bitsLeft;
	int c,fc,oc,bufcnt;

	if((size = GetByte(input)) < 0)
		return false;
	if(size < 2 || size > 9)
		return false;
	curSize = size + 1;
	topSlot = 1 << curSize;
	clear = 1 << size;
	ending = clear + 1;
	slot = newCodes = ending + 1;
	bytesLeft = bitsLeft = 0;
	oc = fc = 0;
	buf = new unsigned char[lineWidth + 1];
	sp = stack;
	bufptr = buf;
	bufcnt = lineWidth;
	curByte = 0;
	while((c = GetNextCode(input,curSize,bytesLeft,bitsLeft,curByte)) != ending) {
		if(c == -TRAILER)
			break;
		if (c < 0) {
			delete [] buf;
			return false;
		}
		if(c == clear) {
			curSize = size + 1;
			slot = newCodes;
			topSlot = 1 << curSize;
			while((c = GetNextCode(input,curSize,bytesLeft,bitsLeft,curByte)) == clear);
			if(c == ending)
				break;
			if(c >= slot)
				c = 0;
			oc = fc = c;
			*bufptr++ = c;
			if(--bufcnt == 0) {
				OutLine(buf,lineWidth);
				bufptr = buf;
				bufcnt = lineWidth;
			}
		}
		else {
			code = c;
			if(code >= slot) {
				code = oc;
				*sp++ = fc;
			}
			while(code >= newCodes) {
				*sp++ = suffix[code];
				code = prefix[code];
			}
			*sp++ = code;
			if(slot < topSlot) {
				suffix[slot] = fc = code;
				prefix[slot++] = oc;
				oc = c;
			}
			if(slot >= topSlot) {
				if (curSize < 12) {
					topSlot <<= 1;
					++curSize;
				}
			}
			while(sp > stack) {
				*bufptr++ = *(--sp);
				if(--bufcnt == 0) {
					OutLine(buf,lineWidth);
					bufptr = buf;
					bufcnt = lineWidth;
				}
			}
		}
	}
	if(bufcnt != lineWidth)
		OutLine(buf,(lineWidth - bufcnt));
	delete [] buf;
	return true;
}

int
GpGifImage::GetNextCode(GIStream& input,int curSize,int& bytesLeft,int& bitsLeft,unsigned char& curByte)
{
	static unsigned char byteBuf[257];
	static unsigned char *pBytes = byteBuf;
	int ret;

	if(bitsLeft == 0) {
		if(bytesLeft <= 0) {
			pBytes = byteBuf;
			if((bytesLeft = GetByte(input)) < 0)
				return bytesLeft;
			if(bytesLeft)
				input.Read(byteBuf,bytesLeft);
		}
		curByte = *pBytes++;
		bitsLeft = 8;
		--bytesLeft;
	}
	ret = curByte >> (8 - bitsLeft);
	while(curSize > bitsLeft) {
		if(bytesLeft <= 0) {
			pBytes = byteBuf;
			if((bytesLeft = GetByte(input)) < 0)
				return bytesLeft;
			if(bytesLeft == TRAILER)
				return -bytesLeft;
			if(bytesLeft)
				input.Read(byteBuf,bytesLeft);
		}
		curByte = *pBytes++;
		ret |= (static_cast<int>(curByte) << bitsLeft);
		bitsLeft += 8;
		--bytesLeft;
	}
	bitsLeft -= curSize;
	ret &= mCodeMask[curSize];
	return ret;
}

void
GpGifImage::OutLine(unsigned char *pixels,int length)
{
	if(mLine >= mHeight)
		return;
	memcpy(&mData[mLine * length],pixels,length);
	if(mInterlaced) {
		switch(mGroup) {
		case 1:
			mLine += 8;
			if(mLine >= mHeight) {
				mLine = 4;
				mGroup = 2;
			}
			break;
		case 2:
			mLine += 8;
			if(mLine >= mHeight) {
				mLine = 2;

				mGroup = 3;
			}
			break;
		case 3:
			mLine += 4;
			if(mLine >= mHeight) {
				mLine = 1;
				mGroup = 4;
			}
			break;
		case 4:
			mLine += 2;
			break;
		}
	}
	else {
		mLine++;
	}
}

GAUGE3D_DECLARE_INTERFACES(GpGifImage)
	GAUGE3D_DECLARE_BASECLASS(GImage)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
