/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: tgaimage.cpp
 *      Package: TGA image file plugin
 *    Component: ImgPak
 *  Description: Loads TGA Images
 * Dependencies: GImage
 *               GFile
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "tgaimage.h"

// system ------------------------------------------------------------
// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

struct TgaHeader
{
	uint8  mIdLength;
	uint8  mColorMapType;
	uint8  mImageType;
	uint16 mCMFirstEntryIndex;
	uint16 mCMLength;
	uint8  mCMSize;
	uint16 mXOrigin;
	uint16 mYOrigin;
	uint16 mWidth;
	uint16 mHeight;
	uint8  mBPP;
	uint8  mDescriptor;
};

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GpTgaImage::GpTgaImage(pGFile tgafile)
	: mReadOk(false)
{
	TgaHeader tgaHeader;
	int BitsPerPixel,BytesPerPixel,LinePitch,i;

	pGRawIStream inSrc = tgafile->GetRawInputStream();
	if(inSrc == NULL)
		return;
	GIStream input(inSrc);
	input.Endianness(GIStream::LITTLE);
	
	//Hi Steve!  Here's what I fixed:
	input.Read(&tgaHeader.mIdLength, 3);
	input.Read(&tgaHeader.mCMFirstEntryIndex, 2);
	input.Read(&tgaHeader.mCMSize, 1);
	input.Read(&tgaHeader.mXOrigin, 4);
	input.Read(&tgaHeader.mBPP, 2);
	//That's all.  It used to read the whole header in one command, but that
	// is bad because:
	// (1) Bytes would not be swapped properly on big-endian systems.
	// (2) It requires that the header structure be tightly backed, which is
	//    a compiler-specific feature (which broke it for me).
	//Keep up the good work!
	
	if(tgaHeader.mIdLength)
		input.Seek(tgaHeader.mIdLength,GIStream::CURRENT);
	BitsPerPixel = (tgaHeader.mColorMapType == 1) ? tgaHeader.mCMSize : tgaHeader.mBPP;
	if(BitsPerPixel == 16)
		BitsPerPixel = 24;
	mFormat.mNumChannels = 3;
	if(tgaHeader.mDescriptor & 0x0F)
		++mFormat.mNumChannels;
	mFormat.mBitsPerChannel = BitsPerPixel / 4;
	mFormat.mWidth = tgaHeader.mWidth;
	mFormat.mHeight = tgaHeader.mHeight;
	if(tgaHeader.mBPP == 16)
		BytesPerPixel = 3;	// We do 16 to 24 bit conversion here to simplify output format
	else
		BytesPerPixel = (tgaHeader.mBPP + 7) / 8;
	LinePitch = BytesPerPixel * mFormat.mWidth;
	if(tgaHeader.mColorMapType == 1) {
		mPalette = GArray<uint8>((tgaHeader.mCMFirstEntryIndex + tgaHeader.mCMLength) * 3);
		switch(tgaHeader.mCMSize) {
		case 16:
			for(i=tgaHeader.mCMFirstEntryIndex;i<tgaHeader.mCMLength;i++) {
				uint16 color;
				input.Read(&color,1);
				mPalette[i * 3] = ((color >> 10) & 0x1F) << 3;
				mPalette[i * 3 + 1] = ((color >> 5) & 0x1F) << 3;
				mPalette[i * 3 + 2] = (color & 0x1F) << 3;
			}
			break;
		case 24:
			for(i=tgaHeader.mCMFirstEntryIndex;i<tgaHeader.mCMLength;i++) {
				uint8 color[3];
				input.Read(color,3);
				mPalette[i * 3] = color[2];
				mPalette[i * 3 + 1] = color[1];
				mPalette[i * 3 + 2] = color[0];
			}
			break;
		}
	}
	else if(tgaHeader.mBPP == 8) {	// Gray-scale palette if 8-bit with no palette
		mPalette = GArray<uint8>(768);
		for(int c=0;c<256;c++)
			memset(&mPalette[c * 3],c,3);
	}
	mInverted = (tgaHeader.mDescriptor & 0x20) == 0;
	mReversed = (tgaHeader.mDescriptor & 0x10) != 0;
	mData = GArray<uint8>(mFormat.mWidth * mFormat.mHeight * BytesPerPixel);
	int Index = 0;
	while(Index < LinePitch * mFormat.mHeight) {
		if(tgaHeader.mImageType <= 3) {
			if(tgaHeader.mBPP == 16) {
				for(int c=0;c<mFormat.mWidth;c++) {
					uint16 color;
					input.Read(&color,1);
					mData[Index++] = (color & 0x1f) << 3;
					mData[Index++] = ((color >> 5) & 0x1f) << 3;
					mData[Index++] = ((color >> 10) & 0x1f) << 3;
				}
			}
			else {
				Index += input.Read(&mData[Index],LinePitch);
			}
		}
		else {
			uint8 packet;
			input.Read(&packet,1);
			int size = (packet & 0x7F) + 1;
			if(packet & 0x80) {
				switch(tgaHeader.mBPP) {
				case 1:
				case 8:
					{
						uint8 color;
						input.Read(&color,1);
						memset(&mData[Index],color,size);
						Index += size;
					}
					break;
				case 16:
					{
						uint16 color;
						input.Read(&color,1);
						uint8 conv_color[3];
						conv_color[0] = (color & 0x1f) << 3;
						conv_color[1] = ((color >> 5) & 0x1f) << 3;
						conv_color[2] = ((color >> 10) & 0x1f) << 3;
						for(i=0;i<size;i++) {
							mData[Index++] = conv_color[0];
							mData[Index++] = conv_color[1];
							mData[Index++] = conv_color[2];
						}
					}
					break;
				case 24:
					{
						uint8 color[3];
						input.Read(color,3);
						for(i=0;i<size;i++) {
							mData[Index++] = color[0];
							mData[Index++] = color[1];
							mData[Index++] = color[2];
						}
					}
					break;
				}
			}
			else {
				switch(tgaHeader.mBPP) {
				case 1:
				case 8:
					for(i=0;i<size;i++) {
						uint8 color;
						input.Read(&color,1);
						mData[Index++] = color;
					}
					break;
				case 16:
					for(i=0;i<size;i++) {
						uint16 color;
						input.Read(&color,1);
						mData[Index++] = (color & 0x1f) << 3;
						mData[Index++] = ((color >> 5) & 0x1f) << 3;
						mData[Index++] = ((color >> 10) & 0x1f) << 3;
					}
					break;
				case 24:
					for(i=0;i<size;i++) {
						uint8 color[3];
						input.Read(color,3);
						mData[Index++] = color[0];
						mData[Index++] = color[1];
						mData[Index++] = color[2];
					}
					break;
				}
			}
		}
	}
	mReadOk = true;
}

GpTgaImage::~GpTgaImage()
{
}

// operations --------------------------------------------------------
//Data() retuns the raw image data in the format given by GetFormat().
//mipmapLevel is the mip map to retun.  For example, for a 64x64 image,
// level 0 is 64x64, 1 is 32x32, 2 is 16x16, etc.  Images that don't provide
// mip maps should ignore this.
GArray<uint8>
GpTgaImage::Data(int mipmapLevel)
{
	GArray<uint8> ImageData;
	int ImageSize = mFormat.mWidth * mFormat.mHeight;
	int XStart,XEnd,XDir;
	int YStart,YEnd,YDir;

	XStart = mReversed ? mFormat.mWidth - 1 : 0;
	XEnd = mReversed ? -1 : mFormat.mWidth;
	XDir = mReversed ? -1 : 1;
	YStart = mInverted ? mFormat.mHeight - 1: 0;
	YEnd = mInverted ? -1 : mFormat.mHeight;
	YDir = mInverted ? -1 : 1;
	if(mPalette.Size()) {
		ImageData = GArray<uint8>(ImageSize * 3);
		uint8 *Output = &ImageData[0];
		for(int y=YStart;y!=YEnd;y+=YDir) {
			for(int x=XStart;x!=XEnd;x+=XDir) {
				int ind = mData[y * mFormat.mWidth + x] * 3;
				*(Output++) = mPalette[ind++];
				*(Output++) = mPalette[ind++];
				*(Output++) = mPalette[ind++];
			}
		}
	}
	else {
		if(mFormat.mNumChannels == 4) {
			ImageData = GArray<uint8>(ImageSize * 4);
			uint8 *Output = &ImageData[0];
			for(int y=YStart;y!=YEnd;y+=YDir) {
				for(int x=XStart;x!=XEnd;x+=XDir) {
					uint8 *Source = &mData[(y * mFormat.mWidth * 4) + (x * 4)];
					*(Output++) = Source[3];
					*(Output++) = Source[2];
					*(Output++) = Source[1];
					*(Output++) = Source[0];
				}
			}
		}
		else {
			ImageData = GArray<uint8>(ImageSize * 3);
			uint8 *Output = &ImageData[0];
			for(int y=YStart;y!=YEnd;y+=YDir) {
				for(int x=XStart;x!=XEnd;x+=XDir) {
					uint8 *Source = &mData[(y * mFormat.mWidth * 3) + (x * 3)];
					*(Output++) = Source[2];
					*(Output++) = Source[1];
					*(Output++) = Source[0];
				}
			}
		}
	}
	return ImageData;
}

GAUGE3D_DECLARE_INTERFACES(GpTgaImage)
	GAUGE3D_DECLARE_BASECLASS(GImage)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
