/* GAUGE^3D - Grand Unified (3D) Game Engine
 * JPEG Image Loading Plugin
 * Copyright (C) 2000 Will Newton
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: jpegimage.cpp
 *      Package: JPEG image file plugin
 *    Component: JpegImg
 *  Description: Loads JPEG Images
 * Dependencies: GImage
 *               GFile
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "jpegimage.h"
#include <stdio.h>

// system ------------------------------------------------------------
// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {


// class member functions ============================================
// lifecycle ---------------------------------------------------------

typedef struct
{
	struct jpeg_source_mgr pub;   /* public fields */
	GIStream *stream;                 /* source stream */
	JOCTET *buffer;               /* start of buffer */
	boolean start_of_file;        /* have we gotten any data yet? */
} my_source_mgr;


void jpeg_error(j_common_ptr cinfo)
{	
}

void init_source(j_decompress_ptr cinfo)
{
	        // NOP
}

void term_source(j_decompress_ptr cinfo)
{
	// NOP
}

boolean fill_buffer(j_decompress_ptr cinfo)
{
	int bytes=0;
	my_source_mgr *src = (my_source_mgr*) cinfo->src;	
	
	bytes = src->stream->Read((void *)(src->buffer), INPUT_BUF_SIZE);

	if (bytes <= 0) {
		if (src->start_of_file)
		{
			return FALSE;
		}
		/* Insert a fake EOI marker */
		src->buffer[0] = (JOCTET) 0xFF;
		src->buffer[1] = (JOCTET) JPEG_EOI;
		bytes = 2;
	}

	src->pub.next_input_byte = src->buffer;
	src->pub.bytes_in_buffer = bytes;
  	src->start_of_file = FALSE;

	return TRUE;
}
	

void skip_input(j_decompress_ptr cinfo, long num_bytes)
{
	my_source_mgr *src = (my_source_mgr*) cinfo->src;
		
        src->stream->Seek((int)num_bytes,GIStream::CURRENT);
}

void GpJpegImage::DecompressFromStream(j_decompress_ptr cinfo,GIStream &inSrc)
{
	cinfo->src = (struct jpeg_source_mgr *)(*cinfo->mem->alloc_small)
			((j_common_ptr) cinfo, JPOOL_PERMANENT,
			(size_t)sizeof(my_source_mgr));

	my_source_mgr *src = (my_source_mgr *)cinfo->src;
	src->stream = &inSrc;
	src->start_of_file = TRUE;
	src->buffer = (JOCTET *)(*cinfo->mem->alloc_small)(
			(j_common_ptr) cinfo, JPOOL_PERMANENT,
			(INPUT_BUF_SIZE) * sizeof(JOCTET));
	src->pub.init_source = init_source;
	src->pub.fill_input_buffer = fill_buffer;
	src->pub.skip_input_data = skip_input;
	src->pub.resync_to_restart = jpeg_resync_to_restart;
	src->pub.term_source = term_source;
	src->pub.bytes_in_buffer = 0;
	src->pub.next_input_byte = NULL;
}


GpJpegImage::GpJpegImage(pGFile jpegfile)
	: mReadOk(false)
{
	struct jpeg_decompress_struct cinfo;
	int row_stride;
	struct jpeg_error_mgr pub;
	
	pGRawIStream inSrc = jpegfile->GetRawInputStream();

	if(inSrc == NULL)
		return;

	GIStream input(inSrc);
        input.Endianness(GIStream::LITTLE);
	
	cinfo.err = jpeg_std_error(&pub);
	pub.error_exit = jpeg_error;
	    
	jpeg_create_decompress(&cinfo);
	DecompressFromStream(&cinfo, input);
	jpeg_read_header(&cinfo, TRUE);
	jpeg_start_decompress(&cinfo);
        
	row_stride = cinfo.output_width * cinfo.output_components;	
	mData = GArray<uint8>(row_stride * cinfo.output_height);
	
	if (mData)
	{
		while (cinfo.output_scanline < cinfo.output_height) {
			uint8 *line = &mData[cinfo.output_scanline*row_stride]; 
			jpeg_read_scanlines(&cinfo, &line, 1);
		}
		mReadOk = true;
	}

	mFormat.mWidth = row_stride/3;
	mFormat.mHeight = cinfo.output_height;
	mFormat.mNumChannels = cinfo.output_components;
	mFormat.mBitsPerChannel = 8;
	mFormat.mHasMipmaps = false;
	mFormat.mEncoding = "raw24";
	
	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);
}

GpJpegImage::~GpJpegImage()
{
}


// operations --------------------------------------------------------
//Data() retuns the raw image data in the format given by GetFormat().
//mipmapLevel is the mip map to retun.  For example, for a 64x64 image,
// level 0 is 64x64, 1 is 32x32, 2 is 16x16, etc.  Images that don't provide
// mip maps should ignore this.
GArray<uint8> GpJpegImage::Data(int mipmapLevel)
{
	return GArray<uint8>(mData);
}

GAUGE3D_DECLARE_INTERFACES(GpJpegImage)
	GAUGE3D_DECLARE_BASECLASS(GImage)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
