/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: loader.cpp
 *      Package:
 *    Component:
 *  Description:
 * Dependencies:
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "loader.h"

// system ------------------------------------------------------------
// external packages -------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {
namespace GMDL {

// class member functions ============================================
GpGMDLLoader::~GpGMDLLoader() {}

GString
GpGMDLLoader::FileType()
{
	return "gmdl";
}

bool
GpGMDLLoader::CanSave(pGObject object)
{
	if(QueryInterface<GSuperModel>(object) != NULL)
		return true;
	return false;
}

pGObject
GpGMDLLoader::Load(pGFile file)
{
	//FIXME:  Check file extension?  Probably not necessary.
	
	pGIStream input = file->GetInputStream();
	
	if(input == NULL)
		return NULL;
	
	input->Endianness(GIStream::BIG);
	
	return Read(input);
}

bool
GpGMDLLoader::Save(pGFile file, pGObject object)
{
	pGSuperModel superModel = QueryInterface<GSuperModel>(object);
	if(superModel == NULL)
		return false;    //not a SuperModel -- We only like SuperModels
	
	Model* model = superModel->ModelData();
	if(model == NULL)
		return false;    //model contained no data
	
	pGOStream output = file->GetOutputStream();
	if(output == NULL)
		return false;    //could not open file
	
	output->Endianness(GOStream::BIG);
	
	ChunkHeader chunk;
	int pos1 = output->TellPos();
	
	chunk.mType = CHUNK_MODEL;
	
	(*output) << chunk; //place holder, will write real chunk later
	
	chunk.mSize = WriteModel(model, output) + sizeof(ChunkHeader);
	
	int pos2 = output->TellPos();
	output->Seek(pos1, GOStream::BEGINNING);
	(*output) << chunk;  //write real chunk
	output->Seek(pos2, GOStream::BEGINNING);
	
	return true;
}

pGObject
GpGMDLLoader::Read(pGIStream input)
{
	ChunkHeader chunk;
	
	//find the model chunk.
	(*input) >> chunk;
	if(chunk.mType != CHUNK_MODEL)
		return NULL;  //could not find a model chunk.
	
	Model* model = new Model;
	ReadModel(model, input, chunk.mSize);
	return new GpGMDLModel(model);
}

int
GpGMDLLoader::Write(pGOStream output, pGObject object)
{
	pGSuperModel superModel = QueryInterface<GSuperModel>(object);
	if(superModel == NULL)
		return -1;    //not a SuperModel -- We only like SuperModels
	
	Model* model = superModel->ModelData();
	if(model == NULL)
		return -1;    //model contained no data
	
	ChunkHeader chunk;
	int pos1 = output->TellPos();
	
	chunk.mType = CHUNK_MODEL;
	
	(*output) << chunk; //place holder, will write real chunk later
	
	chunk.mSize = WriteModel(model, output) + sizeof(ChunkHeader);
	
	int pos2 = output->TellPos();
	output->Seek(pos1, GOStream::BEGINNING);
	(*output) << chunk;  //write real chunk
	output->Seek(pos2, GOStream::BEGINNING);
	
	return chunk.mSize;
}


void
GpGMDLLoader::ReadModel(Model* model, GIStream* input, int size)
{
	ChunkHeader subChunk;
	int curMesh = 0;
	int endPos = input->TellPos() - sizeof(ChunkHeader) + size;
	
	(*input) >> model->mNumMeshes;
	if(model->mNumMeshes > 0)
		model->mMeshes = GStridedArray<Mesh>(new Mesh[model->mNumMeshes]);
	
	while(input->TellPos() < endPos)
	{
		(*input) >> subChunk;
		
		switch(subChunk.mType)
		{
		case CHUNK_MESH:
			ReadMesh(&(model->mMeshes[curMesh]), input, subChunk.mSize);
			++curMesh;
			break;
		default:
			SkipChunk(input, subChunk);
			break;
		}
	}
	
	//just in case...
	input->Seek(endPos, GIStream::BEGINNING);
}

int
GpGMDLLoader::WriteModel(Model* model, GOStream* output)
{
	ChunkHeader subChunk;
	int size = 0;
	
	(*output) << model->mNumMeshes;
	size += sizeof(int32);
	
	subChunk.mType = CHUNK_MESH;
	for(int i = 0; i < model->mNumMeshes; i++)
	{
		(*output) << subChunk;
		
		subChunk.mSize = WriteMesh(&(model->mMeshes[i]), output) + sizeof(ChunkHeader);
		size += subChunk.mSize;
		
		output->Seek(-subChunk.mSize, GOStream::CURRENT);
		(*output) << subChunk;
		output->Seek(subChunk.mSize - sizeof(ChunkHeader), GOStream::CURRENT);
	}
	
	return size;
}

void
GpGMDLLoader::ReadMesh(Mesh* mesh, GIStream* input, int size)
{
	ChunkHeader subChunk;
	int curFrame = 0;
	int endPos = input->TellPos() - sizeof(ChunkHeader) + size;
	int32* indexMap;    //array holding mChangedIndices for all edge reductions
	int32 indexMapSize; //size of array
	int indexMapPos;  //current position in this array
	int i;
	
	(*input) >> mesh->mDynamicFlags >> mesh->mNumVertices;
	
	(*input) >> mesh->mNumFrames;
	if(mesh->mNumFrames > 0)
		mesh->mFrames = GStridedArray<MeshFrame>(new MeshFrame[mesh->mNumFrames]);
	
	(*input) >> mesh->mIndexType;
	(*input) >> mesh->mNumIndices;
	if(mesh->mNumIndices > 0)
	{
		mesh->mIndices = new uint16[mesh->mNumIndices];
		input->Read(mesh->mIndices, mesh->mNumIndices);
	}
	
	while(input->TellPos() < endPos)
	{
		(*input) >> subChunk;
		
		switch(subChunk.mType)
		{
		case CHUNK_MESHFRAME:
			ReadMeshFrame(&(mesh->mFrames[curFrame]), input, subChunk.mSize, mesh->mNumVertices);
			++curFrame;
			break;
		case CHUNK_EDGE_REDUCTIONS:
			(*input) >> indexMapSize;
			indexMap = new int32[indexMapSize];
			input->Read(indexMap, indexMapSize);
			
			mesh->mEdgeReductions = new EdgeReduction[mesh->mNumVertices];
			indexMapPos = 0;
			
			for(i = 0; i < mesh->mNumVertices; i++)
			{
				(*input) >> mesh->mEdgeReductions[i].mLODThreshold;
				(*input) >> mesh->mEdgeReductions[i].mCollapseInto;
				(*input) >> mesh->mEdgeReductions[i].mNumIndices;
				(*input) >> mesh->mEdgeReductions[i].mNumChangedIndices;
				mesh->mEdgeReductions[i].mChangedIndices = indexMap + indexMapPos;
				indexMapPos += mesh->mEdgeReductions[i].mNumChangedIndices;
			}
			break;
		default:
			SkipChunk(input, subChunk);
			break;
		}
	}
	
	//just in case...
	input->Seek(endPos, GIStream::BEGINNING);
}

int
GpGMDLLoader::WriteMesh(Mesh* mesh, GOStream* output)
{
	ChunkHeader subChunk;
	int size = 0;
	int i;
	
	(*output) << mesh->mDynamicFlags << mesh->mNumVertices
	          << mesh->mNumFrames << mesh->mIndexType << mesh->mNumIndices;
	if(mesh->mNumIndices > 0)
		output->Write(mesh->mIndices, mesh->mNumIndices);
	
	size += 5 * sizeof(int32) + mesh->mNumIndices * sizeof(uint16);
	
	if(mesh->mEdgeReductions != NULL)
	{
		int32 indexMapSize = 0;
		
		subChunk.mType = CHUNK_EDGE_REDUCTIONS;
		
		for(i = 0; i < mesh->mNumVertices; i++)
			indexMapSize += mesh->mEdgeReductions[i].mNumChangedIndices;
		
		subChunk.mSize = sizeof(ChunkHeader) + sizeof(int32);   //header and indexMapSize
		subChunk.mSize = indexMapSize * sizeof(int32);          //index map
		
		int edgeReductionSize = sizeof(float32) + 3 * sizeof(int32);
		subChunk.mSize += mesh->mNumVertices * edgeReductionSize;
		
		(*output) << subChunk << indexMapSize;
		for(i = 0; i < mesh->mNumVertices; i++)
		{
			output->Write(mesh->mEdgeReductions[i].mChangedIndices,
			              mesh->mEdgeReductions[i].mNumChangedIndices);
		}
		for(i = 0; i < mesh->mNumVertices; i++)
		{
			(*output) << mesh->mEdgeReductions[i].mLODThreshold;
			(*output) << mesh->mEdgeReductions[i].mCollapseInto;
			(*output) << mesh->mEdgeReductions[i].mNumIndices;
			(*output) << mesh->mEdgeReductions[i].mNumChangedIndices;
		}
	}
	
	subChunk.mType = CHUNK_MESHFRAME;
	for(i = 0; i < mesh->mNumFrames; i++)
	{
		(*output) << subChunk;
		
		subChunk.mSize = WriteMeshFrame(&(mesh->mFrames[i]), output, mesh->mNumVertices)
		                 + sizeof(ChunkHeader);
		size += subChunk.mSize;
		
		output->Seek(-subChunk.mSize, GOStream::CURRENT);
		(*output) << subChunk;
		output->Seek(subChunk.mSize - sizeof(ChunkHeader), GOStream::CURRENT);
	}
	
	return size;
}

void
GpGMDLLoader::ReadMeshFrame(MeshFrame* frame, GIStream* input, int size, int numVertices)
{
	ChunkHeader subChunk;
	int endPos = input->TellPos() - sizeof(ChunkHeader) + size;
	int i;
	GCoordinate* data = new GCoordinate[numVertices * 3];
	
	while(input->TellPos() < endPos)
	{
		(*input) >> subChunk;
		
		switch(subChunk.mType)
		{
		case CHUNK_VERTEX_POSITIONS:
			if(frame->mVertices == NULL)
			{
				//allocate vertices and normals together
				frame->mVertices = new GVector[numVertices * 2];
				frame->mNormals  = frame->mVertices + numVertices;
			}
			
			input->Read(data, numVertices * 3);
			for(i = 0; i < numVertices; i++)
			{
				frame->mVertices[i].x = data[i*3    ];
				frame->mVertices[i].y = data[i*3 + 1];
				frame->mVertices[i].z = data[i*3 + 2];
			}
			break;
		case CHUNK_VERTEX_NORMALS:
			if(frame->mVertices == NULL)
			{
				//allocate vertices and normals together
				frame->mVertices = new GVector[numVertices * 2];
				frame->mNormals  = frame->mVertices + numVertices;
			}
			
			input->Read(data, numVertices * 3);
			for(i = 0; i < numVertices; i++)
			{
				frame->mNormals[i].x = data[i*3    ];
				frame->mNormals[i].y = data[i*3 + 1];
				frame->mNormals[i].z = data[i*3 + 2];
			}
			break;
		case CHUNK_VERTEX_TEXCOORDS:
			frame->mTexCoords = new GVector[numVertices];
			
			input->Read(data, numVertices * 3);
			for(i = 0; i < numVertices; i++)
			{
				frame->mTexCoords[i].x = data[i*3    ];
				frame->mTexCoords[i].y = data[i*3 + 1];
				frame->mTexCoords[i].z = data[i*3 + 2];
			}
			break;
		default:
			SkipChunk(input, subChunk);
			break;
		}
	};
	
	delete [] data;
	
	//just in case...
	input->Seek(endPos, GIStream::BEGINNING);
}

int
GpGMDLLoader::WriteMeshFrame(MeshFrame* frame, GOStream* output, int numVertices)
{
	int size = 0;
	int i;
	ChunkHeader subChunk;
	GCoordinate* data = new GCoordinate[numVertices * 3];
	
	if(frame->mVertices != NULL)
	{
		subChunk.mType = CHUNK_VERTEX_POSITIONS;
		subChunk.mSize = sizeof(GCoordinate) * numVertices * 3 + sizeof(ChunkHeader);
		(*output) << subChunk;
		
		for(i = 0; i < numVertices; i++)
		{
			data[i*3    ] = frame->mVertices[i].x;
			data[i*3 + 1] = frame->mVertices[i].y;
			data[i*3 + 2] = frame->mVertices[i].z;
		}
		
		output->Write(data, numVertices * 3);
		size += subChunk.mSize;
	}
	
	if(frame->mNormals != NULL)
	{
		subChunk.mType = CHUNK_VERTEX_NORMALS;
		subChunk.mSize = sizeof(GCoordinate) * numVertices * 3 + sizeof(ChunkHeader);
		(*output) << subChunk;
		
		for(i = 0; i < numVertices; i++)
		{
			data[i*3    ] = frame->mNormals[i].x;
			data[i*3 + 1] = frame->mNormals[i].y;
			data[i*3 + 2] = frame->mNormals[i].z;
		}
		
		output->Write(data, numVertices * 3);
		size += subChunk.mSize;
	}
	
	if(frame->mTexCoords != NULL)
	{
		subChunk.mType = CHUNK_VERTEX_TEXCOORDS;
		subChunk.mSize = sizeof(GCoordinate) * numVertices * 3 + sizeof(ChunkHeader);
		(*output) << subChunk;
		
		for(i = 0; i < numVertices; i++)
		{
			data[i*3    ] = frame->mTexCoords[i].x;
			data[i*3 + 1] = frame->mTexCoords[i].y;
			data[i*3 + 2] = frame->mTexCoords[i].z;
		}
		
		output->Write(data, numVertices * 3);
		size += subChunk.mSize;
	}
	
	delete [] data;
	
	return size;
}


GAUGE3D_DECLARE_INTERFACES(GpGMDLLoader)
	GAUGE3D_DECLARE_BASECLASS(GFileLoader)
GAUGE3D_END_INTERFACES

} //namespace GMDL
} //namespace GAUGE3D
