/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Direct3D 7 Rendering Plugin
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: dxerror.h
 *      Package: Direct3D 7 Rendering Plugin
 *    Component: internal
 *  Description: handles direct x errors
 * Dependencies:
 */

#ifndef __GAUGE3D_PLUGIN_RENDERER_DIRECT3D7_DXERROR_H__INCLUDED__
#define __GAUGE3D_PLUGIN_RENDERER_DIRECT3D7_DXERROR_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
#ifndef __SYSTEM_DDRAW_H__INCLUDED__
#include <ddraw.h>
#define __SYSTEM_DDRAW_H__INCLUDED__
#endif //redundant include guard

#ifndef __SYSTEM_STRING__INCLUDED__
#include <string>
#define __SYSTEM_STRING__INCLUDED__
#endif

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

class GError
{
public:
	virtual uint32 GetErrorCode(void) = 0;
	virtual std::string GetErrorDescription(void) = 0;
	virtual std::string GetOperation(void) = 0;
	virtual std::string GetAPI(void) = 0;
	virtual int32 GetSourceLine(void) = 0;
	virtual std::string GetSourceFile(void) = 0;
	virtual void DebugOutput(void) = 0;
	virtual void UserOutput(void) = 0;
};

class GGaugeError : public GError
{
public:
	enum {
		OUT_OF_MEMORY,
		INVALID_ARG,
		INVALID_DISPLAY,
		NOT_FOUND
	};
	GGaugeError(uint32 ErrorCode,const char *Operation = NULL);
	virtual uint32 GetErrorCode(void) { return mError; }
	virtual std::string GetErrorDescription(void);
	virtual std::string GetOperation(void) { return mOperation; }
	virtual std::string GetAPI(void) { return "Gauge3D"; }
	virtual int32 GetSourceLine(void) { return -1; }
	virtual std::string GetSourceFile(void) { return ""; }
	virtual void DebugOutput(void);
	virtual void UserOutput(void);
private:
	uint32 mError;
	std::string mOperation;
};

#define CREATE_GAUGE3D_ERROR(err,code,op) \
	if(err) { \
		delete err; \
		err = NULL; \
	} \
	err = new GGaugeError(code,op);

class DXError : public GError
{
public:
	DXError(HRESULT Code,const char *Operation = NULL);

	virtual uint32 GetErrorCode(void) { return mHResult; }
	virtual std::string GetErrorDescription(void);
	virtual std::string GetOperation(void) { return mOperation; }
	virtual std::string GetAPI(void) { return "Direct X"; }
	virtual int32 GetSourceLine(void) { return -1; }
	virtual std::string GetSourceFile(void) { return ""; }
	virtual void DebugOutput(void);
	virtual void UserOutput(void);

private:
	HRESULT mHResult;
	std::string mOperation;
};

#define CREATE_DIRECTX_ERROR(err,code,op) \
	if(err) { \
		delete err; \
		err = NULL; \
	} \
	err = new DXError(code,op);

class Win32Error : public GError
{
public:
	Win32Error(const char *Operation = NULL);

	virtual uint32 GetErrorCode(void) { return mError; }
	virtual std::string GetErrorDescription(void);
	virtual std::string GetOperation(void) { return mOperation; }
	virtual std::string GetAPI(void) { return "Win32"; }
	virtual int32 GetSourceLine(void) { return -1; }
	virtual std::string GetSourceFile(void) { return ""; }
	virtual void DebugOutput(void);
	virtual void UserOutput(void);

private:
	DWORD mError;
	std::string mOperation;
};

#define CREATE_WIN32_ERROR(err,op) \
	if(err) { \
		delete err; \
		err = NULL; \
	} \
	err = new Win32Error(op);

int Debug(const char *str,...);
int ShowError(std::string message,std::string caption);

#define RELEASE_ERROR(err) if(err) { delete err; err = NULL; }

} //namespace Plugins
} //namespace GAUGE3D

#endif // included
