/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Direct3D 7 Rendering Plugin
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: d3d7light.cpp
 *      Package: Direct3D 7 Rendering Plugin
 *    Component: Light
 *  Description: D3D implementation of a light
 * Dependencies:
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "d3d7light.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_WINDOWS_H__INCLUDED__
#include <windows.h>
#define __SYSTEM_WINDOWS_H__INCLUDED__
#endif

#ifndef __SYSTEM_D3DX_H__INCLUDED__
#include <d3dx.h>
#define __SYSTEM_D3DX_H__INCLUDED__
#endif //redundant include guard

// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class member functions ============================================
GpD3D7Light::GpD3D7Light(pGpD3D7Light* link,LPDIRECT3DDEVICE7 device)
	: mEnabled(false),
	  mPosition(0.0, 0.0, 0.0),
	  mDirectional(false),
	  mSpotDirection(0.0, 0.0, 0.0),
	  mSpotExponent(0.0),
	  mSpotWidth(-1.0),
	  mNext(*link),
	  mPrev(link),
	  mDevice(device)
{
	mError = NULL;
	mColor[0]       = 1.0;
	mColor[1]       = 1.0;
	mColor[2]       = 1.0;
	mLightProps[0]  = 1.0;
	mLightProps[1]  = 1.0;
	mLightProps[2]  = 0.0;
	mAttenuation[0] = 0.0;
	mAttenuation[1] = 0.0;
	
	*mPrev = this;
	if(mNext != NULL)
		mNext->mPrev = &mNext;
}

GpD3D7Light::~GpD3D7Light()
{
	RELEASE_ERROR(mError);
}
	
void
GpD3D7Light::Register(int num)
{
	HRESULT hr;
	D3DLIGHT7 light;
	D3DCOLORVALUE color;

	if(!mEnabled) {
		mDevice->LightEnable(num,FALSE);
		return;
	}
	hr = mDevice->LightEnable(num,TRUE);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Enabling light");
		return;
	}
	ZeroMemory(&light,sizeof(D3DLIGHT7));
	if(mDirectional)
		light.dltType = D3DLIGHT_DIRECTIONAL;
	else
		light.dltType = D3DLIGHT_POINT;
	color.r = mColor[0] * mLightProps[1];
	color.g = mColor[1] * mLightProps[1];
	color.b = mColor[2] * mLightProps[1];
	color.a = mColor[3];
	light.dcvDiffuse = color;
	color.r = mColor[0] * mLightProps[2];
	color.g = mColor[1] * mLightProps[2];
	color.b = mColor[2] * mLightProps[2];
	color.a = mColor[3];
	light.dcvSpecular = color;
	color.r = mColor[0] * mLightProps[0];
	color.g = mColor[1] * mLightProps[0];
	color.b = mColor[2] * mLightProps[0];
	color.a = mColor[3];
	light.dcvAmbient = color;
	light.dvPosition = D3DVECTOR(mPosition.x,mPosition.y,mPosition.z);
	light.dvDirection = D3DVECTOR(-mPosition.x,-mPosition.y,-mPosition.z);
	light.dvAttenuation0 = 1.0f;
	light.dvAttenuation1 = mAttenuation[0];
	light.dvAttenuation2 = mAttenuation[1];
	light.dvRange = D3DLIGHT_RANGE_MAX;
	hr = mDevice->SetLight(num,&light);
	if(FAILED(hr)) {
		CREATE_DIRECTX_ERROR(mError,hr,"Setting up light");
		return;
	}
}

// operations ------------------------------------------------------
//All lights start out hidden.  After you have cached the light, call Show() to
// enable it.  Show and Hide are a good way to turn lights on and off.
void
GpD3D7Light::Show()
{
	mEnabled = true;
}
void
GpD3D7Light::Hide()
{
	mEnabled = false;
}
//Call Remove to permenantly remove the light from the scene.
void
GpD3D7Light::Remove()
{
	if(mPrev != NULL)
		*mPrev = mNext;
	if(mNext != NULL)
		mNext->mPrev = mPrev;
	mPrev = NULL;
	mNext = NULL;
}

// access ----------------------------------------------------------
//The origin is the location of the light.
void
GpD3D7Light::Position(const GVector& position)
{
	mPosition = position;
	mDirectional = false;
}

//Instead of an origin, you can specify a direction for an infinite light source.
void
GpD3D7Light::Direction(const GVector& direction)
{
	mPosition = direction;
	mDirectional = true;
}

//Use these to set the color of the light.
void
GpD3D7Light::Color(float red, float green, float blue)
{
	mColor[0] = red;
	mColor[1] = green;
	mColor[2] = blue;
}
void
GpD3D7Light::Color(const float color[])
{
	mColor[0] = color[0];
	mColor[1] = color[1];
	mColor[2] = color[2];
}

//LightProps sets the light properties.
void
GpD3D7Light::LightProps(float ambient, float diffuse, float specular)
{
	mLightProps[0] = ambient;
	mLightProps[1] = diffuse;
	mLightProps[2] = specular;
}
void
GpD3D7Light::LightProps2v(const float lightProps[])
{
	mLightProps[0] = lightProps[0];
	mLightProps[1] = lightProps[1];
	mLightProps[2] = 0.0;
}
void
GpD3D7Light::LightProps3v(const float lightProps[])
{
	mLightProps[0] = lightProps[0];
	mLightProps[1] = lightProps[1];
	mLightProps[2] = lightProps[2];
}

//Attenuation allows you to set linear and quadratic attenuation factors.
//(this makes the light get dimmer as you get further away)
void
GpD3D7Light::Attenuation(float linear, float quadratic)
{
	mAttenuation[0] = linear;
	mAttenuation[1] = quadratic;
}

//Spotlight makes the light into a spotlight pointing in the specified
// direction and having the specified width (angle in degrees).  exponent
// specifies how focused the spotlight is.  To disable, specify -1.0 for width.
void
GpD3D7Light::Spotlight(const GVector& direction, float exponent, float width)
{
	mSpotDirection = direction;
	mSpotExponent  = exponent;
	mSpotWidth     = width;
}

GAUGE3D_DECLARE_INTERFACES(GpD3D7Light)
	GAUGE3D_DECLARE_BASECLASS(GLight)
GAUGE3D_END_INTERFACES

} // namespace Plugins
} // namespace GAUGE3D
