/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Direct3D 7 Rendering Plugin
 * Copyright (C) 2000 Steven Haworth
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: d3d7sprite.h
 *      Package: Direct3D 7 Rendering Plugin
 *    Component: Sprite
 *  Description: D3D implementation of a sprite
 * Dependencies: GSprite, GModel, some 3D math thingies
 */

#ifndef __GAUGE3D_PLUGINS_RENDERER_DIRECT3D7_D3D7SPRITE_H__INCLUDED__
#define __GAUGE3D_PLUGINS_RENDERER_DIRECT3D7_D3D7SPRITE_H__INCLUDED__

#ifndef DIRECTINPUT_VERSION
#define DIRECTINPUT_VERSION 0x0700
#define DIRECT3D_VERSION    0x0700
#endif

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#define __GAUGE3D_STDDEFS_H__INCLUDED__
#endif //redundant include guard

// system ------------------------------------------------------------
#ifndef __SYSTEM_DDRAW_H__INCLUDED__
#include <ddraw.h>
#define __SYSTEM_DDRAW_H__INCLUDED__
#endif //redundant include guard

#ifndef __SYSTEM_D3D_H__INCLUDED__
#define D3D_OVERLOADS
#include <d3d.h>
#define __SYSTEM_D3D_H__INCLUDED__
#endif //redundant include guard

// GAUGE^3D ----------------------------------------------------------
#ifndef __GAUGE3D_RENDERER_SPRITE_H__INCLUDED__
#include "gauge3d/renderer/sprite.h"
#define __GAUGE3D_RENDERER_SPRITE_H__INCLUDED__
#endif //redundant include guard

// local package -----------------------------------------------------
#ifndef __GAUGE3D_PLUGINS_RENDERER_DIRECT3D_D3D7TEXTURE_H__INCLUDED__
#include "d3d7texture.h"
#define __GAUGE3D_PLUGINS_RENDERER_DIRECT3D_D3D7TEXTURE_H__INCLUDED__
#endif

#ifndef __LOCAL_DXERROR_H__INCLUDED__
#include "dxerror.h"
#define __LOCAL_DXERROR_H__INCLUDED__
#endif

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class definition ==================================================
class GpD3D7Sprite;
typedef GSmartPointer<GpD3D7Sprite> pGpD3D7Sprite;
class GpD3D7Sprite: public GSprite
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39782296)
public:
	GpD3D7Sprite(pGpD3D7Sprite* link,LPDIRECT3DDEVICE7 device);
	virtual ~GpD3D7Sprite();
	
	void AddToList(pGpD3D7Sprite* link);
	
	//Draw the object via D3D
	void Draw(GVector viewpoint, GQuaternion viewangle, GFrustum viewFrustum, float geoLod);
	
	// operations ------------------------------------------------------
	//Call Show to tell the renderer to start drawing this object whenever it is visible.
	virtual void Show();
	//Call Hide to tell the renderer not to draw this object at all, and not to use it in
	// any calculations such as shadow casting.
	virtual void Hide();
	//Call Remove to permenantly remove the sprite from the scene.
	virtual void Remove();
	
	// access ----------------------------------------------------------
	//A GModel is used to represent the geometry of the object.
	virtual void Model(pGModel model);
	//The current animation state of the model must also be given.
	virtual void AnimState(GModel::pAnimState anim);
	//Set the current texture like so.  The texture object must have come from
	// the same renderer object as the sprite.
	virtual void Texture(pGTexture texture);
	
	//The origin is the location of the object.
	// Angles is the direction the object is pointing.
	virtual void Position(const GVector& origin);
	virtual void Angle(const GQuaternion& angle);
	
	//Use these to set the solid color of the object.  Set to full white to ignore.
	virtual void Color(float red, float green, float blue);  //(alpha unchanged)
	virtual void Color(float red, float green, float blue, float alpha);
	virtual void Color3v(const float color[]);   //RGB, alpha not changed
	virtual void Color4v(const float color[]);   //RGBA
	virtual void Alpha(float alpha);             //Set alpha separately
	
	//LightProps sets the light reflectiveness properties of the object.
	virtual void LightProps(float specular, float shininess, float emission);
	//If fullbright is true, lighting will be ignored and the object will be drawn
	// at maximum brightness.
	virtual void FullBright(bool fullBright);
	
	// Sets a uniform scale factor for the sprite.
	virtual void Scale(GCoordinate magnification);
	
	bool HasAlpha();

	inline pGpD3D7Sprite Next() {return mNext;}
	
private:
	GError *mError;
	pGModel mModel;
	GModel::pAnimState mAnimState;
	GVector mPosition;
	GQuaternion mAngle;
	float mColor[4];   //RGBA
	float mLightProps[3];  //specular, shininess, and emission
	bool mFullBright;
	bool mShow;
	GCoordinate mScaleFactor;
	pGpD3D7Texture mTexture;
	GModel::pRenderInfo mRenderInfo;

	GVector *mVertexData;
	GVector *mTexcoordData;
	uint16* mIndexData;
	int mVertexDataSize;
	int mTexcoordDataSize;
	int mIndexDataSize;

	pGpD3D7Sprite  mNext;
	pGpD3D7Sprite* mPrev;      //actually a pointer to the previous' mNext field

	LPDIRECT3DDEVICE7 mDevice;

	void ApplyOrigin(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum);
	void ApplyAngle(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum);
	void ApplyScale(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum);
	void ApplyLightProps();
	void ApplyColor();
	void DrawGeometry(GVector& viewpoint, GQuaternion& viewangle,
                           GFrustum& viewFrustum, float geoLod);
};

} // namespace Plugins
} // namespace GAUGE3D

#endif //included
