/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: gldisplay.cpp
 *      Package: OpenGL Rendering Plugin
 *    Component: Display
 *  Description: Represents the rendering window.
 * Dependencies: GpGLScene, GpGLSprite, GpGLLight
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "gldisplay.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_FSTREAM_H__INCLUDED__
#include <fstream>
#define __SYSTEM_FSTREAM_H__INCLUDED__
#endif

#ifndef __SYSTEM_IOSTREAM_H__INCLUDED__
#include <iostream>
#define __SYSTEM_IOSTREAM_H__INCLUDED__
#endif
using namespace std;

#ifdef PLATFORM_WINDOWS
#include <windows.h>        //needed for gl.h  :(
#undef ABSOLUTE  //stupid Win32 API!
#undef RELATIVE
#endif

#ifndef __SYSTEM_GL_GL_H__INCLUDED__
#include <GL/gl.h>
#define __SYSTEM_GL_GL_H__INCLUDED__
#endif //redundant include guard

#ifndef __SYSTEM_GL_GLU_H__INCLUDED__
#include <GL/glu.h>
#define __SYSTEM_GL_GLU_H__INCLUDED__
#endif //redundant include guard

// GAUGE^3D ----------------------------------------------------------
#include "gauge3d/input/messagequeue.h"
#include "gauge3d/scheduler.h"

// local package -----------------------------------------------------
#include "glcamera.h"

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class state definition ============================================
void
GpGLDisplay::SaveScreen()
{
	int size = mWidth * mHeight * 3;
	char* screenShot = new char[size];
	glReadBuffer(GL_FRONT);
	glReadPixels(0, 0, mWidth, mHeight, GL_RGB, GL_UNSIGNED_BYTE, screenShot);
	glReadBuffer(GL_BACK);
	
	fstream file("gauge3d.tga", ios::out);
	
	file << (char)0       //ID field size (none)
	     << (char)0       //colormap type (none)
	     << (char)2;      //data type (RGB)
	file << "     ";      //colormap spec (ignored)
	file << (char)0 << (char)0;  //origin X
	file << (char)0 << (char)0;  //origin Y
	file << (char)mWidth  << (char)(mWidth >> 8);  //width
	file << (char)mHeight << (char)(mHeight >> 8); //height
	file << (char)24;     //bits per pixel
	file << (char)0;      //wierd bitfield
	
	for(int i = 0; i < size; i += 3)
	{
		//TGA format wants BGR ordering
		file.write(screenShot + i + 2, 1);
		file.write(screenShot + i + 1, 1);
		file.write(screenShot + i    , 1);
	}
	file.close();
	delete [] screenShot;
}

void
GpGLDisplay::ProcessEvents()
{
	SDL_Event event;
	pGMessage message, message2;
	int xmove, ymove;
	char c[2];
	c[1] = '\0';
	bool pressed;
	GString name;
	
	while(SDL_PollEvent(&event))
	{
		message = message2 = NULL;
		
		switch(event.type)
		{
		case SDL_KEYDOWN:
			//These direct key interpretations are temporary...
			if((event.key.keysym.sym == SDLK_PRINT)
			|| (event.key.keysym.sym == SDLK_F10))
				SaveScreen();
			else if(event.key.keysym.sym == SDLK_PAGEDOWN)
				mGeoLod *= 4.0f;
			else if(event.key.keysym.sym == SDLK_PAGEUP)
				mGeoLod /= 4.0f;
			
			//Translate the event
			c[0] = (char)event.key.keysym.unicode;
			message = new GKeyboardMessage(this, SDL_GetKeyName(event.key.keysym.sym), c, true);
			break;
		case SDL_KEYUP:
			c[0] = (char)event.key.keysym.unicode;
			message = new GKeyboardMessage(this, SDL_GetKeyName(event.key.keysym.sym), c, false);
			break;
		case SDL_MOUSEMOTION:
			if(mCaptureMouse)
			{
				//mouse is captured, so we warp it back to the center of the
				// window whenever it moves.  So, we measure its movement relative
				// to the center of the window.
				xmove = event.motion.x - (mWidth / 2);
				ymove = event.motion.y - (mHeight / 2);
				
				if(xmove != 0)
					message = new GMouseMotionMessage(this, "mouse x",
					                                  (float)xmove / (float)mWidth,
					                                  -1.0f);
				if(ymove != 0)
					message2 = new GMouseMotionMessage(this, "mouse y",
					                                   (float)ymove / (float)mHeight,
					                                   -1.0f);
				
				if((xmove != 0) || (ymove != 0))
					SDL_WarpMouse(mWidth / 2, mHeight / 2);
			}
			else
			{
				if(event.motion.xrel != 0)
					message = new GMouseMotionMessage(this, "mouse x",
					                                  (float)event.motion.xrel / (float)mWidth,
					                                  (float)event.motion.x    / (float)mWidth);
				if(event.motion.yrel != 0)
					message2 = new GMouseMotionMessage(this, "mouse y",
					                                   (float)event.motion.yrel / (float)mHeight,
					                                   (float)event.motion.y    / (float)mHeight);
			}
			break;
		case SDL_MOUSEBUTTONDOWN:
		case SDL_MOUSEBUTTONUP:
			pressed = (event.button.state == SDL_PRESSED);
			
			switch(event.button.button)
			{
			case 0:
				name = "mouse 0";
				break;
			case 1:
				name = "mouse 1";
				break;
			case 2:
				name = "mouse 2";
				break;
			case 3:
				name = "mouse 3";
				break;
			case 4:
				name = "mouse 4";
				break;
			case 5:
				name = "mouse 5";
				break;
			default:
				name = "strange mouse button";
				break;
			}
			
			message = new GMouseButtonMessage(this, name, pressed);
			
			break;
		case SDL_QUIT:
			message = new GQuitMessage(this);
			break;
		default:
			break;
		}
		
		if(message)
			gMainMessageQueue->EnqueueMessage(message);
		if(message2)
			gMainMessageQueue->EnqueueMessage(message2);
	}
}

bool GpGLDisplay::mSDLOpen = false;

// class member functions ============================================
// lifecycle ---------------------------------------------------------
GpGLDisplay::GpGLDisplay()
	: mGeoLod(0.25f / 16384)
{
	if(mSDLOpen)
	{
		cerr << "GpGLDisplay: ERROR:  Sorry, can only open one display at a time." << endl
		     << "                     (limited by SDL and OpenGL)" << SDL_GetError() << endl;
		return;
	}
	
	if(SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		cerr << "GpGLDisplay: ERROR:  Could not initialize SDL:" << endl
		     << "                     " << SDL_GetError() << endl;
		return;
	}
	
	//Tell SDL to translate each key to its printable version.
	SDL_EnableUNICODE(true);
	
	mSDLOpen = true;
	
	gVideoScheduler->CallEveryFrame(this, 0);
}

GpGLDisplay::~GpGLDisplay()
{
	if(mSDLOpen)
	{
		SDL_Quit();
		
		gVideoScheduler->StopCallingEveryFrame(this);
	}
	mSDLOpen = false;
}

// operations --------------------------------------------------------
//Open resets the display resolution and color depth.
//This function must be called once before any rendering can be done.
//If the display could not be openned, false is returned.
//Passing zero for colorDepth will cause the current display depth to be used.
bool
GpGLDisplay::Open(int width, int height, int colorDepth, bool fullScreen)
{
	if(!mSDLOpen)
		return false;      //already reported error

#if 0	
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
#endif
	
	if(fullScreen)
		mFrameBuffer = SDL_SetVideoMode(width, height, colorDepth, SDL_OPENGL | SDL_FULLSCREEN | SDL_DOUBLEBUF);
	else
		mFrameBuffer = SDL_SetVideoMode(width, height, colorDepth, SDL_OPENGL | SDL_DOUBLEBUF);

	if(mFrameBuffer == NULL)
	{
		cerr << "GpGLDisplay: ERROR:  Could not set video mode:" << endl
		     << "                     " << SDL_GetError() << endl;
		return false;
	}
	
	SDL_EnableUNICODE(true);
//	SDL_ShowCursor(false);
	
	glViewport(0, 0, width, height);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glShadeModel(GL_SMOOTH);
	glFrontFace(GL_CW);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	float ambient[] = {0.0, 0.0, 0.0, 1.0};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
#ifdef GL_VERSION_1_2
	glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
#else
	//On OpenGL 1.0 or 1.1, the implementation may support the
	// GL_EXT_separate_specular_color extension.  If it does, this will enable
	// it (the numbers are the defined values for GL_LIGHT_MODEL_COLOR_CONTROL_EXT
	// and GL_SEPARATE_SPECULAR_COLOR_EXT).  If the implementation doesn't support
	// the extension, this should just cause a harmless GL error that we can ignore.
	//In the future, we will actually check if the extension is available.
	glLightModeli((GLenum)0x81F8, (GLenum)0x81FA);
#endif
	
	mWidth  = width;
	mHeight = height;
	mCaptureMouse = false;
	
	return true;
}

//I think this is self-explanitory...
GArray<GDisplay::DisplayProps>
GpGLDisplay::ListAvailableDisplays()
{
	return GArray<GDisplay::DisplayProps>();    //empty list
}

void
GpGLDisplay::Camera(pGCamera camera)
{
	//FIXME: ERROR-CHECK
	mCamera = (pGpGLCamera)camera;
}

void
GpGLDisplay::DoFrame(tTime time, tTime frametime)
{
	if(mCamera == NULL)
		return;
	
	//TODO:  Figure out clear screen stuff.
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	mCamera->Draw(mGeoLod);
	
	SDL_GL_SwapBuffers();
	
	//TODO:  ProcessEvents() should be scheduled with input, not video.
	ProcessEvents();
}


void
GpGLDisplay::CaptureMouse(bool enable)
{
	mCaptureMouse = enable;
}


GAUGE3D_DECLARE_INTERFACES(GpGLDisplay)
	GAUGE3D_DECLARE_BASECLASS(GDisplay)
GAUGE3D_END_INTERFACES

} // namespace Plugins
} // namespace GAUGE3D
