/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: gllight.cpp
 *      Package: OpenGL Rendering Plugin
 *    Component: Light
 *  Description: GL implementation of a light
 * Dependencies:
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "gllight.h"

// system ------------------------------------------------------------
#ifdef PLATFORM_WINDOWS
#include <windows.h>        //needed for gl.h  :(
#endif

#include <GL/gl.h>

#include <string.h>

// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class member functions ============================================
GpGLLight::GpGLLight(pGpGLLight* link)
	: mEnabled(false),
	  mPosition(0.0, 0.0, 0.0),
	  mSpotDirection(0.0, 0.0, 0.0),
	  mSpotExponent(0.0),
	  mSpotWidth(-1.0),
	  mNext(*link),
	  mPrev(link)
{
	mColor[0]       = 1.0;
	mColor[1]       = 1.0;
	mColor[2]       = 1.0;
	mLightProps[0]  = 1.0;
	mLightProps[1]  = 1.0;
	mLightProps[2]  = 0.0;
	mAttenuation[0] = 0.0;
	mAttenuation[1] = 0.0;
	
	*mPrev = this;
	if(mNext != NULL)
		mNext->mPrev = &mNext;
}

GpGLLight::~GpGLLight()
{
}
	
void
GpGLLight::Register(int num)
{
	GLenum light = (GLenum)(GL_LIGHT0 + num);
	float color[4];
	GCoordinate origin[4];
	int i;
	
	if(!mEnabled)
	{
		glDisable(light);
		return;
	}
	
	glEnable(light);
	
	
	memcpy(origin, &mPosition.x, sizeof(GCoordinate) * 3);
	if(mDirectional)
		origin[3] = 0.0;
	else
		origin[3] = 1.0;
	glLightfv(light, GL_POSITION, origin);
	
	color[3] = mColor[3];
	for(i = 0; i < 3; i++)
		color[i] = mColor[i] * mLightProps[0];
	glLightfv(light, GL_AMBIENT, color);
	for(i = 0; i < 3; i++)
		color[i] = mColor[i] * mLightProps[1];
	glLightfv(light, GL_DIFFUSE, color);
	for(i = 0; i < 3; i++)
		color[i] = mColor[i] * mLightProps[2];
	glLightfv(light, GL_SPECULAR, color);
	
	if(mSpotWidth == -1.0)
		glLightf(light, GL_SPOT_CUTOFF, 180.0);  //disable spotlight
	else
	{
		glLightfv(light, GL_SPOT_DIRECTION, &mSpotDirection.x);
		glLightf(light, GL_SPOT_EXPONENT, mSpotExponent);
		glLightf(light, GL_SPOT_CUTOFF, mSpotWidth / 2.0);
	}
	
	glLightf(light, GL_LINEAR_ATTENUATION, mAttenuation[0]);
	glLightf(light, GL_QUADRATIC_ATTENUATION, mAttenuation[1]);
}

// operations ------------------------------------------------------
//All lights start out hidden.  After you have cached the light, call Show() to
// enable it.  Show and Hide are a good way to turn lights on and off.
void
GpGLLight::Show()
{
	mEnabled = true;
}
void
GpGLLight::Hide()
{
	mEnabled = false;
}
//Call Remove to permenantly remove the light from the scene.
void
GpGLLight::Remove()
{
	if(mPrev != NULL)
		*mPrev = mNext;
	if(mNext != NULL)
		mNext->mPrev = mPrev;
	mPrev = NULL;
	mNext = NULL;
}

// access ----------------------------------------------------------
//The origin is the location of the light.
void
GpGLLight::Position(const GVector& position)
{
	mPosition = position;
	mDirectional = false;
}

//Instead of an origin, you can specify a direction for an infinite light source.
void
GpGLLight::Direction(const GVector& direction)
{
	mPosition = direction;
	mDirectional = true;
}

//Use these to set the color of the light.
void
GpGLLight::Color(float red, float green, float blue)
{
	mColor[0] = red;
	mColor[1] = green;
	mColor[2] = blue;
}
void
GpGLLight::Color(const float color[])
{
	mColor[0] = color[0];
	mColor[1] = color[1];
	mColor[2] = color[2];
}

//LightProps sets the light properties.
void
GpGLLight::LightProps(float ambient, float diffuse, float specular)
{
	mLightProps[0] = ambient;
	mLightProps[1] = diffuse;
	mLightProps[2] = specular;
}
void
GpGLLight::LightProps2v(const float lightProps[])
{
	mLightProps[0] = lightProps[0];
	mLightProps[1] = lightProps[1];
	mLightProps[2] = 0.0;
}
void
GpGLLight::LightProps3v(const float lightProps[])
{
	mLightProps[0] = lightProps[0];
	mLightProps[1] = lightProps[1];
	mLightProps[2] = lightProps[2];
}

//Attenuation allows you to set linear and quadratic attenuation factors.
//(this makes the light get dimmer as you get further away)
void
GpGLLight::Attenuation(float linear, float quadratic)
{
	mAttenuation[0] = linear;
	mAttenuation[1] = quadratic;
}

//Spotlight makes the light into a spotlight pointing in the specified
// direction and having the specified width (angle in degrees).  exponent
// specifies how focused the spotlight is.  To disable, specify -1.0 for width.
void
GpGLLight::Spotlight(const GVector& direction, float exponent, float width)
{
	mSpotDirection = direction;
	mSpotExponent  = exponent;
	mSpotWidth     = width;
}


GAUGE3D_DECLARE_INTERFACES(GpGLLight)
	GAUGE3D_DECLARE_BASECLASS(GLight)
GAUGE3D_END_INTERFACES

} // namespace Plugins
} // namespace GAUGE3D
