/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: glsprite.h
 *      Package: OpenGL Rendering Plugin
 *    Component: Sprite
 *  Description: GL implementation of a sprite
 * Dependencies: GSprite, GModel, some 3D math thingies
 */

#ifndef __GAUGE3D_PLUGINS_RENDERER_OPENGL_GLSPRITE_H__INCLUDED__
#define __GAUGE3D_PLUGINS_RENDERER_OPENGL_GLSPRITE_H__INCLUDED__

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE3D_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
// GAUGE^3D ----------------------------------------------------------
#ifndef __GAUGE3D_RENDERER_SPRITE_H__INCLUDED__
#include "gauge3d/renderer/sprite.h"
#endif //redundant include guard

// local package -----------------------------------------------------
#ifndef __GAUGE3D_PLUGINS_RENDERER_OPENGL_GLTEXTURE_H__INCLUDED__
#include "gltexture.h"
#endif

// begin namespace ===================================================
namespace GAUGE3D {
namespace Plugins {

// class definition ==================================================
class GpGLSprite;
typedef GSmartPointer<GpGLSprite> pGpGLSprite;
class GpGLSprite: public GSprite
{
	GAUGE3D_SET_UID(0x00000048, 0x54846eca, 0x39782237)
public:
	GpGLSprite(pGpGLSprite* link);
	virtual ~GpGLSprite();
	
	void AddToList(pGpGLSprite* link);
	
	//Draw the object via GL
	void Draw(GVector viewpoint, GQuaternion viewangle, GFrustum viewFrustum, float geoLod);
	
	// operations ------------------------------------------------------
	//Call Show to tell the renderer to start drawing this object whenever it is visible.
	virtual void Show();
	//Call Hide to tell the renderer not to draw this object at all, and not to use it in
	// any calculations such as shadow casting.
	virtual void Hide();
	//Call Remove to permenantly remove the sprite from the scene.
	virtual void Remove();
	
	// access ----------------------------------------------------------
	//A GModel is used to represent the geometry of the object.
	virtual void Model(pGModel model);
	//The current animation state of the model must also be given.
	virtual void AnimState(GModel::pAnimState anim);
	//Set the current texture like so.  The texture object must have come from
	// the same renderer object as the sprite.
	virtual void Texture(pGTexture texture);
	
	//The location of the object.
	virtual void Position(const GVector& position);
	//Angle is the direction the object is pointing.  Remember, you can convert
	// GAngles to GQuaternion, so you can use a GAngles object here if you prefer.
	virtual void Angle(const GQuaternion& angle);
	
	//Use these to set the solid color of the object.  Set to full white to ignore.
	virtual void Color(float red, float green, float blue);  //(alpha unchanged)
	virtual void Color(float red, float green, float blue, float alpha);
	virtual void Color3v(const float color[]);   //RGB, alpha not changed
	virtual void Color4v(const float color[]);   //RGBA
	virtual void Alpha(float alpha);             //Set alpha separately
	
	//LightProps sets the light reflectiveness properties of the object.
	virtual void LightProps(float specular, float shininess, float emission);
	//If fullbright is true, lighting will be ignored and the object will be drawn
	// at maximum brightness.
	virtual void FullBright(bool fullBright);
	
	
	virtual void Scale(GCoordinate magnification);
	
	bool HasAlpha();
	
	inline pGpGLSprite Next() {return mNext;}
	
private:
	pGModel mModel;
	GModel::pAnimState mAnimState;
	pGpGLTexture mTexture;
	GVector mPosition;
	GQuaternion mAngle;
	float mColor[4];   //RGBA
	pGpGLTexture mSkin;
	float mLightProps[3];  //specular, shininess, and emission
	bool mFullBright;
	bool mShow;
	GCoordinate mScaleFactor;
	GModel::pRenderInfo mRenderInfo;
	
	pGpGLSprite  mNext;
	pGpGLSprite* mPrev;      //actually a pointer to the previous' mNext field
	
	void ApplyOrigin(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum);
	void ApplyAngle(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum);
	void ApplyScale(GVector& viewpoint, GQuaternion& viewangle, GFrustum& viewFrustum);
	void ApplyLightProps();
	void ApplyColor();
	void ApplyTexture();
	void DrawGeometry(GVector& viewpoint, GQuaternion& viewangle,
	                  GFrustum& viewFrustum, float geoLod);
};

} // namespace Plugins
} // namespace GAUGE3D

#endif //included
