/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: gltexture.cpp
 *      Package: OpenGL Rendering Plugin
 *    Component: Texture
 *  Description: Represents a cached texture
 * Dependencies: GTexture
 *               GImage
 */

// includes ==========================================================
// component header --------------------------------------------------
#include "gltexture.h"

// system ------------------------------------------------------------
#ifndef __SYSTEM_GL_GLU_H__INCLUDED__
#include <GL/glu.h>
#define __SYSTEM_GL_GLU_H__INCLUDED__
#endif //redundant include guard

// GAUGE^3D ----------------------------------------------------------
// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

#ifdef PLATFORM_MSVC
//grr, why can't I find an OpenGL 1.2 SDK for MSVC!?
#define GL_UNSIGNED_SHORT_4_4_4_4 0x8033
#define GL_UNSIGNED_SHORT_5_5_5_1 0x8034
#endif

// class member functions ============================================
GpGLTexture::TexFormat
GpGLTexture::ImageFormat2TexFormat(GImage::FormatDesc imageFormat)
{
	TexFormat texFormat;
	
	//FIXME: Check for bad values?  (is it necessary?  It won't be very hard to tell
	// that something is wrong with the texture if someone messes up. :)
	
	switch(imageFormat.mNumChannels)
	{
	case 1:
		texFormat.format = GL_LUMINANCE;
		switch(imageFormat.mBitsPerChannel)
		{
		case 4:
			//FIXME:  Is it possible to have two pixels per byte with GL?
			//        If so, what type do we use?
			texFormat.internalFormat = GL_LUMINANCE4;
			texFormat.type = GL_UNSIGNED_BYTE;
			texFormat.bitsPerPixel = 4;
			break;
		default:
		case 8:
			texFormat.internalFormat = GL_LUMINANCE8;
			texFormat.type = GL_UNSIGNED_BYTE;
			texFormat.bitsPerPixel = 8;
			break;
		}
		break;
	case 2:
		texFormat.format = GL_LUMINANCE_ALPHA;
		switch(imageFormat.mBitsPerChannel)
		{
		case 4:
			//FIXME:  Is this possible with GL?
			texFormat.internalFormat = GL_LUMINANCE4_ALPHA4;
			texFormat.type = GL_UNSIGNED_BYTE;
			texFormat.bitsPerPixel = 8;
			break;
		default:
		case 8:
			texFormat.internalFormat = GL_LUMINANCE8_ALPHA8;
			texFormat.type = GL_UNSIGNED_BYTE;
			texFormat.bitsPerPixel = 16;
			break;
		}
		break;
	default:
	case 3:
		switch(imageFormat.mBitsPerChannel)
		{
		case 4:
			texFormat.format = GL_RGBA;
			texFormat.internalFormat = GL_RGB4;
			texFormat.type = GL_UNSIGNED_SHORT_4_4_4_4;
			texFormat.bitsPerPixel = 16;
			break;
		case 5:
			texFormat.format = GL_RGBA;
			texFormat.internalFormat = GL_RGB5;
			texFormat.type = GL_UNSIGNED_SHORT_5_5_5_1;
			texFormat.bitsPerPixel = 16;
			break;
		default:
		case 8:
			texFormat.format = GL_RGB;
			texFormat.internalFormat = GL_RGB8;
			texFormat.type = GL_UNSIGNED_BYTE;
			texFormat.bitsPerPixel = 24;
			break;
		}
		break;
	case 4:
		texFormat.format = GL_RGBA;
		switch(imageFormat.mBitsPerChannel)
		{
		case 4:
			texFormat.internalFormat = GL_RGBA4;
			texFormat.type = GL_UNSIGNED_SHORT_4_4_4_4;
			texFormat.bitsPerPixel = 16;
			break;
		case 5:
			texFormat.internalFormat = GL_RGB5_A1;
			texFormat.type = GL_UNSIGNED_SHORT_5_5_5_1;
			texFormat.bitsPerPixel = 16;
			break;
		default:
		case 8:
			texFormat.internalFormat = GL_RGBA8;
			texFormat.type = GL_UNSIGNED_BYTE;
			texFormat.bitsPerPixel = 32;
			break;
		}
		break;
	}
	
	return texFormat;
}

void
GpGLTexture::LoadTexture(pGImage image, int layer)
{
	GImage::FormatDesc imageFormat = image->Format();
	TexFormat texFormat;
	
	texFormat = ImageFormat2TexFormat(imageFormat);
	
	glBindTexture(GL_TEXTURE_2D, mLayerIDs[layer]);
	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
	//check if image's width and height are powers of two.
	int32 w, h;
	bool needScale = false;
	for(w = 1 << 30; w > imageFormat.mWidth; w >>= 1);
	if(w < imageFormat.mWidth)
	{
		w <<= 1;        //lowest power of 2 that is greater than the width.
		needScale = true;
	}
	
	for(h = 1 << 30; h > imageFormat.mHeight; h >>= 1);
	if(h < imageFormat.mHeight)
	{
		h <<= 1;        //lowest power of 2 that is greater than the height.
		needScale = true;
	}
	
	if(needScale)
	{
		//we need to scale the image to make the sides be powers of two.
		GArray<uint8> imageData = image->Data(0);
		uint8* scaledImage = new uint8[texFormat.bitsPerPixel * w * h];
		gluScaleImage(texFormat.format, imageFormat.mWidth, imageFormat.mHeight, texFormat.type,
		              (GLvoid*)&imageData[0], w, h, texFormat.type, scaledImage);
		glTexImage2D(GL_TEXTURE_2D, 0, texFormat.internalFormat, w, h, 0, texFormat.format,
		             texFormat.type, (GLvoid*)scaledImage);
		delete [] scaledImage;
	}
	else
	{
		glTexImage2D(GL_TEXTURE_2D, 0, texFormat.internalFormat,
		             imageFormat.mWidth, imageFormat.mHeight,
		             0, texFormat.format, texFormat.type, (GLvoid*)&(image->Data(0))[0]);
	}
}

// lifecycle ---------------------------------------------------------
GpGLTexture::GpGLTexture()
	: mNumLayers(0),
	  mHasAlpha(false)
{
}

GpGLTexture::~GpGLTexture()
{
	if(mNumLayers > 0)
	{
		glDeleteTextures(mNumLayers, mLayerIDs);
		delete [] mLayerIDs;
	}
}

// operations --------------------------------------------------------
//Bind() sets up the pixel pipeline to apply the texture.
void
GpGLTexture::Bind()
{
	//TODO: Multitexturing
	if(mNumLayers > 0)
		glBindTexture(GL_TEXTURE_2D, mLayerIDs[0]);
}

void
GpGLTexture::Surface(pGSurface surface)
{
	//TODO
}

void
GpGLTexture::Image(pGImage image)
{
	if(mNumLayers > 0)
	{
		glDeleteTextures(mNumLayers, mLayerIDs);
		delete [] mLayerIDs;
	}
	
	mNumLayers = 1;
	mLayerIDs = new GLuint[1];
	
	glGenTextures(mNumLayers, mLayerIDs);
	LoadTexture(image, 0);
	
	int numChannels = image->Format().mNumChannels;
	if(numChannels == 2 || numChannels == 4)
		mHasAlpha = true;
}

GAUGE3D_DECLARE_INTERFACES(GpGLTexture)
	GAUGE3D_DECLARE_BASECLASS(GTexture)
GAUGE3D_END_INTERFACES

} //namespace GAUGE3D
