/* GAUGE^3D - Grand Unified (3D) Game Engine
 * Copyright (C) 2000 Kenton Varda
 * <http://www.gauge3d.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

// ===================================================================

/*
 *     Filename: test_frontend.cpp
 *      Package: test progs
 *    Component:
 *  Description: Frontend (graphics/sound/UI) testing programs.
 * Dependencies:
 */

// includes ==========================================================
// stddefs.h ---------------------------------------------------------
#ifndef __GAUGE_STDDEFS_H__INCLUDED__
#include "gauge3d/stddefs.h"
#endif //redundant include guard

// system ------------------------------------------------------------
#include <string.h>
#include <stdio.h>    //sprintf
#include <iostream>
using namespace std;

// external packages -------------------------------------------------
#include "gauge3d/graphics/spheremodel.h"
#include "gauge3d/graphics/supermodel.h"
#include "gauge3d/graphics/billboardsprite.h"

#include "gauge3d/scheduler.h"
#include "gauge3d/engine.h"
#include "gauge3d/files/diskfile.h"
#include "gauge3d/plugins/objectloader.h"
#include "gauge3d/renderer/renderer.h"
#include "gauge3d/renderer/display.h"
#include "gauge3d/renderer/sprite.h"
#include "gauge3d/renderer/light.h"

#include "gauge3d/input/messagequeue.h"

// local package -----------------------------------------------------

// begin namespace ===================================================
namespace GAUGE3D {

static ostream&
operator<<(ostream& os, GString str)
{
	str = str.ToAscii();
	return os << str.AsciiData();
}

class TestImage: public GImage
{
	GArray<uint8> mData;
public:
	TestImage()
		: mData(256 * 256 * 3)
	{
		int pos;
		
		//make a checkerboard
		for(int i = 0; i < 256; i++)
		{
			for(int j = 0; j < 256; j++)
			{
				pos = (i * 256 + j) * 3;
				if(((i+4) / 8 % 2) ^ ((j+4) / 8 % 2))
				{
					mData[pos  ] = 0xFF;
					mData[pos+1] = 0xFF;
					mData[pos+2] = 0xFF;
				}
				else
				{
					mData[pos  ] = 0x00;
					mData[pos+1] = 0x00;
					mData[pos+2] = 0x00;
				}
			}
		}
	}
	~TestImage() {}
	
	virtual FormatDesc Format()
	{
		FormatDesc format;
		format.mWidth          = 256;
		format.mHeight         = 256;
		format.mNumChannels    = 3;
		format.mBitsPerChannel = 8;
		format.mHasMipmaps     = false;
		format.mEncoding       = "raw24";
		return format;
	}
	
	virtual GArray<uint8> Data(int mipmapLevel = 0)
	{
		return mData;
	}
	
	virtual GArray<uint8> EncodedData(int mipmapLevel = 0)
	{
		return mData;
	}
};

const float LIGHT_COLORS[3][4] = {{1.0f, 1.0f,  0.05f, 1.0f},
                                  {1.0f, 0.05f, 1.0f,  1.0f},
                                  {1.0f, 1.0f,  1.0f,  1.0f}};
float gObjectColor[4] = {1.0f, 1.0f, 1.0f, 1.0f};
const GCoordinate LIGHT_DISTANCES[] = {1.65F, 1.8F, 1.95F};
const GCoordinate LIGHT_SPEEDS[] = {0.5F, 0.7F, -1.1F};

class TestScene: public GObject
{
	enum
	{
		NUM_LIGHTS = 3
	};
	
	pGModel mSquareModel;
	pGSphereModel mBigSphere;
	pGSphereModel mSmallSphere;
	pGSprite mBigObj;
	pGSprite mSmallObj[NUM_LIGHTS];
	pGSprite mSquare;
	pGLight  mLight[NUM_LIGHTS];
	
	GSuperModel::pSuperAnimState mSuperAnim;
	int   mFrame;
	tTime mNextThink;
	
	bool mShowLights;
	
	GQuaternion mAngle;
	
public:
	TestScene(pGScene scene, pGModel model, pGTexture texture, int lightLevel)
		: mSquareModel(new GBillboardSprite(2.0f, 2.0f)),
		  mBigSphere  (new GSphereModel),
		  mSmallSphere(new GSphereModel),
		  mAngle      (GVECTOR_I, (GCoordinate)-PI_2)
	{
		if(lightLevel == 0)
			mShowLights = false;
		else
			mShowLights = true;
		
		mBigObj = scene->NewSprite();
		mSquare = scene->NewSprite();
		
		mSquare->Position(GVector(-3.0F, 0.0F, -4.0F));
		mSquare->Angle(GQuaternion(GVECTOR_I, 0.0f));
		mSquare->FullBright(true);
		mSquare->Color(1.0F, 1.0F, 1.0F);
		mSquare->LightProps(0.0F, 0.7F, 0.7F);
		mSquare->Texture(texture);
		mSquare->Model(mSquareModel);
		mSquare->AnimState(mSquareModel->CreateAnimState());
		mSquare->Show();
		
		mBigObj->Position(GVector(0.0F, 0.0F, -3.0F));
		mBigObj->Angle(mAngle);
		mBigObj->Color4v(gObjectColor);
		mBigObj->Texture(texture);
		
		switch(lightLevel)
		{
		case 0:      //no lighting
			mBigObj->FullBright(true);
			break;
		default:
		case 1:      //diffuse only
			mBigObj->FullBright(false);
			mBigObj->LightProps(0.0F, 0.0F, 0.0F);
			break;
		case 2:      //diffuse and specular
			mBigObj->FullBright(false);
			mBigObj->LightProps(0.8F, 30.0F, 0.0F);
			break;
		}
		
		if(mShowLights)
		{
			for(int i = 0; i < NUM_LIGHTS; i++)
			{
				mSmallObj[i] = scene->NewSprite();
				mSmallObj[i]->FullBright(true);
				mSmallObj[i]->Color4v(LIGHT_COLORS[i]);
				mSmallObj[i]->Model(mSmallSphere);
				mSmallObj[i]->AnimState(mSmallSphere->CreateAnimState());
				mSmallObj[i]->Scale(0.06f);
				mSmallObj[i]->Show();
				
				mLight[i] = scene->NewLight();
				if(lightLevel == 2)
					mLight[i]->LightProps(0.0F, 0.7F, 0.7F);
				else
					mLight[i]->LightProps(0.0F, 1.0F, 0.0F);
				mLight[i]->Color(LIGHT_COLORS[i]);
				mLight[i]->Show();
			}
		}
		
		if(model)
		{
			mBigObj->Model(model);
			GModel::pAnimState anim = model->CreateAnimState();
			mBigObj->AnimState(anim);
			//if anim is a super model animation state, we can animate it!
			mSuperAnim = QueryInterface<GSuperModel::SuperAnimState>(anim);
			if(mSuperAnim != NULL)
			{
				mFrame = 0;
				mNextThink = 0.0f;
			}
		}
		else
		{
			mBigObj->Model(mBigSphere);
			mBigObj->AnimState(mBigSphere->CreateAnimState());
		}
		
		mBigObj->Show();
		
		gMiscScheduler->CallEveryFrame(this, 0);
	}
	
	virtual ~TestScene()
	{
		gMiscScheduler->StopCallingEveryFrame(this);
	}
	
	virtual void DoFrame(tTime time, tTime frametime)
	{
		GVector pos;
		double angle;
		
		if(mShowLights)
		{
			angle = time * LIGHT_SPEEDS[0];
			pos = LIGHT_DISTANCES[0] * GVector(
				cos(angle),
				0.0F,
				sin(angle)
				) - GVector(0.0F, 0.0F, 3.0F);
			mSmallObj[0]->Position(pos);
			mLight[0]->Position(pos);
			
			angle = time * LIGHT_SPEEDS[1];
			pos = LIGHT_DISTANCES[1] * GVector(
				cos(angle) / 2.0F,
				cos(angle) * 1.7320508F / 2.0F,
				sin(angle)
				) - GVector(0.0f, 0.0f, 3.0f);
			mSmallObj[1]->Position(pos);
			mLight[1]->Position(pos);
			
			angle = time * LIGHT_SPEEDS[2];
			pos = LIGHT_DISTANCES[2] * GVector(
				cos(angle) / 2.0f,
				cos(angle) * -1.7320508f / 2.0f,
				sin(angle)
				) - GVector(0.0f, 0.0f, 3.0f);
			mSmallObj[2]->Position(pos);
			mLight[2]->Position(pos);
		}
				
		if(mSuperAnim != NULL)
		{
			if(time > mNextThink)
			{
				if(mNextThink == 0.0f)
					mNextThink = time;
				
				mSuperAnim->SetMeshFrame(mFrame, mFrame + 1, mNextThink, mNextThink + 0.5f);
				mNextThink += 0.5f;     //do this again in 1/2 of a second.
				++mFrame;
			}
		}
		
		
		static int framenum = 0;
		static float lasttime = 0.0f;
		float fps;
		
		if(framenum >= 100)
		{
			framenum = 0;
			fps = 100.0f / (time - lasttime);
			lasttime = time;
			cout << fps << " FPS" << endl;
		}
		++framenum;
	}
	
	void Rotate(GQuaternion amount)
	{
		mAngle = amount * mAngle;
		mBigObj->Angle(mAngle);
	}
};

TestScene* gTheTestScene;

class TestMessageHandler: public GMessageQueue::Handler
{
public:
	virtual bool HandleMessage(pGMessage message)
	{
		pGInputMessage input = QueryInterface<GInputMessage>(message);
		if(input)
		{
			cout << "Input message:\n"
			     << "    Name:  " << input->Name() << endl
			     << "    Value: " << input->Value() << endl;
			if(input->IsAnalog())
			{
				if(input->IsRelative())
					cout << "    Type:  Relative Analog" << endl;
				else
					cout << "    Type:  Absolute Analog" << endl;
			}
			else
				cout << "    Type:  Digital" << endl;
			
			pGKeyboardMessage keymessage = QueryInterface<GKeyboardMessage>(input);
			if(keymessage)
			{
				cout << "    Character: " << keymessage->ModifiedKey() << endl;
			}
			
			if(input->Value() == 1.0)
			{
				if(input->Name() == "escape")
					GEngine::Quit();
				else if(input->Name() == "[8]")
					gTheTestScene->Rotate(GQuaternion(GVECTOR_I, -PI / 8.0));
				else if(input->Name() == "[2]")
					gTheTestScene->Rotate(GQuaternion(GVECTOR_I, PI / 8.0));
				else if(input->Name() == "[4]")
					gTheTestScene->Rotate(GQuaternion(GVECTOR_J, -PI / 8.0));
				else if(input->Name() == "[6]")
					gTheTestScene->Rotate(GQuaternion(GVECTOR_J, PI / 8.0));
			}
			
			return true;
		}
		else
		{
			if(QueryInterface<GQuitMessage>(message))
				GEngine::Quit();
			return true;
		}
		
		return false;
	};
};

void
TestFrontend(int resX, int resY, bool fullscreen, int lightLevel, GString model_filename,
             GString image_filename, bool captureMouse)
{
	gMainMessageQueue->AddHandler(new TestMessageHandler);
	
	pGModel model;
	if(model_filename != "")
	{
		cout << model_filename << endl;
		pGDiskFile modelfile(new GDiskFile(model_filename, GDiskFile::NOCREATE));
		model = (pGModel)gMainObjectLoader->LoadObjectForFile(modelfile, "GModel");
		if(!model)
			cout << "couldn't load model " << model_filename << endl;
	}
	
	pGImage image;
	if(image_filename != "")
	{
		pGDiskFile imagefile(new GDiskFile(image_filename,GDiskFile::NOCREATE));
		image = (pGImage)gMainObjectLoader->LoadObjectForFile(imagefile,"GImage");
		if(!image)
		{
			cout << "couldn't load image" << image_filename << endl;
			image = new TestImage;
		}
	}
	else
	{
		image = new TestImage;
	}

	pGRenderer renderer(gMainObjectLoader->LoadObjectOfType("GRenderer"));
	if(renderer == NULL)
	{
		cout << "ERROR: Could not find a rendering plugin." << endl;
		return;
	}
	
	pGDisplay display(renderer->NewDisplay());
	if(display == NULL)
	{
		cout << "ERROR: Could not get a display." << endl;
		return;
	}
	if(!display->Open(resX, resY, 0, fullscreen))
	{
		cout << "ERROR: Could not open display." << endl;
		return;
	}
	if(captureMouse)
		display->CaptureMouse(true);
	
	pGTexture texture = renderer->NewTexture();
	if(texture != NULL)
		texture->Image(image);
	
	pGScene   scene(renderer->NewScene());
	TestScene test(scene, model, texture, lightLevel);
	gTheTestScene = &test;
	pGCamera  camera(scene->NewCamera());
	camera->Direction(GQuaternion(GVECTOR_I, -PI_2));
	
	display->Camera(camera);
	
	GEngine::MainLoop();
}

} // namespace GAUGE3D

void PrintUsage()
{
	cout << "Usage:\ngauge3dfe [OPTIONS]" << endl;
	cout << "OPTIONS can be any combination of:" << endl;
	cout << "  -res <x> <y>    Set the display resolution to <x> by <y>." << endl;
	cout << "  -f              Display fullscreen." << endl;
	cout << "  -m <file>       Use model file <file>." << endl;
	cout << "  -i <file>       Use image file <file>." << endl;
	cout << "  -color <red> <green> <blue>  Set the color of the main object" << endl;
	cout << "  -alpha <value>  Set the opacity of the main object" << endl;
	cout << "  -nolight        Turn off lighting." << endl;
	cout << "gauge3dfe will display the given model using the given image as a texture." << endl;
	cout << "Plugins must be available which can load the model and texture formats." << endl;
	cout << "If no model is given, a sphere is used." << endl;
	cout << "If no texture is given, a simple checkerboard texture is used." << endl;
	cout << "When running, you can press up and down to increase and decrease the" << endl;
	cout << " geometric level of detail of the model.  Note that some models do not" << endl;
	cout << " support this." << endl;
}

int
main(int argc, char* argv[])
{
	bool fullscreen = false;
	bool captureMouse = false;
	int lightLevel = 1;
	GAUGE3D::GString model_filename, image_filename;
	int resX = 800, resY = 600;
	
	for(int i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "-f"))
			fullscreen = true;
		else if(!strcmp(argv[i], "-m"))
			model_filename = argv[++i];
		else if(!strcmp(argv[i], "-i"))
			image_filename = argv[++i];
		else if(!strcmp(argv[i], "-nolight"))
			lightLevel = 0;
		else if(!strcmp(argv[i], "-specular"))
			lightLevel = 2;
		else if(!strcmp(argv[i], "-capture"))
			captureMouse = true;
		else if(!strcmp(argv[i], "-res"))
		{
			resX = int(GAUGE3D::GString(argv[++i]));
			resY = int(GAUGE3D::GString(argv[++i]));
		}
		else if(!strcmp(argv[i], "-color"))
		{
			GAUGE3D::gObjectColor[0] = float(GAUGE3D::GString(argv[++i]));
			GAUGE3D::gObjectColor[1] = float(GAUGE3D::GString(argv[++i]));
			GAUGE3D::gObjectColor[2] = float(GAUGE3D::GString(argv[++i]));
		}
		else if(!strcmp(argv[i], "-alpha"))
		{
			GAUGE3D::gObjectColor[3] = float(GAUGE3D::GString(argv[++i]));
		}
		else
		{
			PrintUsage();
			return 1;
		}
	}
	
	GAUGE3D::GEngine::Init(GAUGE3D::GEngine::VIDEO |
	                       GAUGE3D::GEngine::INPUT |
	                       GAUGE3D::GEngine::OBJECT_LOADER);
	
	GAUGE3D::TestFrontend(resX, resY, fullscreen, lightLevel, model_filename,
	                      image_filename, captureMouse);
	
	GAUGE3D::GEngine::Shutdown();
	
	return 0;
}
